package nxm.audio.net;

import java.rmi.*;
import java.rmi.server.*;
import nxm.audio.lib.*;

/**
    Remote player object for RMI communication between Mixer and Player.
    A stub of this object is returned when the Mixer does a lookup on the player.

    @author Lisa Tursky
    @version $Id: RemotePlayerDelegate.java,v 1.3 2013/05/23 20:45:26 ntn Exp $
*/
public class RemotePlayerDelegate extends UnicastRemoteObject
                                 implements PlayerIntf {

  static final long serialVersionUID = 5578538881323037726L;

  private playerdaemon controller;
  private hwplayer thePlayer;

  public RemotePlayerDelegate(playerdaemon playerThread) throws RemoteException {
    controller = playerThread;
    thePlayer = playerdaemon.getPlayer();
  }

/************* methods for Player interface *****************/

   /**
    * Returns the name that is used to bind the player in the rmi registry.
    */
    public String getName() throws RemoteException {
        return "HWPlayer";
    }

   /**
    * Sets the host and port for the audio stream played on the given channel
    * of the hwplayer.
    */
    public void setHostPort(String host,int port,int channel,String audiotype,boolean on)
        throws RemoteException {
        controller.setHostPort(host, port, channel, audiotype, on);
    }

    public void shutdown() throws RemoteException {

        thePlayer.shutdown();
        thePlayer = null;
        controller.shutdown();
        controller = null;
    }

   /**
    * Closes the player.
    */
    public void close() throws RemoteException {

        controller.close();
        controller = null;
        thePlayer = null;
    }

   /**
    * Prints out the channel number, host, and port information for all
    * streams currently being played on the hwplayer. For debug purposes.
    */
    public String getPlayerChannelInfo() throws RemoteException {
        return controller.getPlayerChannelInfo();
    }

   /**
    * Adds a datagramreader to the hwplayer playing the audio stream at host:port.
    * Returns the channel number of the stream.
    */
    public int addReader(String host,int port,String audiotype,int bufferSize) throws RemoteException {
        return controller.addReader(host,port,audiotype,bufferSize,1);
    }

   /**
    * Removes the give channel from the hwplayer.
    */
    public void removeReader(int channel) throws RemoteException {
        controller.removeReader(channel);
    }

    /**
    * Test to see if Alive
    */
    public void isAlive() throws RemoteException { }

   /**
    * Get the current volume value for the indicated channel in the player.
    */
    public float getVolume(int lineNum) throws RemoteException {
        return thePlayer.getVolume(lineNum);
    }

   /**
    * Get the current pan value for the indicated channel in the player.
    */
    public float getPan(int lineNum) throws RemoteException {
        return thePlayer.getPan(lineNum);
    }

   /**
    * Get the current Off value for the indicated channel in the player.
    */
    public boolean getOff(int lineNum) throws RemoteException {
        return thePlayer.getOff(lineNum);
    }

   /**
    * Sets the volume to value for the indicated channel in the player.
    */
    public void setVolumeRemote(int volume,int lineNum) throws RemoteException {
        setVolumeRemote(volume, lineNum, false);
    }

   /**
    * Sets the volume to value for the indicated channel in the player.
    * Use the defined conversion scheme if the boolean convert is true.
    */
    public void setVolumeRemote(int volume,int lineNum, boolean convert)
        throws RemoteException {
        double cvol = volume;
        if (convert) {
            if(volume > 50) {
                cvol = (volume-50)*.40;
            } else {
                cvol = (volume*.80)-40;
            }
            volume = (int)cvol;
        }

        controller.setVolume(volume, lineNum);
    }

   /**
    * Sets the pan for the indicated channel in the player.
    */
    public void setPanRemote(float pan,int lineNum) throws RemoteException {
        setPanRemote(pan, lineNum, false);
    }

   /**
    * Sets the pan for the indicated channel in the player.
    * Uses the defined conversion scheme if the boolean convert is true.
    */
    public void setPanRemote(float pan,int lineNum, boolean convert) throws RemoteException {
        double p = pan;
        if (convert) {
            p = pan / 100.0;
        }
        pan = (float)p;
        controller.setPan(pan, lineNum);
    }

   /**
    * Sets the off for the indicated channel in the player.
    */
    public void setOffRemote(boolean off, int lineNum) throws RemoteException {
        controller.setOff(off, lineNum);
    }

}
