package nxm.sys.lib;

import java.io.*;
import java.net.*;

import nxm.sys.inc.InternalUseOnly;

/** A new resource that will used the current Midas Context's ClassLoader to access the files in
    the classpath.

    NOTE: This new resource type will provide read-only access consistent with the Java
    ClassLoader to access the files. Becuase we are using other resources according to there
    URL type. Some of this resources may have read/wride access.
    @author  Juan Velasquez
    @since NeXtMidas 3.7.1
 */
public class ClassPathResource extends IOResource {
  private URL newurl;
  private String path_name;
  // IOResource to be used when URL is found on Classpath
  private IOResource resource;

  public ClassPathResource() { }

  /**
   * This method calls the getResource method to find the
   * URL on the java ClassLoader. If the URL path is found it
   * uses the CoreIO class to find the IOResource to used according
   * to the type of file from the URL.
   *
   */
  @Override
  public void setURL(String urlstr) {
    path_name = urlstr.replaceFirst("^\\w+:", ""); // retrieve only the path specified on CLASSPATH:
    newurl = this.getResource(path_name, this.getClass());
    if (newurl != null) {
      urlstr = newurl.toExternalForm();
      // Used the Midas CoreIO to find the resource needed for the URL found on the classpath
      resource = M.getIO().findResource(urlstr, DIR_INOUT);
    }
    else { M.warning("ClassPathResource: Error accessing Resource URL="+getURL()); }
  }

  /** test existence of the resource */
  @Override
  public boolean exists() {
    return resource.exists();
  }

  /** open the resource */
  @Override
  public boolean open () {
    boolean op = false;
    try {
      op = resource.open();
    } catch (Exception e){
      M.warning(e);
    }
    return op;
  }

  /** read the resource data */
  @Override
  public int read (byte[] buf, int boff, int bytes) {
    int read = -1;
    try {
      read = resource.read(buf, boff, bytes);
    } catch(Exception e) {
      M.warning(e);
    }
    return read;
  }

  /**
   * write the result. Currently may not be supported following the java ClassLoader
   * read-only access. It depends on the type of IOResource URL open.
   */
  @Override
  public int write (byte[] buf, int boff, int bytes) {
    int write= -1;
    try {
      write = resource.write(buf, boff, bytes);
    } catch (Exception e){
      M.warning(e);
    }
    return write;
  }

  /**
   * Set the current seek offset
   * @param off the offset value
   * @return true on success, false on failure.
   */
  @Override
  public boolean seek (long off) {
    boolean seek_flag = false;
    try {
      seek_flag = resource.seek(off);
    } catch(Exception e) {
      M.warning(e);
    }
    return seek_flag;
  }

  /** Get List of container entries. (This is used by the <tt>FILES</tt> command.)
  @param filter A filter to select matching entries (null=AcceptAll).
  @return The entries that match the filter or null if this is not a container.
  */
  @Override
  public String[] getEntryList (FilenameFilter filter) {
    String[] entry = null;
    try {
      entry = resource.getEntryList(filter);
    } catch (Exception e) {
      M.warning(e);
    }
    return entry;
  }

  /** close the resource */
  @Override
  public boolean close() {
    boolean close = false;
    try {
      close = resource.close();
      newurl=null;
      path_name="";
      resource=null;
    } catch (Exception e) {
      M.warning(e);
    }
    return close;
  }

  /** Method for getting the length of the resource file */
  @Override
  public long getLength() {
    return resource.getLength();
  }

  @Override
  public long avail () {
    return resource.avail();
  }

  @Override
  public long lastModified () {
    return resource.lastModified();
  }

  /** @return the InputStrem from the loaded resource */
  @Override
  public InputStream  getInputStream()  { return resource.getInputStream(); }

  /** Return the OutputStream if there is one available */
  @Override
  public OutputStream getOutputStream () { return resource.getOutputStream(); }


  /** Return a string containing a description of this object. */
  @Override
  public String toString () {
    return "ClassPathResource: " + resource.toString();
  }

  /** get entry from Container */
  @Override
  public IOResource getEntry (String name) { return resource.getEntry(name); }

  @Override
  public boolean isContainer() { return resource.isContainer(); }

  /**
   * This is a convenience method to load a resource as a stream. <p/> The
   * algorithm used to find the resource is given in getResource()
   *
   * @param resourceName The name of the resource to load
   * @param callingClass The Class object of the calling object
   *
  */
  @InternalUseOnly
  private InputStream getResourceAsStream(String pathName, Class<?> callingClass) {
    URL url = getResource(pathName, callingClass);
    try {
        return (url != null) ? url.openStream() : null;
    } catch (IOException e) {
        M.warning(e);
        return null;
    }
  }

  /**
   * Load a given resource. This method will try to load the resource
   * using the following methods (in order):
   *
   * Thread.currentThread().getContextClassLoader()
   * Shell.class.getClassLoader()
   * callingClass.getClassLoader()
   *
   * @param pathName The name of the resource to load
   * @param callingClass The Class object of the calling object
   *
   */
  @InternalUseOnly
  private URL getResource(String pathName, Class<?> callingClass) {
    // the current thread classLoader
    URL url = Thread.currentThread().getContextClassLoader().getResource(pathName);
    if (url == null && pathName.startsWith(File.separator)) {
      url = Thread.currentThread().getContextClassLoader()
          .getResource(pathName.substring(1));
    }

    // the NeXtMidas shell classLoader
    ClassLoader cluClassloader = Shell.class.getClassLoader();
    if (cluClassloader == null) {
      cluClassloader = ClassLoader.getSystemClassLoader();
    }
    if (url == null) {
      url = cluClassloader.getResource(pathName);
    }
    if (url == null && pathName.startsWith(File.separator)) {
      url = cluClassloader.getResource(pathName.substring(1));
    }


    // the current class classLoader
    if (url == null) {
      ClassLoader cl = callingClass.getClassLoader();

      if (cl != null) {
        url = cl.getResource(pathName);
      }
    }

    if (url == null) {
      url = callingClass.getResource(pathName);
    }

    if ((url == null) && (pathName != null) && (pathName.charAt(0) != File.pathSeparatorChar)) {
      return getResource(File.separator + pathName, callingClass);
    }

    return url;
  }
}