package nxm.sys.lib;

import nxm.sys.inc.*;
import java.text.DecimalFormat;
import java.text.FieldPosition;
import java.lang.Math;

/** Class to perform alignable Decimal Formatting with Units. This is used internally by the numeric
    valuators and has special methods used for character alignment in a GUI. Most users should use
    the {@link MFormat} class for general formatting.

    @author Jeff Schoen
    @version $Id: Format.java,v 1.23 2014/10/01 20:07:56 ntn Exp $
    @see MFormat
*/
public class Format extends DecimalFormat {

  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 5256251046960840573L;
  private String pattern;
  private String units = "";
  private FieldPosition fp = new FieldPosition(INTEGER_FIELD);
  private int fones, ones, exp, imult, digitSel;
  private double mult=1, digitInc; // <-- digitInc, is not used?
  private char cmult=' ';
  private boolean hms,isSelected;
  private boolean dms;

  public Format () {
  }

  /** Constructor with given pattern
      @param pattern A Java format pattern or HH:MM:SS or DMS
  */
  public Format (String pattern) {
    this();
    this.setFormat(pattern);
  }

  /** Set the format to a Java pattern or a special pattern.
     <pre>
       Special patterns and characters are:
          HH:MM:SS - Show in hours:minutes:seconds
          DMS      - Show in deg'min'sec format (i.e 33'22'11)
          ?        - Question marks are replaced with blanks and trigger the
                     use of a multiplier

       Units are defined as the String to the right of the last blank space in the
       input String. (e.g. "## MHz" the units are MHz).
     </pre>
    @param pattern A Java format pattern or HH:MM:SS or DMS
  */
  public void setFormat (String pattern) {
    this.pattern = pattern;
    hms = pattern.equals("HH:MM:SS");
    dms = pattern.equals("DMS");
    fones = pattern.indexOf('.');
    if (fones<0) fones = pattern.lastIndexOf('#')+1;
    if (fones<=0) fones = pattern.lastIndexOf('0')+1;
    imult = pattern.indexOf('?');
    if (imult>=0) pattern = pattern.replace('?',' ');
    if (imult<0) imult = 0;
    if (!hms && !dms) applyPattern(pattern);
    int space = pattern.lastIndexOf(' ');
    if (space >=0 ) units=pattern.substring(space+1);
  }

  /** Get the pattern string. */
  public String getFormat() { return pattern; }

  /** Sets the exponent in the given a number or multiplier string.
     <pre>
       Arguments can be passed in three ways:
         1) n, where n is one of {-12,-9,-6,-3,3,6,9,12}: (e.g. 6 is Mega)
         2) One of { "pico", "nano", "micro", "milli","Kilo", "Mega", "Giga", "Tera" }
         --OR--
         3) One of {'p','n','u','m','K','M','G','T'}
     </pre>
    @param obj Number or string representing the  multiplier
  */
  public void setExp (Object obj) {
    int imult = 0;
    if (obj instanceof Data) {
      int val = ((Data)obj).toL();
      if (val < -12 || val > 12 || (val%3 > 0)) {
        Shell.warning("Invalid argument for Format.setExp: " + obj);
        return;
      }
      else imult = val;
    }
    else if (obj instanceof String) {
      int index = -1;
      String arg = ((String)obj).trim();
      if (arg.length() > 1) {
        for (int i = 0; i < Units.expNames.length ; i++) {
          if (Units.expNames[i].equalsIgnoreCase(arg)) {
            imult = Units.expValues[i];
            index = i;
            break;
          }
        }
      }
      else {
        char cmult = arg.charAt(0);
        for (int i = 0; i < Units.expAbbr.length ; i++) {
          if (Units.expAbbr[i] == cmult) {
            imult = Units.expValues[i];
            index = i;
            break;
          }
        }
      }
      if (index == -1) {
        Shell.warning("Invalid argument for Format.setExp: " + arg);
        return;
      }
    }
    setExp(imult);
  }

  /** Get the exponent */
  public int getExp () { return exp; }

  /** Set the best engineering exponent for a range of numbers
    @param dmin Minimum of the range
    @param dmax Maximum of the range
  */
  public void setExp (double dmin, double dmax) {
    double absmax; int kengr;
    absmax = Math.max( Math.abs(dmin),Math.abs(dmax) );
    if (absmax == 0.0) kengr=0;
    else {
      kengr = (int)( 0.1447648 * Math.log(absmax*.5) );
      //if (absmax<1.0) kengr--;
    }
    setExp (kengr*3);
  }

  /** Set the exponent.  If not &gt;-12, &lt;12 and a multiple of 3, the multiplier character is '?'.
    @param exp the exponent
  */
  public void setExp (int exp) {
    if (imult==0) exp = 0;
    this.exp = exp;
    mult = Math.pow(10.0,-exp);
    switch (exp) {
     case  0:  cmult=' '; break;
     case  3:  cmult='k'; break;
     case  6:  cmult='M'; break;
     case  9:  cmult='G'; break;
     case 12:  cmult='T'; break;
     case -3:  cmult='m'; break;
     case -6:  cmult='u'; break;
     case -9:  cmult='n'; break;
     case -12: cmult='p'; break;
     default:  cmult='?';
    }
  }

  /** Get the multiplier in String form. */
  public String getCmult() {
    if (cmult == ' ') {
      return "";
    } else  return "" + cmult;
  }

  /** Get the units in string form. */
  public String getUnits() {
    return units;
  }

  /** The un-overridable format() call to format a double number.
      @param dval Number to be formatted
   */
  public String form (double dval) {
    String s;
    if (hms) {
      s = Time.toString(dval,Time.STD,0);
      ones = s.length();
    }
    else if (dms) {
      s = Convert.deg2dmsUnicode(dval);
      ones = 2;
    }
    else {
      StringBuffer sb = new StringBuffer(pattern.length());
      sb = super.format(dval*mult,sb,fp);
      ones = fp.getEndIndex();
      int i = imult + sb.length() - pattern.length();
      if (imult==0);
      else if (exp==0) sb.deleteCharAt(i);
      else sb.setCharAt(i,cmult);
      s = sb.toString();
      i = s.indexOf('E');
      if (imult==0 && i>=0) {
        exp = Integer.parseInt(s.substring(i+1));
        if (s.endsWith("E0")) s = s.substring(0,i);
      }
    }
    return s;
  }

  /** Increment the selected digit.
    www.fffEee
    digit is the position of the dot
    Fractional digits (negative) are to the right
    Whole digits (positive) are to the left
  */
  public void incDigitPos () { digitSel++; }

  /** Decrement the selected digit.
    www.fffEee
    digit is the position of the dot
    Fractional digits (negative) are to the right
    Whole digits (positive) are to the left
  */
  public void decDigitPos () { digitSel--; }

  /** Set the digit position.
    @param pos the digit pos
  */
  public void setDigitPos (int pos) {
    int digit = ones-pos+exp;
//System.out.println("Set pos ds="+digitSel+" d="+digit+" e="+exp+" o="+ones+" p="+pos);
    if (pos<ones) digit--;
    isSelected = pos>=0 && pos!=ones && !(isSelected && digit==digitSel);
    digitSel = digit;
  }

  /** Get the digit position of currently selected digit. */
  public int getDigitPos() {
    if (!isSelected) return -1;
    int digit = (digitSel-exp);
    int pos = ones-digit;
    if (digit>=0) pos--;
//System.out.println("Digit pos ds="+digitSel+" d="+digit+" e="+exp+" o="+ones+" p="+pos);
    return pos;
  }

  /** Get the digit increment.  Returns 0 is not selected. */
  public double getDigitInc() {
    return (isSelected)? getDigitInc(digitSel-exp) : 0;
  }
  /** Test whether a digit is selected. */
  public boolean isDigitSel() { return isSelected; }

  /** Get the digit position given a digit (ones - digit) . */
  public int getDigitPos (int digit) { return ones-digit; }

  /** Get the digit index (ones-pos). */
  public int getDigitIndex (int pos) { return ones-pos; }

  /** Gets the increment that corresponds to the given digit. Digit is relative to the ones
      place and 0 corresponds to the ones place and a negative digit indicates right of the ones
      place. For HMS the increment is in seconds and for DMS the increment is in degrees.
    @param digit The digit place 0 to n
  */
  public double getDigitInc (int digit) {
    double inc=0;
//jph -- methinks we do not need this because the else clause contains the same equation
//    if (digit<0) inc = Math.pow(10.0,exp+digit);
//    else
    if (hms) switch (digit) {
        case 0: inc=1; break;
        case 1: inc=10; break;
        case 3: inc=60; break;
        case 4: inc=600; break;
        case 6: inc=3600; break;
        case 7: inc=36000; break;
        case 10: inc=86400; break;
        case 11: inc=864000; break;
    }
    else if (dms) switch (digit) {
        case  0: inc=1;       break;
        case  1: inc=10;      break;
        case -1: inc=1/6.0;   break;
        case -2: inc=1/60.0;  break;
        case -4: inc=1/360.0; break;
        case -5: inc=1/3600.0;break;
    }
    else inc = Math.pow(10.0,exp+digit);
    return inc;
  }

  /** @deprecated Since NeXtMidas 2.5.0: Use the {@link MFormat} class. */
  @Deprecated public static Format getDefaultFormatFor (String format, int units) {
    // final char mode = format.charAt(0);
    final char type = format.charAt(1);
    Format fmt = null;

    if (type == 'A') { /* ASCII/String */
      fmt = null;

    } else {
      String unitsStr = DataFile.getUnitsNameShort(units);
      String floatStr = null; /* The format to use for float/double values. */
      String intStr   = null; /* The format to use for long/int/short/byte values. */

      switch(units) {
        case Units.LONGITUDE:  /* FALLTHROUGH */
        case Units.LATITUDE:   floatStr = "##0.0000";      intStr = "###"; unitsStr = ""; break;
        default:               floatStr = "##0.#####E000"; intStr = "0";                  break;
      }

      if (unitsStr.length() > 0) {
        unitsStr = "'" + unitsStr + "'"; /* Protect the units. */
      }

      if ((type == 'D') || (type == 'F')) {
        fmt = new nxm.sys.lib.Format(floatStr + unitsStr);
      } else {
        fmt = new nxm.sys.lib.Format(intStr + unitsStr);
      }
    }

    return fmt;
  }
}
