package nxm.sys.lib;

import java.io.FilenameFilter;

/**
  Resource for interacting with a NeXtMidas Pipe.

  @author Jeff Schoen
  @version $Id: PipeResource.java,v 1.30 2013/03/12 14:43:44 ntn Exp $
*/
public class PipeResource extends IOResource {

  /** the Pipe entity */
  public Pipe pipe;
  /** 0=writer, 1:N for readers */
  public int outlet;
  /** key in pipe table */
  public String name;
  /** parent DataFile */
  public DataFile df;
  /** default pipe size (in bytes) */
  public int pipeSize;
  /** default pipe pause time (in seconds)*/
  public double pause=0.0125;
  public static boolean POLLING = true;

  /** initial reader's seek position into pipe (set on first connect in {@link #initTimecode()}. */
  private double initialConnectSeekPos = 0;

  public PipeResource () { }

  /** return if this resource is a stream resources */
  public boolean isStream () { return true; }

  /** return if this resource is currently streaming */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  public boolean isStreaming () {
    if (pipe==null) return true; // not connected yet
    DataFile df = pipe.getDataFile(outlet);
    if (df==null) return true; // in transition
    if (df.isOpen) return true; // still writing
    else if (df.getNextDataFile()!=null) return false; // no input, but chained
    else if (pipe.inlet==0) return false; // all done
    else return true;	// no input, but not ready to chain
  }

  /** test existence of the resource */
  public boolean exists () {
    name = urlstr.substring(4).toUpperCase();
    return M.pipes.containsKey(name);
  }

  /** open the resource */
  public synchronized boolean open () { // since 3.1.1 added synchronized to address BUG 2457
    // 2010-09-13 JWB: The line below was MOVE from the two blocks below to here because if there
    // Since 2.9.0:    is a restart of an output pipe then the dir could be >0 and
    //                 the pipe could be null but still exist in the pipes table. See BUG 1166.
    pipe = (Pipe) M.pipes.get(name);

    if (dir <= DIR_INOUT) { // if dir is for INOUT(0), INPUT(-1), INPUT_TRUNC(-2), etc.
      //pipe = (Pipe) M.pipes.get(name);  // MOVED above - See 2010-09-13 comment above
      outlet = pipe.attach(dir,df);
      initTimecode();
    }
    else if (pipe==null) {  // this happens when dir is for OUPUT(1) or OUTPUT_TRUNC(2)
      int maxOutlets = Pipe.DEFAULT_MAX_OUTLETS;
      if (qualTable != null) {
        maxOutlets = qualTable.getL("MAXOUTLETS", maxOutlets); // since 3.3.2 to allow setting max number of readers on a pipe
      }
      pipe = new Pipe(M, name, maxOutlets); // since 3.3.2: pass maxOutlet to allow setting max number of readers
      pipe.setSize(pipeSize);
      outlet = pipe.attach(dir,df);
      M.pipes.put (name,pipe);
    }
    else {                 // have existing pipe (on a RESTART)
      //pipe = (Pipe) M.pipes.get(name);  // MOVED above - See 2010-09-13 comment above
      outlet = pipe.attach(dir,df);
      M.pipes.put (name,pipe);
    }
    pause = M.pause;
    if (M.macro!=null) pause = M.macro.lines[1].getD("/POLL",pause);
    isOpen = (outlet >= 0);
    return isOpen;
  }

  /** close the resource */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  public synchronized boolean close () { // since 3.1.1 added synchronized to address BUG 2457
    if (pipe==null) return true;
    pipe.detach(outlet,df);
    if (outlet==0 && pipe.inlet==0) M.pipes.remove(name);
    return true;
  }

  /** delete the resource */
  public boolean delete () {
    M.pipes.remove(name);
    return true;
  }

  /** get length of resource */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  public long getLength () {
    return (long)pipe.inPtr;
  }

  /** seek to the specified offset in file */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  public boolean seek (long offset) {
    if (outlet>0) pipe.outPtr[outlet]=(double)offset;
    this.offset = offset;
    return true;
  }

  /** get the current seek offset */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  public long seek () {
    if (outlet>0) return (long)pipe.outPtr[outlet];
    else return (long)pipe.inPtr;
  }

  /** get currently available length of resource */
  public long avail() {
//System.out.println("avail = "+(long)pipe.avail(outlet));
    return (long)pipe.avail(outlet);
  }

  @Override
  public boolean isReady (int numBytes) {
    return pipe.isReady(outlet, numBytes);
  }

  /** read data from resource */
  public int read (byte[] buf, int boff, int bytes) {
    return read(buf,0L,boff,bytes);
  }

  /** write data to resource */
  public int write (byte[] buf, int boff, int bytes) {
    return write(buf,0L,boff,bytes);
  }

  /** read data from resource */
  public int read (long lbuf, int boff, int bytes) {
    return read(null,lbuf,boff,bytes);
  }

  /** write data to resource */
  public int write (long lbuf, int boff, int bytes) {
    return write(null,lbuf,boff,bytes);
  }

  /** read data from resource */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  private int read (byte[] buf, long lbuf, int boff, int bytes) {
    int status=0, off=boff;
    while (bytes>0) {
      status = pipe.read(buf,lbuf,off,bytes,outlet);
      if (status>0) { bytes-=status; off+=status; }
      else if (status<0) break;
      else if ((df.cmd != null) && (df.cmd.isStateChanged())) break; // Added in 2.5.0
      else if (POLLING) Time.sleep(pause);
      else try { pipe.wait(); } catch (Exception e) { };
    }
    offset = (long)pipe.outPtr[outlet];
    if (!POLLING) pipe.awaken();
    if (status<0 && off==boff) return status;
    propagateTimecode();  // JC4: Added in 2.5.0 - propagate timecode for the pipe
    return off-boff;
  }

  /** write data to resource */
  @SuppressWarnings("deprecation") // for usage of pipe fields what should become package-private
  private int write (byte[] buf, long lbuf, int boff, int bytes) {
    int status=0, off=boff;
    while (bytes>0) {
      status = pipe.write(buf,lbuf,off,bytes,outlet);
      if (status>0) { bytes-=status; off+=status; }
      else if (status<0) break;
      //else if ((df.cmd != null) && (df.cmd.isStateChanged())) break; // Added in 2.5.0 <-- JC4: I think we want this, but I don't have a test case for it.
      else if (POLLING) Time.sleep(pause);
      else try { pipe.wait(); } catch (Exception e) { };
    }
    offset = (long)pipe.inPtr;
    if (!POLLING) pipe.awaken();
    if (status<0 && off==boff) return status;
    return off-boff;
  }

  /** get List of container entries */
  public String[] getEntryList (FilenameFilter filter) {
    return filterEntries(filter, M.pipes.getKeys());
  }

  /** get entry from Container */
  public IOResource getEntry (String name) { return null; }

  /** @since NeXtMidas 3.3.0 */
  @Override
  public Time lastModifiedTime () {
    return new Time(pipe.lastModified);
  }

  /** @since NeXtMidas 3.3.0 */
  @Override
  public long lastModified () {
    return (long) Time.toJ1970(pipe.lastModified); // convert J1950 to J1970
  }

  /** implement connect function (attach/detach from pipe) */
  public void connect (int flag) {
    if (flag>0 && outlet==Pipe.DISCON) outlet = pipe.attach(dir,df);
    if (flag<0 && outlet>0) outlet = pipe.detach(outlet,df);
  }

  public int readHeader (byte[] hb) {
    return pipe.readHeader (hb,outlet);
  }

  public int writeHeader (byte[] hb) {
    return pipe.writeHeader (hb,outlet);
  }

  /** Initialize the timeline in a pipe reader. This is only called by {@link #open()} for a pipe reader
      AFTER the call to attach().
   */
  private void initTimecode () {
    DataFile previous = df.getPreviousDataFile();

    if (previous == null) { // Outlet on first connect
      df.newTimeLineFromDF(pipe.getDataFile(outlet));
      initialConnectSeekPos = pipe.getDataFile(outlet).seek(); // 2010-12-20 NTN: for BUG 2265
      propagateTimecode();

    } else {                // Outlet or Inlet on reconnect
      Time stopTime  = previous.getTimeAt(previous.seek(), null);
      Time startTime = stopTime.addSec(previous.getDelta());

      df.newTimeLineFromDF(previous);
      df.setTimeAt(0.0, startTime);
    }
  }

  /** Propagate the timecode from pipe writer to pipe reader. */
  private void propagateTimecode () {
    DataFile in = pipe.getDataFile(outlet); // Inlet to the pipe

    double seekBytes = df.seek() * df.getBPE();             // Outlet's seek in bytes
    double seekElem  = Math.floor(seekBytes / in.getBPE()); // Inlet's position in elements
    Time time = in.getTimeAt(initialConnectSeekPos+seekElem, null); // Time at current seek() + initialConnectSeekPos (2010-12-20 for BUG 2265)

    df.setTimeAt(time);
  }
}
