package nxm.sys.lib;

import java.awt.Color;
import java.util.Vector;

import nxm.sys.inc.MidasReference;
import nxm.sys.inc.PlotFile;
import nxm.sys.inc.Units;
import nxm.sys.libg.Layer;
import nxm.sys.libg.LayerShape;

/** A class to handle ESRI Shape files. This file matches the "ESRI Shapefile Technical Description"
    dated July 1998 and corresponding to files with the version number matching {@link #VERSION}.

    @author Jeff Schoen
    @version $Id: ShapeFile.java,v 1.18 2014/04/15 15:38:16 ntn Exp $
*/
public class ShapeFile extends BaseFile implements PlotFile {

  /** ShapeFile Code Number (value={@value}) */
  public static final int FILECODE=9994;

  /** @deprecated Since NeXtMidas 2.3.1: Use {@link #FILECODE} */
  @Deprecated public static final int SHAPEFILE_ID=FILECODE;

  /** ShapeFile Version Number (value={@value}) */
  public static final int VERSION=1000;

  /** ShapeFile Shape Type: Null Shape (ID={@value}) */
  public static final int NULLSHAPE = 0;

  /** ShapeFile Shape Type: Point (ID={@value}) */
  public static final int POINT = 1;

  /** ShapeFile Shape Type: PolyLine (ID={@value}) */
  public static final int POLYLINE = 3;

  /** ShapeFile Shape Type: Polygon (ID={@value}) */
  public static final int POLYGON = 5;

  /** ShapeFile Shape Type: MultiPoint (ID={@value}) */
  public static final int MULTIPOINT = 8;

  /** ShapeFile Shape Type: PointZ (ID={@value}) */
  public static final int POINTZ = 11;

  /** ShapeFile Shape Type: PolyLineZ (ID={@value}) */
  public static final int POLYLINEZ = 13;

  /** ShapeFile Shape Type: MultiPointZ (ID={@value}) */
  public static final int MULTIPOINTZ = 18;

  /** ShapeFile Shape Type: PointM (ID={@value}) */
  public static final int POINTM = 21;

  /** ShapeFile Shape Type: PolyLineM (ID={@value}) */
  public static final int POLYLINEM = 23;

  /** ShapeFile Shape Type: PolygonM (ID={@value}) */
  public static final int POLYGONM = 25;

  /** ShapeFile Shape Type: MultiPointM (ID={@value}) */
  public static final int MULTIPOINTM = 28;

  /** ShapeFile Shape Type: MultiPatch (ID={@value}) */
  public static final int MULTIPATCH = 31;

  /** @deprecated Since NeXtMidas 2.3.1: Use {@link #POLYLINE} */
  @Deprecated public static final int ARC=POLYLINE;

  /** @deprecated Since NeXtMidas 2.3.1: Use {@link #POLYLINEM} */
  @Deprecated public static final int ARC_M=POLYLINEM;

  private int     fileCode;    // Internal file code for the file (see FILECODE)
  private int     fileLength;  // Length of the file (number of 16-bit words) including header
  private int     fileOffset;  // Read offset into the file (starts at 50=LengthOfHeader)
  private int     version;     // ShapeFile Version Number (see VERSION)
  public  int     shapeType;   // Shape file type code
  public  double  x1;          // XMin from bounding box in ShapeFile header
  public  double  y1;          // YMin from bounding box in ShapeFile header
  public  double  x2;          // XMax from bounding box in ShapeFile header
  public  double  y2;          // YMax from bounding box in ShapeFile header
  private double  timecode;    // Timecode set with setTimeCode(..), not in the ShapeFile.
  public  DbfFile dbf;         // Accompanying DBF file.
  private boolean dbfOK;       // Use the DBF file?

  protected byte[] ghb = new byte[1024];

  /** Creates a new instance. */
  public ShapeFile () { }

  /** Creates a new instance.
      @deprecated since NeXtMidas 3.5.0 use {@link #ShapeFile(MidasReference, Object)}
   */
  @Deprecated
  public ShapeFile (Object ref, Object filename) {
    if (ref instanceof MidasReference) {
      init((MidasReference) ref, filename);
    } else {
      init(Convert.ref2Midas(ref), filename);
    }
  }

  /** Creates a new instance.
      @since NeXtMidas 3.5.0
   */
  public ShapeFile (MidasReference ref, Object filename) {
    init(ref,filename);
  }


  public String toString() {
    String str = "ShapeFile ";
    if (io!=null) str += io.getTypeString();
    str += " Resource";
    return str;
}

  public void connect (int mode) { }

  public Data getDataBuffer (int size) { return null; }

  /** Gets the internal file code for the file (see {@link #FILECODE}). To get the shape type
      use {@link #getShapeType()}.
   */
  public int getType() { return fileCode; }

  /** Always returns 1 (scalar data). */
  public int getMode() { return 1; }

  /** Currently does nothing and returns -1. */
  public int read (Data data, int ndo) { return -1; }

  public String listHeader() {
    String list =
      "ShapeFile   :  "+getURL()+"\n"
    + "Code        :  "+fileCode+"\n"
    + "Version     :  "+version+"\n"
    + "Length      :  "+fileLength+"\n"
    + "Size        :  "+getSize()+" bytes\n"
    + "Type        :  "+getShapeType()+"\n"
    + "X1,X2       :  ("+x1+","+x2+")\n"
    + "Y1,Y2       :  ("+y1+","+y2+")\n"
    + "DB File     :  "+dbf.filename+", exists="+dbfOK+"\n"
    + "Comment     :  "+getComment()+"\n";
    return list;
  }

  /** Gets the shape type, same as {@link #getShapeType()}. */
  public String getTypeString() {
    // Replaced this with getShapeType() to prevent confusion with getType() and
    // IOResource.getTypeString() which are NOT related to this method.
    return getShapeType();
  }

  /** Gets the shape type, returns "unknown" if not found. */
  public String getShapeType() {
    return getShapeType(shapeType,"unknown");
  }

  /** Maps a ShapeFile type code to its type string, returns null if not found. */
  public static String getShapeType(int type) {
    return getShapeType(type,null);
  }

  /** Maps a ShapeFile type code to its type string, returns def if not found. */
  private static String getShapeType(int type, String def) {
    String str = def;
    switch (type) {
      case NULLSHAPE:   str = "NullShape";   break;
      case POINT:       str = "Point";       break;
      case POLYLINE:    str = "PolyLine";    break;
      case POLYGON:     str = "Polygon";     break;
      case MULTIPOINT:  str = "MultiPoint";  break;
      case POINTZ:      str = "PointZ";      break;
      case POLYLINEZ:   str = "PolyLineZ";   break;
      case MULTIPOINTZ: str = "MultiPointZ"; break;
      case POINTM:      str = "PointM";      break;
      case POLYLINEM:   str = "PolyLineM";   break;
      case POLYGONM:    str = "PolygonM";    break;
      case MULTIPOINTM: str = "MultiPointM"; break;
      case MULTIPATCH:  str = "MultiPatch";  break;
    }
    return str;
  }

  /** Always returns "SD". */
  public String getFormat() { return "SD"; }

  public double getStart()  { return 0; }
  public double getDelta()  { return 1; }
  public int    getUnits()  { return Units.NONE; }
  public double getXStart() { return 0; }
  public double getXDelta() { return 1; }
  public int    getXUnits() { return Units.NONE; }
  public int    getXFrame() { return 1; }
  public double getYStart() { return 0; }
  public double getYDelta() { return 1; }
  public int    getYUnits() { return Units.TIME; }
  public int    getYFrame() { return 1; }

  /** Returns the TimeCode set with {@link #setTimeCode(double)}, not in the file. */
  public double getTimeAt(double offset) { return timecode; }

  public void setFormat (String value) { Shell.warning("ShapeFile: setFormat(v) not implemented"); }
  public void setXStart (double value) { Shell.warning("ShapeFile: setXStart(v) not implemented"); }
  public void setXDelta (double value) { Shell.warning("ShapeFile: setXDelta(v) not implemented"); }
  public void setXUnits (int value)    { Shell.warning("ShapeFile: setXUnits(v) not implemented"); }
  public void setXFrame (int value)    { Shell.warning("ShapeFile: setXFrame(v) not implemented"); }
  public void setYStart (double value) { Shell.warning("ShapeFile: setYStart(v) not implemented"); }
  public void setYDelta (double value) { Shell.warning("ShapeFile: setYDelta(v) not implemented"); }
  public void setYUnits (int value)    { Shell.warning("ShapeFile: setYUnits(v) not implemented"); }
  public void setYFrame (int value)    { Shell.warning("ShapeFile: setYFrame(v) not implemented"); }
  public void setTimeCode (double value) { timecode=value; }

  public boolean open() {
    boolean ok = super.open();
    if (!ok) return ok;
    byte[] hb = ghb;
    read (hb,0,100); // Read header
    Convert.swap4(hb,0,1);  // Special case: FileCode   is bigendian, swap to match rest of header
    Convert.swap4(hb,24,1); // Special case: FileLength is bigendian, swap to match rest of header
    fileCode   = Convert.unpackL(hb, 0,EEEI);
    fileLength = Convert.unpackL(hb,24,EEEI);
    version    = Convert.unpackL(hb,28,EEEI);
    shapeType  = Convert.unpackL(hb,32,EEEI);
    x1 = Convert.unpackD(hb,36,EEEI);
    y1 = Convert.unpackD(hb,44,EEEI);
    x2 = Convert.unpackD(hb,52,EEEI);
    y2 = Convert.unpackD(hb,60,EEEI);
    fileOffset = 50; // Header = 50 x 16-bit words

    FileName fn = getFileName();
    String dbfName = fn.getPath() + fn.getRoot() + ".dbf";
    dbf = new DbfFile(M, dbfName);
    try {
      dbfOK = dbf.open();
    } catch (MidasException e) {
      // this is ok, it doesnt have to exist
      dbfOK = false;
    }

    if (version != VERSION) {
      Shell.getSharedMidasContext().warning("ShapeFile: Shape file "+getName()+" has Version="+version
                                    + " but only Version="+VERSION+" is supported by ShapeFile.");
    }
    if (fileCode != FILECODE) {
      Shell.getSharedMidasContext().warning("ShapeFile: Shape file "+getName()+" has FileCode="+fileCode
                                    + " but only FileCode="+FILECODE+" is supported by ShapeFile.");
    }

    return ok;
  }

  @Override
  public void close() {
    super.close();
    if (dbf != null)
      dbf.close(); // close underlying file, BUG 2393
  }

  /** Gets the array of shapes in the file. */
  public Shape[] getShapes() {
    int numPolygons=0;
    Vector<Shape> v = new Vector<Shape>(1024);
    for (; fileOffset<fileLength; ) {
      Shape shape = new Shape();
      fileOffset += shape.readFrom(this);
      numPolygons += shape.numParts;
      v.addElement(shape);
    }
    Shape[] shapes = v.toArray(new Shape[0]);
    return shapes;
  }

  public Layer getLayer() {
    return new LayerShape();
  }

  /** Represents a shape in a ShapeFile. */
  public static class Shape {
    public double x1,x2,y1,y2;
    public int recordNumber, contentLength;
    public int shapeType, numParts, numPoints;
    public int[] parts;
    public double[] points;
    public int maxPartLength=0;
    public String label;

    /** The default color for the shape. */
    private Color color = null;

    public int getLength() {
      return (4 + contentLength);
    }

    public int readFrom (ShapeFile sf) {
      byte[] hb=sf.ghb;

      sf.read (hb,0,12);
      Convert.swap4(hb,0,2);
      recordNumber  = Convert.unpackL(hb,0,EEEI);
      contentLength = Convert.unpackL(hb,4,EEEI);
      shapeType     = Convert.unpackL(hb,8,EEEI);

      switch (shapeType) {
        case POINT:
          sf.read (hb,0,16);
          x1 = Convert.unpackD(hb,0,EEEI);
          y1 = Convert.unpackD(hb,8,EEEI);
          x2 = x1;
          y2 = y1;
          maxPartLength = 1;
          break;
        case POLYLINE:
          /* FALLTHROUGH */
        case POLYGON:
          sf.read (hb,0,40);
          x1 = Convert.unpackD(hb, 0,EEEI);
          y1 = Convert.unpackD(hb, 8,EEEI);
          x2 = Convert.unpackD(hb,16,EEEI);
          y2 = Convert.unpackD(hb,24,EEEI);
          numParts  = Convert.unpackL(hb,32,EEEI);
          numPoints = Convert.unpackL(hb,36,EEEI);
          parts = new int[numParts];
          points = new double[numPoints*2];
          if (numParts*4>hb.length) hb = new byte[numParts*4];
          sf.read (hb,0,numParts*4);
          for(int i = 0; i < numParts; i++) {
            parts[i] = Convert.unpackL(hb, i*4, EEEI);
          }
          if (numPoints*16>hb.length) hb = new byte[numPoints*16];
          sf.read (hb,0,numPoints*16);
          for(int i = 0; i < numPoints*2; i++) {
            points[i] = Convert.unpackD(hb, i*8, EEEI);
          }
          maxPartLength = numPoints;
          if (numParts>0) maxPartLength = numPoints-parts[numParts-1];
            for (int i=0; i<numParts-1; i++) {
            int nPoints = parts[i+1]-parts[i];
            if (nPoints>maxPartLength) maxPartLength=nPoints;
          }
          break;
        case MULTIPOINT:
          break;
        case POLYLINEM:
          break;
        case NULLSHAPE:
          /* FALLTHROUGH */
        default:
          int nullread = ((contentLength+4)*2) - 12;
          sf.read (hb,0,nullread);
      }

      if (sf.dbfOK) {

      }
      sf.ghb=hb;

      return getLength();
    }

    /** Sets the color of the shape.
        @param color The new color (null for default).
        @since NeXtMidas 2.1.0
     */
    public void setColor(Color color) { this.color = color; }

    /** Gets the color of the shape.
        @return The color (null for default).
        @since NeXtMidas 2.1.0
     */
    public Color getColor() { return this.color; }
  } // end class Shape
} // end class ShapeFile
