package nxm.sys.lib;

import java.util.LinkedHashMap;
import java.util.Map;

import nxm.sys.inc.*;

/** Stores a series of time vs index for general use by DataFiles. <br>
    <br>
    The nominal delta between values is typically specified as the x-delta in a {@link DataFile}.
    If the deltas differ more than the given tolerance ({@link #getTolerance()}), a new entry is
    placed in the time line. <br>
    <br>
    This class was initially designed to handle relatively smooth-flowing data that had occasional
    gaps (caused by brief interruptions in the data). This  class ignores minor deviations in the
    timecode (via tolerance) and adds entries to handle the gaps.<br>
    <br>
    Users frequently get into trouble when they use this class for other than its intended purpose
    (e.g. data where the delta is not constant) or, more commonly, use this class with a delta or
    tolerance that do not match the actual data.<br>
    <br>
    If the delta and tolerance do not match the actual data rate users may observe the following
    behavior:
    <pre>
      (1) The TimeLine overflows the max number of entries (see {@link #getAlloc()}
          causing the beginning of the file to have no timecode entries
          (in a PLOT this will often appear in the readout as "t=0").

      (2) The computed time values between TimeLine entries may be wrong.
          This is often seen when time updates are less frequent than the
          data (e.g. timecode is updated only with every Nth frame). This
          can cause several problems:

            (a) The extrapolated values after the last entry in the
                TimeLine are higher/lower than the actual time.

            (b) The extrapolated values in the middle of a TimeLine
                are not monotonically increasing and could exhibit
                points where the TimeLine "goes back in time."

            (c) In a PLOT the readout and features displayed on it
                may appear incorrect. This is a direct result of
                PLOT using extrapolated values that are just plain
                wrong. (Since the readout in a PLOT is based on mouse
                position, it rarely matches one-for-one with frame
                boundaries and ends up using extrapolated values most
                of the time.)

            (d) In the case of a pipe, all "down stream" primitives
                can end up with bad timecode since they are always
                using the last timeline entry or an extrapolation
                from it based on a bad delta.

      (3) If the number of TimeLine entries is allowed to grow (setAlloc(-1)
          in NextMidas 2.5.0 and later), the TimeLine may grow extremely
          large since each frame of data could necessitate a new entry in
          the TimeLine.
    </pre>
    In short, if you are going to use this class, you need to set delta and tolerance correctly.

    @author  Jeff Schoen
    @author  Neon Ngo - added support for high precision times
    @version $Id: TimeLine.java,v 1.27 2015/11/11 15:46:57 jtm Exp $
*/
public class TimeLine extends Object {
  /** The default time line length (TLL) to use. */
  public static final int    DEF_ALLOC     = 128;
  /** The default time line tolerance (TLT) to use. */
  public static final double DEF_TOLERANCE = 0.1; // Default tolerance
  /** The microsecond precision time line tolerance (TLT) to use. */
  static final double MICROSECOND_TOLERANCE = 1e-6;

  // Note: In cases where the timeline length is fixed (alloc!=-1) the
  //       entries will "wrap around" the array. 'current' always indicates
  //       the last entry written.

  private int      current;    // Index of last time line entry
  private int      alloc;      // User-set TimeLineLength (TLL), (-1 = Grow as needed)
  private int      size;       // Number of time line entries used
  private double[] offsets;    // Byte offsets into file for time entries (see times).
                               //   In bytes for compatibility with PIPES where output
                               //   pipe can use different formats.
  private Time[]   times;      // Time values for time line entries (see offsets)
  private double   dt;         // Time delta (sec/byte)
  private double   tolr;       // Time tolerance (sec)
  private boolean  interp;     // Use interpolation rather than extrapolation.

  /** timeline keyword input/output format option. @since NeXtMidas 3.3.0 */
  static enum Format {
    /** Use NeXtMidas 3.2.x and prior TIMELINE keyword which has microsecond precision (double values - seconds since J1950 epoch). */
    MICROSECOND_PRECISION,
//    /** (TODO: Not implemented) Use NeXtMidas 3.3.x+ high precision TIMELINE keyword (double values - seconds since HCB.timecode + TC_PREC Keyword epoch + ?XStart?). */
//    HIGH_PRECISION,
    /** Use Platinum BLUE Data Exchange Format Standard timeline keywords for very high precision. */
    PLATINUM_HIGH_PRECISION,
  };

  /** current time line input/output format to use (defaults to microsecond precision) */
  Format format = Format.MICROSECOND_PRECISION;

  /** Creates a new timeline with default values for length and tolerance.
      @param delta The delta between entries in sec/byte.
      @since NeXtMidas 2.5.0
   */
  public TimeLine (double delta) {
    this(DEF_ALLOC, delta, DEF_TOLERANCE);
  }

  /** Creates a new timeline with given TimeLine Length (TLL) and TimeLine Tolerance (TLT).
      @param tll   Max number of TimeLine entries (-1=Grow As Needed).
      @param delta The delta between entries in sec/byte.
      @param tlt   Allowable tolerance (smooths out small errors), in sec.
   */
  public TimeLine (int tll, double delta, double tlt) {
    current = -1;
    size = 0;
    setAlloc(tll);
    setDelta(delta);
    setTolerance(tlt);
  }

  /** Adds a time+offset entry to the end of the timeline. This is intended to
      be used for internal initialization similar to:
      <pre>
        TimeLine tl = new TimeLine(tll, delta, tlt);
        for (...) {
          addTimeOffset(...);
        }
      </pre>
      This function presumes that the tll used is sufficient and that no other
      entries have been added yet. This function is private since it is for
      internal use only and is very different from the normal <tt>setTimeAt(..)</tt>
      functions and is not intended for user interaction.
   */
  private void addTimeOffset (Time time, double offset) {
    current++;
    size++;
    times[current]   = time;
    offsets[current] = offset;
  }

  /** Should interpolation be used rather than extrapolation between TiimeLine entries?
      @since NeXtMidas 2.5.0
   */
  public boolean getUseInterpolation () { return interp; }

  /** Indicates if interpolation should be used rather than extrapolation between TiimeLine
      entries. Note that extrapolation is always used after the last entry in the TimeLine.
      This is should only be turned on when both of the following are true:
      <pre>
        o The time delta ({@link #getDelta()}) is likely to be inaccurate.
        o The data is continuous (no gaps).
      </pre>
      @param interp Turns interpolation on (true) or off (false).
      @since NeXtMidas 2.5.0
   */
  public void setUseInterpolation (boolean interp) { this.interp = interp; }

  /** get the time delta between offsets (in sec/byte) */
  public double getDelta () { return dt; }

  /** Set the time delta between offsets (in sec/byte) (aka rate of change for extrapolation).
      The TimeLine uses a single delta, changing this in the middle of the TimeLine will change
      the output of the {@link #getOffsetAt(double)} and {@link #getTimeAt(double)} methods.
      @param delta The time delta (rate) between bytes (in sec/byte).
      @throws IllegalArgumentException if delta is invalid: infinity, NaN, or zero (since 3.3.0)
   */
  public void setDelta (double delta) {
    if (Double.isInfinite(delta) || Double.isNaN(delta) || delta == 0)
      throw new IllegalArgumentException("Invalid TimeLine delta (sec/byte) = "+delta);
    dt = delta;
  }

  /** get the time tolerance (in seconds) */
  public double getTolerance () { return tolr; }

  /** set the time tolerance (in seconds) */
  public void setTolerance (double tolerance) { tolr = tolerance; }

  /** Return number of elements used in the timeline (see also {@link #getAlloc()}). */
  public int getSize () { return size; }

  /** Return max number of number of entries permitted in the timeline. Since NeXtMidas 2.5.0
      this will return <tt>-1</tt> if there is no fixed limit (i.e. the TimeLine will grow as
      needed). See {@link #getSize()} for the current number of entries used.
   */
  public int getAlloc () { return alloc; }

  /** Sets the max number of number of entries permitted in the timeline. Since NeXtMidas 2.5.0
      can be set to <tt>-1</tt> so that the TimeLine will grow as needed. Care should be taken
      when setting this to <tt>-1</tt> since the TimeLine could grow extremely large if the
      delta ({@link #getDelta()}) or tolerance ({@link #getTolerance()}) is wrong.
      @param newSize max number of entries (-1 = Grow As Needed).
      @since NeXtMidas 2.5.0
   */
  public void setAlloc (int newSize) {
    if (newSize == -1) { // Since 2.5.0 alloc=-1 to grow as needed
      alloc = newSize;
      realloc(Math.max(DEF_ALLOC, size));
    } else if (newSize < 1) {
      alloc = 1;
      realloc(1);
      Shell.getSharedMidasContext().deprecate("TimeLine: Since NeXtMidas 2.5.0: Use of alloc="+alloc
                                       +" is deprecated. Use alloc > 0; or alloc=-1 for auto.");
    } else {
      alloc = newSize;
      realloc(alloc);
    }
  }

  /** Reallocates the offsets and times arrays. This method is ONLY used internally, everyone
      else should call setAlloc(..).
   */
  private void realloc (int toAlloc) {
    double[] oldOffsets = offsets;
    Time[]   oldTimes   = times;

    if (oldOffsets == null) {
      // Initial setup -- fast, no copy needed
      offsets = new double[toAlloc];
      times   = new Time[toAlloc];
    } else if (oldOffsets.length == toAlloc) {
      // nothing to do
    } else {
      offsets = new double[toAlloc];
      times   = new Time[toAlloc];

      // Be careful to handle wrap-around correctly.
      boolean noWrap = (size < oldOffsets.length) || (current == oldOffsets.length);
      if (noWrap && (offsets.length > oldOffsets.length)) {
        // Be fast for the 90% case
        System.arraycopy(oldOffsets, 0, offsets, 0, oldOffsets.length);
        System.arraycopy(oldTimes,   0, times,   0, oldOffsets.length);
      } else {
        int start = 0;
        if (oldOffsets.length > offsets.length) {
          start = offsets.length - oldOffsets.length; // Decrease size
        }
        for (int i = 1, j = start; i <= oldOffsets.length; i++, j++) {
          int index = (current + i) % oldOffsets.length;

          if (j >= 0) {
            offsets[j] = oldOffsets[index];
            times[j]   = oldTimes[index];
          }
        }
      }
    }
    current = size-1;
  }

  /** Add new entry (up to microseconds precision) if it cannot be described within tolerance by the last entry.
      @param offset file offset (in bytes).
      @param time   time to set (number of J1950 seconds).
      @see #setTimeAt(double, Time) setTime(double offset, Time time) for a high precision version
   */
  public void setTimeAt (double offset, double time) {
    setTimeAt(offset, new Time(time));
  }

  /** Add new high precision entry if it cannot be described within tolerance by the last entry.
      @param offset file offset (in bytes).
      @param time   high precision time to set if NOT within tolerance of last entry;
                    (null and zero/negative entries are ignored).
      @since NeXtMidas 3.3.0
   */
  public void setTimeAt (double offset, Time time) {
    if (time == null || time.getSec() <= 0) return;
    int i = current;
    if (i<0) { }                        // first entry
    else if (offset==offsets[i]) i--;   // overwrite last entry
    else {                              // is this sufficiently described by the last entry
      double terr = time.diff(times[i]) - (offset - offsets[i])*dt;
      if (Math.abs(terr)<=tolr) return; // last entry is OK (within tolerance)
      if (terr < -1e7) return;          // back in time - probably invalid
    }
    i = i + 1;
    if (i == offsets.length) {
      if (alloc == -1) realloc(offsets.length*2); // resize
      else             i = 0;                     // just wrap around and overwrite
    }
    offsets[i]=offset;
    if (times[i] == null) times[i] = new Time(time); // copy into new instance
    else                  times[i].fromTime(time);   // set in existing instance
    size=Math.max(i+1,size);
    current=i;
  }

  /** Get time (high precision) entry for specified offset in file.
      @param offset file offset (in bytes).
      @param time   {@link Time} object to populate (null to create a new instance)
      @return Time entry for specified offset (zero Time when timeline is empty).
              This is the same as input param when it is not null.
      @since NeXtMidas 3.3.0
   */
  public Time getTimeAt (double offset, Time time) {
    if (time == null) time = new Time();
    int i = current; // Current index
    if (i < 0) {     // is the timeline empty?
      return time.fromJ1950x(0.0, 0.0); // matches previous implementation of returning 0.0;
    }

    double iOffset  = offsets[i]; // i-th offset
    if (offset >= iOffset) {      // is this equal to/later than the last time entry?
      time.fromTime(times[i]);
      return time.addSec( (offset-iOffset)*dt ); // Extrapolation from previous entry using dt
    }

    // go back and look for it
    for (int j=0; j<size; j++) {
      int k = i;       // Last index (for extrapolation)
      if (--i<0) i = size - 1;
      iOffset = offsets[i];
      if (offset >= iOffset) {
        time.fromTime(times[i]);
        if (interp && (j != 0)) {
          // Linear interpolation between two points
          double add = (times[k].diff(time) * (offset-iOffset)) / (offsets[k]-iOffset);
          return time.addSec(add);
        } else {
          // Extrapolation from previous point using dt
          return time.addSec( (offset-iOffset)*dt );
        }
      }
    }
    return time.fromJ1950x(0.0, 0.0); // not found (matches previous implementation of returning 0.0;)
  }

  /** get the time (seconds since J1950) from the current entries
      @param offset file offset (in bytes).
      @see #getTimeAt(double,Time) getTime(double offset, Time time) for a high precision version
   */
  public double getTimeAt (double offset) {
    return getTimeAt(offset, new Time()).getSec();
  }

  /** get the file offset (in bytes) for a given time from the current entries.
      @param time time to set (number of J1950 seconds).
      @return file offset for specified time (in bytes).
      @see #getOffsetAt(Time) getOffset(Time) for a high precision version
   */
  public double getOffsetAt (double time) {
    return getOffsetAt(new Time(time));
  }

  /** get the file offset (in bytes) for a given time from the current entries.
      @param time high precision time to find offset (cannot be null).
      @return file offset for specified time (in bytes), 0 if timeline is empty.
      @since NeXtMidas 3.3.0 (extracted originally from getOffsetAt(double))
   */
  public double getOffsetAt (Time time) {
    int i = current; // Current index

    // is the timeline empty?
    if (i<0) return 0.0;

    // is this later than the last time entry ?
    double tdiff = time.diff(times[i]);
    if (tdiff >= 0) return Math.floor(offsets[i] + tdiff/dt);

    // go back and look for it
    for (int j=0; j<size; j++) {
      int k = i;       // Last index (for extrapolation)
      if (--i<0) i=size-1;
      tdiff = time.diff(times[i]);
      // protect from possibly invalid times[i] entry
      if (tdiff>=0 && tdiff<1e7) {
        if (interp && (k != current)) {
          // Linear interpolation between two points
          return Math.floor((((offsets[k]-offsets[i]) * tdiff) / (times[k].diff(times[i]))) + offsets[i]);
        } else {
          // Extrapolation from previous point using dt
          return Math.floor(offsets[i] + tdiff/dt);
        }
      }
    }
    return 0.0;
  }

  /** Used to purge entries from the TimeLine. */
  public void purge (double toOffset) {
    if (toOffset<=0) return;
    int i = current;
    for (; i>0 && offsets[i]>toOffset; i--) { }
    if (i==0) return;
    size = current+1-i;
    for (int j=0; j<size; j++) {
      offsets[j] = offsets[j+i];
      times[j] = times[j+i];
    }
  }

  /** <b>INTERNAL USE ONLY:</b> Get as Data object for keywords. This is for internal use between
      {@link DataFile} and {@link Keywords}, it is <tt>public</tt> so that it can be used by custom
      file handlers in other option trees. The {@link Data} object holds <tt>"SD"</tt> data in
      the following form:
      <pre>
        data[0]     = {@link #getDelta()}
        data[1]     = {@link #getTolerance()}
        data[2*i+2] = offsets[i]
        data[2*i+3] = times[i]

        Note: The times and offsets do not start at i=0. When the TimeLine gets full
              and wraps around the offsets and times begin overwriting at i=0. For
              example:
                                                v
                  times = { 12, 13, 14, 15, 16, 5, 6, 7, 8, 9, 10, 11 }
                                                ^
        Note2: The times are seconds since J1950 epoch (giving microsecond precision)
      </pre>
      This is the same form that is stored in the extended header keywords of a NeXtMidas BLUE file.
   */
  @InternalUseOnly
  public Data toData () {
/*
        Note2: When format == {@link Format#MICROSECOND_PRECISION} then the times are seconds since J1950 epoch.
BEGIN TODO: NOT IMPLEMENTED -->
               Otherwise (format == {@link Format#HIGH_PRECISION}) and the times are seconds since file's start time (timecode + TC_PREC keyword + xstart?) as epoch,
               which allows high precision times: e.g.
                 2^(-40) (~9.09E-13) gives picosecond precision
                   2^(53-40) = 2^13 = 8192 max seconds
                   8192 seconds / 60 seconds/minute ~ 136 minutes ~ 2 hours
                 2^(-30) (~9.31E-10) gives nanosecond precision
                   2^(53-30) = 2^23 = 8388608 max seconds
                   8388608 seconds / 86400 seconds/day ~ 97 days
(picosecond precision for 75 minute duration from epoch ??? 2^13 (8192) seconds / 60 min/sec ~ 69 minute duration <-- TODO double check
-or- nanosecond precision for ~52 day duration from epoch) ->> vs 2^22 / 86400 sec/day ~ 48 day duration <-- TODO double check
END TODO: NOT IMPLEMENTED

      NOTE: This is the non-high precision version (up to microsecond precision)
            for backwards compatibility with NeXtMidas versions prior to 3.3.0.
 */
    int j=0;
    Data data = new Data("SD",size*2+2);
    data.setD( j++, dt );
    data.setD( j++, tolr );
    for (int i=0; i<size; i++) {
      data.setD( j++, offsets[i] );
      data.setD( j++, times[i].getSec() );
    }
    return data;
  }

  /** <b>Internal use only:</b> Convert this TimeLine to Map that is suitable to store as Keywords
      in BLUE files with high precision time entries (up to picosecond precision).
      This follows the Platinum BLUE Data File Exchange Format Standard rev 4.
      <br>
      Typical Platinum use is with timeEpoch=1994-01-01 and timeDelta=1e-10 for clock ticks.
      @param timeEpoch time epoch to use.
      @param timeDelta time delta (in seconds) of each clock tick (resolution) use time units of {@link Units#COUNTS}.
                       -or- 0.0 for seconds from epoch to use time units of {@link Units#TIME_S}.
      @param dbpe  bytes per element (to convert file offset in bytes to offset in elements).
      @throws IllegalStateException when this timeline does not have any entries
      @return Map of keywords to store the high precision timeline in the extended header of a BLUE file
      @since NeXtMidas 3.3.0
      @see #fromPlatinumKeywords(Keywords, double, double)
   */
  @InternalUseOnly
  Map<String,Object> toPlatinumKeywords (Time timeEpoch, double timeDelta, double dbpe) throws IllegalStateException { // package-private
    if (current < 0) throw new IllegalStateException("TimeLine is empty");

    Map<String,Object> map = new LinkedHashMap<String,Object>();
    final int units = (timeDelta == 0)? Units.TIME_S : Units.COUNTS;
    Data offsetData = new Data("SX", size);
    Data timeData   = new Data(((units == Units.COUNTS)? "SX" : "SD"), size);
    for (int ii=0; ii<current; ii++) {
      long elemOffset = (long) (offsets[ii] * dbpe); // convert to offset in elements (from bytes)
      offsetData.setX(ii, elemOffset);
      if      (units == Units.COUNTS) timeData.setNumber(ii, times[ii].toTimeTicks(timeEpoch, timeDelta));
      else if (units == Units.TIME_S) timeData.setNumber(ii, times[ii].diff(timeEpoch));
    }

    map.put("TIMETAG.OFFSET",     offsetData);              // array of offsets type X (long)
    map.put("TIMETAG.TIME",       timeData);                // array of times type X or D (long or double)
    map.put("TIMETAG.TIME.UNITS", units);                   // type L (integer)
    map.put("TIME_EPOCH",         timeEpoch.toString(Time.ISO8601, 0)); // string in ISO8601 format
    if (units == Units.COUNTS) {
      map.put("TIME_DELTA", timeDelta);                     // type D (double)
      map.put("TIME_TICKS", 1L/timeDelta);                  // type X (long) (inverse of delta)
    }
    return null;
  }

  /** Try to create high precision timeline from a {@link Keywords} object
      that follows the Platinum BLUE Data File Exchange Format Standard rev 4.
      @param keywords The Keywords from DataFile with Platinum time-line keywords.
      @param dfDelta  The DataFile's time delta (XDELTA/YDELTA) per element.
      @param dbpe  bytes per element (to convert file offset in bytes to offset in elements).
      @return new TimeLine instance from Platinum time-line keywords if successful,
              otherwise null if unable to extract Platinum time-line from DataFile/Keywords.
      @see #toPlatinumKeywords(Time, double, double)
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  static TimeLine fromPlatinumKeywords (Keywords keywords, double dfDelta, double dbpe) { // package-private
    // KEYWORD VALUES
    Object  _timeData   = keywords.get("TIMETAG.TIME");
    Object  _offsetData = keywords.get("TIMETAG.OFFSET");
    Object  _timeEpoch  = keywords.get("TIME_EPOCH");
    Double  timeDelta   = keywords.getD("TIME_DELTA",          null); // optional (double)
    Integer unitsKW     = keywords.getUnitsFor("TIMETAG.TIME", null); // optional (integer)

    if ((_timeEpoch  == null) || !(_timeEpoch  instanceof String)) return null; // missing/invalid TIME_EPOCH
    if ((_timeData   == null) || !(_timeData   instanceof Data  )) return null; // missing/invalid TIMETAG.TIME
    if ((_offsetData == null) || !(_offsetData instanceof Data  )) return null; // missing/invalid TIMETAG.OFFSET

    // DERIVED VALUES
    Time   timeEpoch; // set below
    int    units      = (timeDelta == null)? Units.TIME_S : Units.COUNTS;
    double tolerance  = (timeDelta == null)? dfDelta      : timeDelta;
    Data   offsetData = (Data)_offsetData; // array of offsets (usually of type X (long))
    Data   timeData   = (Data)_timeData;   // array of times   (usually of type X (long) or D (double))

    // ERROR CHECKING
    try {
      timeEpoch = Time.toTime(_timeEpoch, Time.ISO8601);
    }
    catch (Exception ex) {
      Shell.printStackTrace("Invalid value in TIME_EPOCH keyword ["+_timeEpoch+"]!", ex);
      return null;
    }
    if (unitsKW != null && unitsKW != units) {
      Shell.warning("Mismatch time units: TIMETAG.TIME.UNIT(S)="+unitsKW+", using time units="+units
                  + " derived from TIME_DELTA keyword");
    }
    if (offsetData.size != timeData.size) {
      Shell.warning("TIMETAG.OFFSET["+offsetData.size+"] and TIMETAG.TIME["+timeData.size
                  + "] MUST have same number of entries (size)");
      return null;
    }
//    else if (offsetData.spa != timeData.spa) {
//      Shell.warning("TIMETAG.OFFSET["+offsetData.spa+"] and TIMETAG.TIME["+timeData.spa+"] MUST have same number of entries (spa)");
//      return null;
//    }


    // CREATE & INIT TIMELINE
    TimeLine timeline = new TimeLine(timeData.getSize(), dfDelta / dbpe, tolerance);

    for (int ii=0; ii < timeData.getSize(); ii++) {
      Number td     = timeData.getNumber(ii);
      Time   time   = (units == Units.COUNTS)? new Time(td.longValue(), timeEpoch, timeDelta)
                                             : new Time(timeEpoch).addSec(td.doubleValue());
      double offset = offsetData.getNumber(ii).doubleValue() / dbpe;

      timeline.addTimeOffset(time, offset);
    }
    timeline.format = Format.PLATINUM_HIGH_PRECISION;
    return timeline;
  }

  /** <b>INTERNAL USE ONLY:</b> Load from a Keywords Data object. This is for internal use between
      {@link DataFile} and {@link Keywords}, it is <tt>public</tt> so that it can be used by custom
      file handlers in other option trees. See {@link #toData()} for details on the structure
      of the {@link Data} object.
      WARNING: MUST call {@link #setAlloc(int)} with enough space to store time entries from input
               data before calling this (and alloc SHOULD NOT be set to -1) otherwise the time
               entries will be ignored or truncated to size of alloc.
   */
  @InternalUseOnly
  public void fromData (Data data) {
    int j=0;
    size    = Math.min((data.size*data.spa-2)/2, alloc);
    dt      = data.getD(j++);
    tolr    = data.getD(j++);
    current = -1;
    double tmax = 0;
    double time;
    for (int i=0; i<size; i++) {
      offsets[i] = data.getD(j++);
      time       = data.getD(j++);
      times[i]   = new Time(time);
      if (time>tmax) { tmax=time; current=i; }
    }
    format = Format.MICROSECOND_PRECISION;
  }

  /** <b>INTERNAL USE ONLY:</b> list entire contents of TimeLine (used by STATUS/T) */
  @InternalUseOnly
  public String list () {
    return list(3);
  }

  /** <b>INTERNAL USE ONLY:</b> list entire contents of TimeLine (used by STATUS/TIME)
      @param decimalPlaces number of decimal places to show for time.
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  public String list (int decimalPlaces) {
    StringBuilder s = new StringBuilder(80 + size*(56+decimalPlaces)); // header (usually <80) + entries (usually <56 chars each)
    s.append("TimeLine Size=").append(size);
    s.append(" Alloc=").append(alloc);
    s.append(" Delta=").append(dt).append(" (sec/byte)");
    s.append(" Tolr=").append(tolr).append(" (sec)\n");
    for (int i=0; i<size; i++) {
      s.append("Index=").append(i).append(" Off=").append(offsets[i]);
      s.append(" Time=").append(times[i].toString(Time.STD,decimalPlaces));
      if (i == 0 && (offsets[0] != 0.0)) {
        s.append(" <-- Does not start at beginning of file!");
      }
      s.append("\n");
    }
    if ((size == offsets.length) && (offsets[0] != 0.0)) {
      // JC4: 15MAY07 Added this warning since it is a very common mistake.
      s.append("WARN: TimeLine is full, some entries may have been discarded. It is likely that\n");
      s.append("WARN: the Delta and Tolerance values do not match the observed data rate.\n");
    }
    return s.toString();
  }

  @Override
  public String toString () {
    int decimalPlaces = 3;
    // String s = "TimeLine Size="+size+" Of="+alloc;
    StringBuilder sb = new StringBuilder("TimeLine Size="+size+" Alloc="+alloc+" Delta="+dt+"(sec/byte) Tolr="+tolr+" (sec)");
    if (size==0) return sb.toString();
    Time tmin=times[0], tmax=tmin;
    for (int i=1; i<size; i++) {
      if (times[i].compareTo(tmax) > 0) tmax=times[i];
      if (times[i].compareTo(tmin) < 0) tmin=times[i];
    }
    sb.append("\n From=").append(tmin.toString(Time.STD,decimalPlaces));
    sb.append("\n To  =").append(tmax.toString(Time.STD,decimalPlaces));
    return sb.toString();
  }

  /** <b>INTERNAL USE ONLY:</b> access to read the array of Time objects in the Timeline.
      NOTE: this method is not use anywhere, so we could remove it or keep it as package-private.
      @return {@link #times}, the array of times in the timeline.
      @since NeXtMidas 3.3.0 changed return type and made package-private
   */
  @InternalUseOnly("since initial version")
  Time[] getTimes () { return times; }

}
