package nxm.sys.libg;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import nxm.sys.inc.Drawable;
import nxm.sys.inc.ProvisionalUseOnly;
import nxm.sys.lib.Position;
import nxm.sys.lib.Table;

import static java.lang.Math.max;
import static java.lang.Math.min;

/** A class to group together Drawable objects. <br>
 *  <br>
 *  The purpose of this class is to allow users to treat multiple Layers and/or Features as a single
 *  Layer, so that performing functions such as enable/disable on them as a group is simpler.
 *
 *  @author J. Bauer, J. Czechowski IV
 *  @since NeXtMdias 3.5.0
 */
@ProvisionalUseOnly("Added in NeXtMidas 3.5.0")
public final class LayerDrawable extends Layer {
  // The following variables hold the set of all drawables in the order in which they are to be
  // drawn. The second one (unmodifiableDrawables) maps to the first and provides the unmodifiable
  // view exposed to users. The default size of 'drawables' is 64 as that seems like a good default
  // that ballances memory use for small sets against number of reallocations for larger sets.
  // The CopyOnWriteArrayList is used rather than ArrayList to avoid ConcurrentModificationException
  // in the draw method caused by users modifying the underlying array. The CopyOnWriteArrayList is
  // slower than ArrayList during add/remove, but the number of those is far below the number of
  // draw calls expected, which would otherwise need to be synchronized (which is slower).
  private final List<Drawable> drawables             = new CopyOnWriteArrayList<Drawable>();
  private final List<Drawable> unmodifiableDrawables = Collections.unmodifiableList(drawables);

  /** Cycles through the child Layers inside this layerDrawable, and calls getData() in each of them.
   *  @return Typically returns Drawable.FULL, unless there are only Layers (no Features) in this Group,
   *          and each of those Layers does not return drawable.FULL.
   */
  @Override
  public int getData (int mode) {
    if (drawables.isEmpty()) return Drawable.FULL;
    int stat = Drawable.FAST;
    for (Drawable d : drawables) {
      if (d instanceof Feature)  {
        stat = Drawable.FULL;  // return FULL so Plot.process() doesn't get in a loop
      }
      if (d instanceof Layer) {
        int tmpStat = ((Layer)d).getData(mode);
        if (tmpStat >= Drawable.FULL && tmpStat < stat) stat = tmpStat;  // set to most drawing needed
      }
    }
    return stat;
  }

  /** @since NeXtMidas 3.5.4 */
  @Override
  public void findRange () {
    boolean first = true;
    for (Drawable d : drawables) {
      if (d.isEnabled(0)) {
        d.findRange();
        if (first) setRange(d);
        else       addRange(d);
        first = false;
      }
    }
    if (first) {
      // When empty (as is often the case before features are added), or when all features are
      // not enabled (i.e. all are "-Global"), just mimic the range of the base layer since we
      // don't want to use the default setting (x1=0,x2=1,y1=0,y2=1) and mess up the auto-scaling
      // in the PLOT which is otherwise based on the other layers.
      MP.getBaseLayer().findRange();
      setRange(MP.getBaseLayer());
    }
    // In situations where the layer consists only of features it is possible to get x1=x2 and/or
    // x1=y2 which can cause scaling issues within PLOT. As such, if we detect an x1=x2 or y1=y2
    // situation we adjust the values by shifing them 1% of the width/height of the base layer.
    // We use the base layer as a reference rather than some arbitrary "small delta" since it
    // maintains the order-of-magnitude of the PLOT.
    if (x1 == x2) {
      double dx = Math.abs(MP.getBaseLayer().getX2() - MP.getBaseLayer().getX1());
      if (dx == 0) dx = 1; // rare (bad) situation where layer has zero size
      x1 -= dx/100;
      x2 += dx/100;
    }
    if (y1 == y2) {
      double dy = Math.abs(MP.getBaseLayer().getY2() - MP.getBaseLayer().getY1());
      if (dy == 0) dy = 1; // rare (bad) situation where layer has zero size
      y1 -= dy/100;
      y2 += dy/100;
    }
    // Set Minimum and Maximum actual auto-scale value (since 3.7.0 for NXM-2984)
    a1 = y1;
    a2 = y2;
  }

  /** Sets the range based on the given drawable. */
  private void setRange (Drawable d) {
    x1 = d.getXMin();  x2 = d.getXMax();
    y1 = d.getYMin();  y2 = d.getYMax();
    z1 = d.getZMin();  z2 = d.getZMax();
  }

  /** Update the range to include the given drawable. */
  private void addRange (Drawable d) {
    x1 = min(x1,d.getXMin());  x2 = max(x2,d.getXMax());
    y1 = min(y1,d.getYMin());  y2 = max(y2,d.getYMax());
    z1 = min(z1,d.getZMin());  z2 = max(z2,d.getZMax());
  }

  /** @since NeXtMidas 3.5.4 */
  @Override
  public void draw (int flag) {
    clearNeedRefresh();                         // <-- clear our local refresh flag
    for (Drawable d : drawables) {
      // use MP.dowDraw(Drawable, int, int) with new DO_FORCE_DRAW flag Since NeXtMidas 3.5.4 Bug 2913
      MP.doDraw(d, flag, MPlot.DO_FORCE_DRAW|MPlot.DO_DRAW);  // <-- calls clearNeedRefresh() on the item
    }
  }

  @Override
  public boolean getNeedRefresh () {
    if (super.getNeedRefresh()) return true;
    for (Drawable d : drawables) {
      if (d.getNeedRefresh()) return true;
    }
    return false;
  }

  @Override
  public boolean getNeedsRescale () {
    if (super.getNeedsRescale()) return true;
    for (Drawable d : drawables) {
      if ((d instanceof Layer) && (((Layer)d).getNeedsRescale())) return true;
    }
    return false;
  }


  /** Add a Drawable object to the array of Drawable objects in this Layer.
   *  @param d The Drawable to add
   */
  public void addDrawable (Drawable d) {
    drawables.add(d);
    needRefresh = true;
  }

  /** Gets a Drawable with the given name.
   *  @param name The name of the drawable.
   *  @return The drawable object or null if not found.
   *  @since NeXtMidas 3.5.4
   */
  public Drawable getDrawable (String name) {
    for (Drawable d : drawables) {
      if (name.equals(d.getName())) return d;
    }
    return null;
  }

  /** Removes a drawable from this layer.
   *  @param name The name of the drawable(s) to remove or "*" to remove all.
   *  @return TRUE if the Drawable is removed, else FALSE.
   */
  public boolean removeDrawable (String name) {
    return (name.equals("*"))? clearAllDrawables() : removeDrawable(getDrawable(name));
  }

  /** Removes all drawables from this layer.
   *  @return TRUE if the Drawable is removed, else FALSE.
   *  @since NeXtMidas 3.5.4
   */
  public boolean clearAllDrawables () {
    if (drawables.isEmpty()) return false;
    drawables.clear();
    needRefresh = true;
    return true;
  }

  /** Remove a Drawable object from the array of Drawable objects in this Layer.
   *  @param d The drawable to remove.
   *  @return TRUE if the Drawable is removed, else FALSE.
   */
  public boolean removeDrawable (Drawable d) {
    if (d == null) return false;
    boolean anyRemoved = drawables.remove(d);
    needRefresh = anyRemoved;
    return anyRemoved;
  }

  /** Remove a Drawable object from the array of Drawable objects in this Layer.
   *  @param index The index number of the drawable.
   *  @return TRUE if the Drawable is removed, else FALSE
   *  @throws IndexOutOfBoundsException if the index is out of range.
   */
  public boolean removeDrawable (int index) {
    drawables.remove(index);
    needRefresh = true;
    return true;
  }

  /** Return a list of the Drawable Objects in the LayerDrawable in Table form.
   *  @param type     "FEATures", "LAYers", or "ALL" [DEF="ALL"]
   *  @param nameOnly if true, the Table only contains a list of the names of the Objects,
   *                  not the Objects themselves
   *  @return A Table containing the Drawables themselves or a list of their names, or null
   *          if there are no Drawables.
   */
  public Table getDrawables (String type, boolean nameOnly) {
    // TODO: Is this useful? Since removeDrawable(..) works off the drawable's name, why not use
    //       that as the key? Suggest revising in NeXtMidas 3.7.x series
    int itype; // 0==all (int for performance)
    String dtype = type.toUpperCase();
    if      (dtype.startsWith("FEAT")) itype = 0x1;  // features
    else if (dtype.startsWith("LAY"))  itype = 0x2;  // Layers
    else if (dtype.equals("ALL"))      itype = 0x3;  // ALL [default]
    else throw new IllegalArgumentException("Invalid type '"+type+"', expected FEATURES, LAYERS, or ALL.");

    Table t = new Table();
    int featIndex = 0;
    int layIndex  = 0;
    for (Drawable d : drawables) {
      String key = null;

      if (((itype & 0x1) != 0) && (d instanceof Feature)) key = "FEAT_"+featIndex++;
      if (((itype & 0x2) != 0) && (d instanceof Layer  )) key = "LAY_"+layIndex++;

      if (key!=null) {
        if (nameOnly)  t.put(key, d.getName());
        else           t.put(key, d);
      }
    }
    return t;
  }

  /** Returns a COPY of the array containing the drawables objects.
   *  @return A COPY of the array containing the drawables objects.
   *  @deprecated Since NeXtMidas 3.5.4 use getAllDrawables()
   */
  @Deprecated // Since NeXtMidas 3.5.4 use getAllDrawables()
  public Drawable[] getDrawables () {
    return drawables.toArray(new Drawable[drawables.size()]);
  }

  /** Returns an unmodifiable view of the drawables objects held by this layer.
   *  @return An unmodifiable view of the drawables objects held by this layer (not null).
   *  @since NeXtMidas 3.5.4
   */
  public List<Drawable> getAllDrawables () {
    return unmodifiableDrawables;
  }

  /** Gets the drawable instances that are visible at the specified position. <br>
   *  <br>
   *  This uses a very rough check on the bounds of each drawable object and adds a fudge-factor
   *  to get small drawable features even is the user's cursor isn't exactly on it. As a result
   *  this may return features that are not strict matches. <br>
   *  <br>
   *  This check omits any drawable items that are not enabled, but may select ones that are
   *  only nominally visible (e.g. "black on black").
   *  @param pos The position to check for.
   *  @return The set of drawables at the given position in the order they appear (not null).
   *  @since NeXtMidas 3.5.4
   */
  public List<Drawable> getDrawablesAt (Position pos) {
    List<Drawable> matches = new ArrayList<Drawable>(8); // 8=Good guess for max number of overlapping

    for (Drawable d : drawables) {
      if (!MP.doDraw(d, 0, MPlot.DO_CHECK)) continue;
      if (d.isAtPosition(pos, -1, -1)) matches.add(d);
    }
    return matches;
  }
}