package nxm.sys.libg;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.util.Arrays;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.prim.plot;

/**
 * A class to hold and plot type 5000 location data.
 * @author Beth Hassler / Jeff Schoen / Jay Hallam / J. Czechowski IV
 * @version $Id: LayerT5.java,v 1.93 2015/12/21 15:39:17 ntn Exp $
 */
public class LayerT5 extends Layer {
  /** Factor to multiply by to convert meters to deg lat. */
  private static final double METERS_TO_DEG_LAT = 180.0 / (Constants.POLAR_RADIUS_IN_METERS * 2.0 * Math.PI);

  /** Factor to multiply by to convert meters to deg lon. */
  private static final double METERS_TO_DEG_LON = 360.0 / (Constants.EARTH_EQUATORIAL_RADIUS * 2.0 * Math.PI);

  /** The list of selection modes. (Value: <code>{@value}</code>) */
  public static final String selectionModes = "Pixel,GreatArc";

  /** Field enable list. (Value: <code>{@value}</code>) */
  public static final String fieldList = "Symbols,Labels,Elps,Cones,Connect";

  /** The list of available flags. (Value: <code>{@value}</code>) */
  public static final String flagsList = "ConeFlagHoriz,ConeFlagVert,TimeScaleLabels,AutoScaleLabels,FillCones";

  /** Selection mode: Nearest using pixel coord. */
  public static final int NEAREST_BY_PIXELS = 1;

  /** Selection mode: Nearest using great arc. */
  public static final int NEAREST_BY_GREAT_ARC = 2;

  /** Field option: Show symbols. */
  public static final int SYMBOLS = 0x01;

  /** Field option: Show labels. */
  public static final int LABELS = 0x02;

  /** Field option: Show ellipses. */
  public static final int ELLIPSES = 0x04;

  /** Field option: Show cones. */
  public static final int CONES = 0x08;

  /** Field option: Connect points. */
  public static final int CONNECT = 0x10;

  /** Flag: Cones should extend to the horizon. */
  public static final int CONE_FLAG_HORIZ = 0x1;

  /**
   * Flag: The flag value indicating that cones should be clipped at a plane vertical to their
   * intersection with the Earth's surface. If both {@link #CONE_FLAG_HORIZ} and
   * {@link #CONE_FLAG_VERT} are set then the vertical is shown as a dotted line.
   */
  public static final int CONE_FLAG_VERT = 0x2;

  /** Flag: Labels should be shown for time-scale plots. */
  public static final int SHOW_TIME_SCALE_LABELS = 0x4;

  /** Flag: Labels should be auto-scaled (hides labels until the user zooms in to avoid cluttering up the plot). */
  public static final int AUTO_SCALE_LABELS = 0x8;

  /** Flag: Cones and ellipses should be drawn filled rather than outlines. */
  public static final int FILL_CONES = 0x10;

  /** Matches ',' or "|" plus any whitespace before/after. */
  private static final String SPLIT_REGEX = "\\s*[,|]\\s*";

  /** Array of points to be displayed. */
  private Pnt[] points = null;

  /** Number of points in the points array. */
  private int numPoints = 0;

  /** Number of points in the points array that are null. */
  private int nullPoints = 0;

  /** Start index of the points array (-1=Variable size). */
  private int firstPoint = -1;

  /** The selection mode used by AutoSelect. */
  private int autoSelMode = NEAREST_BY_GREAT_ARC;

  /** The selection range used by AutoSelect. */
  private double autoSelRange = 500000.0;

  /** The AutoSelect enable flags. */
  private int autoSelEnable = 0x00;

  /** The enables fields, defaults to all fields on.
   * @deprecated Since NeXtMidas 3.7.0 - use {@link #getField()} and {@link #setField(String)} instead
   */
  @Deprecated
  public int field = -1;

  /** Index of selected point (-1 for none selected).
   * @deprecated Since NeXtMidas 3.7.0 - use {@link #getSelectedIndex()} and {@link #setSelectedIndex(int)} instead
   */
  @Deprecated
  public int select = -1;

  /**
   * Toggle Label Region (show text if {@code viewable < tlreg }).
   * @deprecated Since NeXtMidas 3.7.0 - use {@link #getTLReg()} and {@link #setTLReg(double)} instead
   */
  @Deprecated
  public double tlreg = 10.0;

  /** The keys to include when using getSelectedText(). */
  private String[] selectedKeys = null;

  /** The component byte offset of the name field. */
  private int bName = -1;

  /** The component byte offset of the position field. */
  private int bPos = -1;

  /** The component byte offset of the position pointer field. */
  private int bPosp = -1;

  /** The component byte offset of the cone field. */
  private int bCone = -1;

  /** The component byte offset of the ellipse field. */
  private int bElps = -1;

  /** The component byte offset of the flags field. */
  private int bFlag = -1;

  /** The component byte offset of the connect flag field. */
  private int bConn = -1;

  /** The component byte offset of the note field. */
  private int bNote = -1;

  /** The component byte offset of the heading field. */
  private int bHdg = -1;

  /** The component byte offset of the vector field. */
  private int bVec = -1;

  /** The component dimensions of the vector field. */
  private int dVec = -1;

  /** The component format for the name field. */
  private String fName = null;

  /** The component format for the note field. */
  private String fNote = null;

  /** The component format for the cone field. */
  private String fCone = null;

  /** Transform to use (based on DataFile quadwords). */
  private Transform transform = null;

  /** Using a time axis? */
  private boolean timeAxis = false;

  /** Font size (0=Auto). */
  private int fontsize = 0;

  /** Font to use. */
  private Font font = null;

  /** Pix to use for {@code RWC -> pixel conversions}. */
  private Pix pix = null;

  /** The data file being plotted */
  private DataFile df = null;

  /** Number of conic sections. */
  private int conic = 1;

  /** Conic divisions. */
  private int cdiv = 60;

  /** Conic angle. */
  private double cangle = 1;

  /** Conic range. */
  private double crange = 0;

  /** Labels should be on the left? */
  private boolean leftLabels = false;

  /** Color of the selected point. */
  private Color selectColor = Color.RED;

  /** Any flags set. */
  private int flags = 0x09;

  /** Indicates if initialization has been completed. */
  private boolean initialized = false;

  /** Indicates if always convert to LonLatAlt. */
  private boolean toLLA = false;

  /** Only used inside draw methods: Zoomed in enough to show labels. */
  private boolean _labelZoom = false;

  /** Only used inside draw methods: show symbols. */
  private boolean _symbols = false;

  /** Only used inside draw methods: show labels. */
  private boolean _labels = false;

  /** Only used inside draw methods: connect points. */
  private boolean _connect = false;

  /** Only used inside draw methods: connect all points. */
  private boolean _connectAll = false;

  /** Only used inside draw methods: is it selected. */
  private boolean _selected = false;

  /** Only used inside draw methods: color to use. */
  private Color _color = null;

  /** Only used inside draw methods: show all points. */
  private boolean _showAll = false;

  /** Have minLon/maxLon/... been set yet? */
  private boolean minMaxSet = false;

  /** Minimum longitude. */
  private double minLon = 0.0;

  /** Maximum longitude. */
  private double maxLon = 0.0;

  /** Minimum latitude. */
  private double minLat = 0.0;

  /** Maximum latitude. */
  private double maxLat = 0.0;

  /** Minimum altitude. */
  private double minAlt = 0.0;

  /** Maximum altitude. */
  private double maxAlt = 0.0;

  /** Vector scale factor. */
  private double vscale = 0.0;

  /** whether or not to optimize drawing (warning: prevents some lines from displaying). */
  private boolean quickDraw = false;

  private Pnt timePoint = null;

  /** Default constructor. */
  public LayerT5() {
    super();
    name = "T5";
    enable &= ~MPlot.EN_BSTORE;
    line.setRadius(3);
    line.setSymbol("CROSS");
    line.getSymbol().setSize(4);
    line.setType(Line.SOLID | Line.SYMBOL);
    needRefresh = true; // JC4: Prevents race condition.
    pix = new Pix(1, 0);
  }

  /**
   * Constructor with default size.
   * @param size The default size.
   */
  public LayerT5(int size) {
    this();
    setSize(size);
  }

  /**
   * Set the fields to be enabled.
   * @param value This is a mask value of the field elements.
   */
  public void setField(int value) {
    field = value;
    refresh();
  }

  /**
   * Set the fields to be enabled.
   * @param value This is a mask value of the field elements in {@link #fieldList}.
   */
  public void setField(String value) {
    setField(Parser.mask(fieldList, value, field));
  }

  /**
   * Get the fields to be enabled.
   * @return The fields enabled (see {@link #fieldList}).
   */
  public String getField() {
    return Parser.mask2s(fieldList, field);
  }

  /**
   * Set the file input.
   * @param pf The plot file to read.
   * @param tab <i>Not Used</i>
   * @param MA <i>Not Used</i>
   * @return <code>true</code> when the file has been opened, <code>false</code> if the file can not
   *         be opened (usually indicates the file is not a Type 5000 file).
   */
  @Override
  public synchronized boolean setFile(PlotFile pf, Table tab, Args MA) {
    setFile(pf);
    if (typeClass != 5)
      return false;

    if (pf instanceof DataFile) {
      df = (DataFile) pf;
      DataFile df = (DataFile) pf;

      int iPos  = df.findRec("POS");
      int iPosp = df.findRec("POSP");
      int iCone = df.findRec("CONE");
      int iName = df.findRec("NAME");
      int iNote = df.findRec("NOTE");
      int iFlag = df.findRec("FLAG");
      int iElps = df.findRec("ELPS");
      int iConn = df.findRec("CONN");
      int iHdg  = df.findRec("HDG");

      bPos = (iPos >= 0) ? df.getCompOffset(iPos) : -1;
      bPosp = (iPosp >= 0) ? df.getCompOffset(iPosp) : -1;
      bCone = (iCone >= 0) ? df.getCompOffset(iCone) : -1;
      bName = (iName >= 0) ? df.getCompOffset(iName) : -1;
      bNote = (iNote >= 0) ? df.getCompOffset(iNote) : -1;
      bFlag = (iFlag >= 0) ? df.getCompOffset(iFlag) : -1;
      bElps = (iElps >= 0) ? df.getCompOffset(iElps) : -1;
      bConn = (iConn >= 0) ? df.getCompOffset(iConn) : -1;
      bHdg = (iHdg >= 0) ? df.getCompOffset(iHdg) : -1;

      fCone = df.getCompFormat(iCone);
      fName = df.getCompFormat(iName);
      fNote = df.getCompFormat(iNote);
      if (df.isPipe())
        firstPoint = 0; // Set to use a fixed number of points
    }
    setSize((int) pf.getSize());
    setData(pf.getDataBuffer(size));
    transform = new Transform(df); //TODO df == null case
    timeAxis = (pf.getXUnits() == Units.TIME);
    toLLA = !df.getReferenceFrame().equals("NONE"); //TODO df == null case
    quickDraw = MA.getState("/QUICKDRAW", quickDraw);
    return true;
  }

  public void setVName(String name) {
    int iVec = df.findRec(name); //TODO df == null case
    bVec = (iVec >= 0) ? df.getCompOffset(iVec) : -1; //TODO df == null case
    dVec = (iVec >= 0) ? Data.getSPA(df.getCompFormat(iVec).charAt(0)) : 1; //TODO df == null case
  }

  public void setVScale(double scale) {
    vscale = scale;
  }

  @Override
  public synchronized void setSize(int size) {
    super.setSize(size);
    if (df == null) {
      // This is for cases that we have no file and are just holding things in memory
      // e.g. xgeoplot  - Since 3.7.0 for Bug 2917 & Bug 2965
      initialized = true; // no init needed
      if (points == null) points = new Pnt[size];
      else if (points.length != size) points = Arrays.copyOf(points, size);
    }
    else {
      // At this point initialized should be false, so that we go and initialize it.
      // There are two special cases: file size is 0 and file is a pipe; in both
      // of these cases the initialization skips out, so we force initialized = true
      // so the plot can start up, and we can later add data to them.
      initialized = ((size == 0) || (df.isPipe()));
      points = null;
    }
    colors = MColor.getColorMap(MColor.T5, -1);
    line.setColor(colors[MColor.T5_INDEX_TRACE5000]);
  }

  @Override
  public synchronized void process() {
    super.process();
    if (time > 0 && !initialized) {
      initialized = true;
      matchTimePoint();
      return;
    }

    if (df == null) {
      // This is for cases that we have no file and are just holding things in memory
      // e.g. xgeoplot - Since 3.7.0 for Bug 2917 & Bug 2965
      initialized = true;
      return;
    }

    for (int i = 0; i < data.getSize(); i++) {
      int boff = i * data.bpe;
      Pnt point = new Pnt();
      if (bPos != -1) { // get the Position Vector
        int off = bPos + boff;
        double[] pos = new double[] { data.unpackD(off), data.unpackD(off + 8),
            data.unpackD(off + 16) };
        if (toLLA)
          transform.toLonLatAlt(pos, 1);
        point.setPos(pos[2], pos[1], pos[0]);
      }
      if (bVec != -1) { // get the Position Pointer Vector
        int off = bPos + boff;
        int voff = bVec + boff;
        double[] posp = new double[] { data.unpackD(off + 0), data.unpackD(off + 8),
            data.unpackD(off + 16) };
        for (int j = 3 - dVec; j < 3; j++, voff += 8)
          posp[j] += vscale * data.unpackD(voff);
        if (toLLA)
          transform.toLonLatAlt(posp, 1);
        point.setPosp(posp[2], posp[1], posp[0]);
      }
      if (bPosp != -1) { // get the Position Pointer Vector
        int off = bPosp + boff;
        double[] posp = new double[] { data.unpackD(off), data.unpackD(off + 8),
            data.unpackD(off + 16) };
        if (toLLA)
          transform.toLonLatAlt(posp, 1);
        point.setPosp(posp[2], posp[1], posp[0]);
      }
      if (bCone != -1) { // get the Conic Vector
        int off = bCone + boff;
        if (fCone.charAt(0) == 'V') {
          point.setCone(data.unpackD(off), data.unpackD(off + 8), data.unpackD(off + 16));
        }
        else { // For old files with single-element cone.
          // JC4 -- Changed from unpackD(off+16) in 2.4.1 to unpackD(off) in 2.5.0
          point.setCone(0.0, 0.0, data.unpackD(off));
        }
      }
      if (bElps != -1) { // get the Ellipse Vector
        int off = bElps + boff;
        point.setElps(data.unpackD(off), data.unpackD(off + 8), data.unpackD(off + 16));
      }
      if (bFlag != -1) { // get the flags
        int off = bFlag + boff;
        point.setFlagNameLen(data.unpackB(off + 0));
        point.setFlagNamePos(data.unpackB(off + 1));
        point.setFlagFgColor(data.unpackB(off + 2));
        point.setFlagBgColor(data.unpackB(off + 3));
        point.setFlagCone(data.unpackB(off + 4));
        point.setFlagInvert(data.unpackB(off + 5));
        point.setFlagSymCode(data.unpackB(off + 6));
        point.setFlagSymSize(data.unpackB(off + 7));
      }
      if (bConn != -1) { // get the connect flags
        int off = bConn + boff;
        point.setConn(data.unpackB(off));
      }
      if (bHdg != -1) { // get the heading value
        int off = bHdg + boff;
        point.setHdg(data.unpackF(off));
      }
      if (bName != -1) { // get the point name
        int off = bName + boff;
        int len = Data.getBPA(fName);
        point.setName(data.unpackS(off, len));
      }
      if (bNote != -1) { // get the point note
        int off = bNote + boff;
        int len = Data.getBPA(fNote);
        point.setNote(data.unpackS(off, len));
      }
      addPoint(point);
    }
    initialized = true;
  }

  /** Since 3.1.0: Method assigns a specific point if a time value has been given. */
  private void matchTimePoint() {
    Pnt point = new Pnt();

    // Retrieve point, Transformed and interpolated (if needed)
    double[] p = StateVector.getPosVelAccAtTime(df, time, data, toLLA);
    point.setPos(p[2], p[1], p[0]);

    // get current lat/lon values for optimization purposes
    if (timePoint != null) {
      minMaxSet = false;
      updateRange(timePoint);
    }
    timePoint = point;
    updateRange(point);
    addPoint(point);
  }

  /**
   * TBA
   * @param attr Table of attributes (Color, Diameter, ElevAngle)
   */
  public void setConeAttr(Table attr) {
    // This method was added in 3.3.3 to support the WorldWind Option Tree in
    // defining cones for LayerT5 space objects. See nxm.worldwind.prim.worldwind.java
    // for implementation details
  }

  /**
   * Set the reference time. This is the file-relative, for cases where the file starts at midnight
   * of a given day, this will be the number of seconds since midnight. To match X-Midas, it is
   * assumes that very large numbers (greater than 86400*100) are meant to be J1950 and thus this
   * method will pass-through to {@link #setTimeCode(double)}.
   * @param value Time in seconds.
   * @see #setTimeCode
   * @since NeXtMidas 3.1.1
   */
  @Override
  public void setTime(double value) {
    // this method added in 3.1.1 to allow for plotting a single position from the file without using LayerGeo.

    // for large numbers, assume they meant to set J1950 time
    if (value > (86400 * 100)) {
      setTimeCode(value);
    }
    else
      doSetTime(value);
  }

  /**
   * Set the time via {@link Layer#setTime(double)}, then set {@link #points} to null and call
   * {@link #matchTimePoint()}
   * @param value Time as a double
   */
  private void doSetTime(double value) {
    super.setTime(value);
    points = null; // remove old points from array
    if (initialized)
      matchTimePoint();
  }

  /**
   * Get the reference time. This is the file-relative, for cases where the file starts at midnight
   * of a given day, this will be the number of seconds since midnight.
   * @return Time in seconds.
   * @see #getTimeCode
   * @since NeXtMidas 3.1.1
   */
  @Override
  public double getTime() {
    return super.getTime();
  }

  /**
   * Sets the timecode in J1950 time. This computes the file-relative time and then calls {@link #setTime}.
   * @param value Time in seconds since midnight, 1 Jan 1950.
   * @return true if the timecode is within the range of the current state vector file, false if not.
   *         This can be used to identify cases where the file is no longer valid and must be changed.
   * @since NeXtMidas 3.1.1
   */
  public boolean setTimeCode(double value) {
    if ((df == null) || (data == null)) {
      setTime(value);
      return false;
    }

    double start = df.getTimeCode();
    double end = start + (df.getXDelta() * df.getSize());
    doSetTime(value - start);
    return ((value >= start) && (value <= end));
  }

  /**
   * Gets the timecode in J1950 time. This is computed based on the file-relative time from {@link #getTime}.
   * @return Time in seconds since midnight, 1 Jan 1950.
   * @since NeXtMidas 3.1.1
   */
  public double getTimeCode() {
    if ((df == null) || (data == null))
      return getTime();

    return getTime() + df.getTimeCode();
  }

  /** Updates the internal min/max values, see findRange(). */
  private void updateRange(Pnt point) {
    if (!minMaxSet) {
      minLon = point.minLon;
      maxLon = point.maxLon;
      minLat = point.minLat;
      maxLat = point.maxLat;
      minAlt = point.minAlt;
      maxAlt = point.maxAlt;
      minMaxSet = true;
    }
    else {
      minLon = Math.min(minLon, point.minLon);
      maxLon = Math.max(maxLon, point.maxLon);
      minLat = Math.min(minLat, point.minLat);
      maxLat = Math.max(maxLat, point.maxLat);
      minAlt = Math.min(minAlt, point.minAlt);
      maxAlt = Math.max(minAlt, point.maxAlt);
    }
  }

  @Override
  public void findRange() {
    // JC4: When plotting a pipe this is called following every pipe read that has new dats.
    // To avoid recomputing for the large number of unchanged points, use updateRange(..)
    // to update min/max values as new points come in. The only "problem" with this is
    // that the range does not rescale as points drop off. My guess is that such is not
    // an issue since LayerT5 is usually overlaid on a map where the bounds of the map
    // dominate the plot.
    if (minMaxSet) {
      x1 = minLon;
      x2 = maxLon;
      y1 = minLat;
      y2 = maxLat;
      z1 = minAlt;
      z2 = maxAlt;
      a1 = y1;
      a2 = y2;
    }
  }

  /** Sets the Color of the selected point
   *  @param value The color
   *  @see MColor#getColor(Object)
   *  @since NeXtMidas 3.5.3 for Bug 2888
   */
  public void setSelectColor (Object value) {
    if ((value instanceof String) && (value.toString().length()==0)) return;
    Color c = MColor.getColor(value);
    setSelectColor(c);
  }

  /** Sets the Color of the selected point
   *  @param c the Color to use for selected points
   *  @since NeXtMidas 3.5.3 for Bug 2888
   */
  public void setSelectColor (Color c) {
    if (c.equals(selectColor)) return; // no change
    selectColor = c;
    refresh();
  }

  /** Gets the Color used for selected points
   *  @return selectColor Color used for selected points
   *  @since NeXtMidas 3.5.3 for Bug 2888
   */
  public Color getSelectColor(){
    return selectColor;
  }

  @Override
  public synchronized void draw(int flag) {
    if (!initialized || (size == 0 && (pf == null))) { // add check for size - Since 3.7.0 for Bug 2917 & Bug 2965
      return; // JC4: Prevents race condition
    }
    if (MP.gc == null) {
      return;
    }
    // if (needRefresh && !df.isPipe()) process(); // JC4: This may read in data after end of the pipe
    super.draw(flag); // resets needRefresh
    Line lin = line.copy();

    // Set vreg (viewable region)
    double vreg;
    if (MP.view == MPlot.V_GEODETIC || MP.view == MPlot.V_POLAR) {
      vreg = 50.0 * (Math.min(MP.rx2 - MP.rx1, MP.ry2 - MP.ry1));
    }
    else {
      vreg = Math.min(MP.rx2 - MP.rx1, MP.ry2 - MP.ry1);
    }

    // Determine what needs to be drawn
    _labels = showField(LABELS) && (bName != -1) && ((enable & MPlot.EN_LABELS) != 0);
    _labelZoom = (vreg < tlreg);
    _symbols = showField(SYMBOLS) && lin.isTypeSet(Line.SYMBOL);
    _connect = showField(CONNECT) && (bConn != -1);
    _connectAll = timeAxis && ((flags & SHOW_TIME_SCALE_LABELS) == 0);

    boolean coneFile = (conic != 0) && (bPos != -1) && (bPosp != -1);
    boolean elpsFile = (bElps != -1) && (bPos != -1);
    boolean doPos = (bPos != -1) && (bVec == -1);
    ;
    boolean doPosp = (bPosp != -1);
    boolean doVec = (bVec != -1);
    boolean doCones = showField(CONES) && coneFile;
    boolean doEllipses = showField(ELLIPSES) && elpsFile;

    // not sure how to optimize these, default to it being shown
    _showAll = (!MP.is2D) || _connect || _connectAll || doCones;

    if (!_showAll) {
      // Are all points visible (this is common when zoomed out)?
      _showAll = (minLat >= MP.ry1) && (maxLat <= MP.ry2) && (minLon >= MP.rx1) && (maxLon <= MP.rx2);
    }

    // Set the font and color.
    Font f = null;
    if (fontsize > 0 && _symbols) {
      f = MP.gc.getFont();
      MP.gc.setFont(font);
    }
    MP.gc.setColor(lin.getColor());

    // Draw everything.
    Pnt point = null;
    Pnt last = null;
    int saveType = lin.getTypeInt();
    Color saveColor = MP.gc.getColor();

    for (int i = 0; i < numPoints; i++) {

      point = getVisiblePoint(i, doPosp || doVec, doCones, doEllipses);
      if (point != null && !point.isHidden()) {
        point.prep(doPosp);
        _selected = (select == i);
        _color = saveColor;
        if (bFlag != -1) {
          _color = colors[Math.min(point.flagFgColor, colors.length - 1)]; // subtract one from colors.length
                                                                           // because it is zero based
        } else if (point.flagFgColor > 0) { // (BUG 2823) allow control for files without FLAG component
          _color = colors[Math.min(point.flagFgColor, colors.length - 1)];
        }
        if (_selected)
          _color = selectColor;

        MP.gc.setColor(_color);
        lin.setColor(_color);
        lin.setType(Line.SOLID);

        if (doVec && last != null && Math.abs(point._ixPos - last._ixPos) < 5
            && Math.abs(point._iyPos - last._iyPos) < 5)
          continue;
        if (doVec)
          drawVec(lin, point);
        if (doPos)
          drawLabel(lin, point, last, true);
        if (doPosp)
          drawLabel(lin, point, null, false);
        if (doCones)
          drawCone(lin, point);
        if (doEllipses)
          drawEllipse(lin, point);
      }
      if (last != null)
        last.done();
      last = (point != null && point.isHidden()) ? null : point;

      MP.gc.setColor(saveColor);
      lin.setColor(saveColor);
      lin.setType(saveType);
    }
    if (last != null)
      last.done();

    // Restore the font and color
    if (f != null)
      MP.gc.setFont(f);
    MP.gc.setColor(colors[0]);
  }

  /** Draws the labels. -- only used by draw(). */
  private void drawLabel(Line lin, Pnt point, Pnt last, boolean isPos) {
    if (MP.gc == null)
      return;
    int ix = (isPos) ? point._ixPos : point._ixPosp;
    int iy = (isPos) ? point._iyPos : point._iyPosp;
    Symbol symbol = lin.getSymbol();
    float prevSymAng = lin.getSymbolHeading();
    String name = point.name;
    boolean sym = true;
    boolean conn = (_connect && point.conn) || _connectAll;
    boolean noAuto = (flags & AUTO_SCALE_LABELS) == 0;
    boolean lbl = (noAuto || _selected || _labelZoom);

    if (bFlag != -1) {
      // Set the defaults for files with flags
      lbl = true;
      sym = true;

      if (_symbols && isPos) {
        String symbolName = XmUtil.getXmSymbolName(point.flagSymCode);
        int symbolSize = point.flagSymSize;

        if (symbolName == null) {
          warn("Invalid symbol ID (" + point.flagSymCode + ") for plot (" + name + ") using default.");
        } else if ((symbolSize <= 0) || (symbolName.equals("NULL"))) {
          sym = false; // Turn the symbol off
        } else {
          symbol = Symbol.get(symbolName, point.flagSymSize);
        }
      }

      if (_labels && isPos) {
        byte nameLength = point.flagNameLen;

        if (nameLength == 0) {
          lbl = false;
        }
        else if (name.length() <= nameLength) {
          // Do Nothing - Name is Fine
        }
        else if ((nameLength > 0) && (nameLength <= name.length())) {
          name = name.substring(0, nameLength);
        }
        else {
          warn("Invalid name length (" + nameLength + ") for plot (" + name + ").");
        }
      }
    }

    if (conn && (last != null) && isPos) {
      if (!MP.is2D || Math.abs(point.posLon - last.posLon) < 180)
        MP.gc.drawLine(last._ixPos, last._iyPos, ix, iy);
    }

    if (noAuto || (bFlag != -1) || (ix > MP.ix1 && ix < MP.ix2 && iy > MP.iy1 && iy < MP.iy2)) {
      if (_symbols && sym && isPos && (symbol != null)) {
        if (bHdg != -1)
          lin.setSymbolHeading(point.hdg);
        symbol.draw(ix, iy, MP.gc);
      }
      if (_labels && isPos && lbl) {
        if (_selected) {
          name = getSelectedText();
        }
        // Determine the x-offset of the label; left vs right of point
        int xoff = getLabelxoffset(lin.getRadius(), name);
        if (name.indexOf('\n') < 0) {
          MP.gc.drawString(name, ix + xoff, iy + lin.getRadius());
        }
        else {
          String[] lines = name.split("\n");
          int lineHeight = MP.gc.getFontMetrics().getHeight();
          for (int lineNum = 0; lineNum < lines.length; lineNum++) {
            int delta = lineHeight * lineNum;
            MP.gc.drawString(lines[lineNum], ix + xoff, iy + lin.getRadius() + delta);
          }
        }
      }
    }
    lin.setSymbolHeading(prevSymAng); // restore previous angle
  }

  /** Draws the ellipses. -- only used by draw(). */
  private void drawEllipse(Line lin, Pnt point) {
    if ((bFlag == -1) || (point.flagCone != 0)) {
      // Estimate if the ellipse needs to be drawn. If the symbol is drawn and the ellipse is
      // smaller than the symbol, there is no need to draw it.
      double rad = Math.max(point.elpsSemiMajor, point.elpsSemiMinor);
      double dy = rad * METERS_TO_DEG_LAT * MP.myy;
      double dx = rad * METERS_TO_DEG_LON * MP.mxx;
      int r = lin.getSymbol().getSize();

      if (!_symbols || (dx > r) || (dy > r)) {
        if ((flags & FILL_CONES) != 0) {
          lin.setType(Line.FILL);
          lin.setColor(new Color(0x70FFFFFF & _color.getRGB(), true));
          MP.drawEllipse(point.posLon, point.posLat, point.posAlt, point.elpsSemiMajor,
              point.elpsSemiMinor, (90.0 - point.elpsTilt), -1, lin);
          lin.setType(Line.SOLID);
        }
        lin.setColor(_color);
        MP.drawEllipse(point.posLon, point.posLat, point.posAlt, point.elpsSemiMajor,
            point.elpsSemiMinor, (90.0 - point.elpsTilt), -1, lin);
      }
    }
  }

  /** Draws the cones -- only used by draw(). */
  private void drawCone(Line lin, Pnt point) {
    // Determine which cone format(s) to draw
    boolean drawHoriz = (flags & CONE_FLAG_HORIZ) != 0;
    boolean drawVert = (flags & CONE_FLAG_VERT) != 0;

    // Set the line type to use.
    int hLineType = Line.SOLID;
    int vLineType = (drawHoriz && drawVert) ? Line.POINT : Line.SOLID;

    // Init the positions.
    Position pPos = (bFlag != -1) ? point.getPosp() : point.getPos();
    Position qPos = (bFlag == -1) ? point.getPosp() : point.getPos();

    if (bCone != -1)
      cangle = point.coneAngle;

    // DRAW CONE
    if ((bFlag == -1) || (point.flagCone != 0)) {
      if ((flags & FILL_CONES) != 0) {
        lin.setType(Line.FILL);
        lin.setColor(new Color(0x70FFFFFF & _color.getRGB(), true));
        if (drawVert) {
          MP.drawCone(pPos, qPos, conic, cdiv, cangle, crange, lin, 0);
        }
        if (drawHoriz) {
          MP.drawCone(pPos, qPos, conic, cdiv, cangle, crange, lin, 1);
        }
        lin.setColor(_color);
      }
      if (drawVert) {
        lin.setType(vLineType);
        MP.drawCone(pPos, qPos, conic, cdiv, cangle, crange, lin, 0);
      }
      if (drawHoriz) {
        lin.setType(hLineType);
        MP.drawCone(pPos, qPos, conic, cdiv, cangle, crange, lin, 1);
      }
      lin.setType(Line.SOLID);
    }
  }

  /** Draws a vector -- only used by draw(). */
  private void drawVec(Line lin, Pnt point) {
    MP.gc.drawLine(point._ixPos, point._iyPos, point._ixPosp, point._iyPosp);
  }

  /* Determine the offset necessary to place LABEL to left or right of point */
  private int getLabelxoffset(int radius, String label) {
    if (MP.gc == null)
      return -1;
    int xoff = 2 + radius;
    if (getLeft()) {
      FontMetrics fm = MP.gc.getFontMetrics();
      xoff = -1 * (fm.bytesWidth(label.getBytes(), 0, label.length()) + xoff);
    }
    return xoff;
  }

  /** Gets the number of cuts (slices of the cone) to show for a file with cones. */
  public int getConic() {
    return conic;
  }

  /** Gets the number of divisions between POS and POSP to draw the cuts on (typically around 60). */
  public int getConeDiv() {
    return cdiv;
  }

  /**
   * Gets the cone angle (in deg). This value is typically superseded by the CONE field in the file.
   */
  public double getConeAngle() {
    return cangle;
  }

  /** Gets the cone range (0=Compute range based on POS, POSP). */
  public double getConeRange() {
    return crange;
  }

  /**
   * Sets the number of cuts (slices of the cone) to show for a file with cones.
   * @param value The number of cuts (must be at least one). A value of 1 will display the cone only
   *          where it hits the surface of the Earth.
   */
  public void setConic(int value) {
    if (conic != value) {
      conic = value;
      refresh();
    }
  }

  /** Sets the number of divisions between POS and POSP to draw the cuts on (typically around 60). */
  public void setConeDiv(int value) {
    if (cdiv != value) {
      cdiv = value;
      refresh();
    }
  }

  /** Gets the cone angle (in deg). This value is typically superseded by the CONE field in the file. */
  public void setConeAngle(double value) {
    if (cangle != value) {
      cangle = value;
      refresh();
    }
  }

  /** Sets the cone range (0=Compute range based on POS, POSP). */
  public void setConeRange(double value) {
    if (crange != value) {
      crange = value;
      refresh();
    }
  }

  /**
   * Gets the Toggle Label Region. This controls the point, when zooming in, at which text labels
   * are shown (show label if {@code viewable < tlreg }). This value is ignored if the file has a
   * FLAG field that specifies if a label is shown.
   */
  public double getTLReg() {
    return tlreg;
  }

  /** Sets the Toggle Label Region, see {@link #getTLReg()} for details. */
  public void setTLReg(double value) {
    if (tlreg != value) {
      tlreg = value;
      refresh();
    }
  }

  /** get and set the Labels left placement */
  public boolean getLeft() {
    return leftLabels;
  }

  public void setLeft(int value) {
    if (value == 1) {
      leftLabels = true;
    }
    else {
      leftLabels = false;
    }
    refresh();
  }

  /** get and set the toggle Label region */
  public int getFontSize() {
    return fontsize;
  }

  public void setFontSize(int value) {
    fontsize = value;
    if (fontsize > 0)
      font = MP.newFont(MP.font, "", -1, fontsize);
  }

  // ////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods to set/get the selected point
  // ////////////////////////////////////////////////////////////////////////////////////////////////
  /**
   * Selects the nearest point.
   * @deprecated Since NeXtMidas 2.1.0: Use <code>selectNearest(pos,NEAREST_BY_PIXELS)</code>.
   */
  @Deprecated
  public void select(Position pos) {
    // deprecated("select(Position) method is deprecated use selectNearest(Position,\"Pixel\").");
    double[] pp = new double[3];
    pp[0] = pos.x;
    pp[1] = pos.y;
    pp[2] = pos.z;
    select(pp);
  }

  /**
   * Selects the nearest point.
   * @deprecated Since NeXtMidas 2.1.0: Use <code>selectNearest(pos,NEAREST_BY_PIXELS)</code>.
   */
  @Deprecated
  public void select(double[] pos) {
    setSelectedIndex(getNearestByPixels(pos, 25 * 25));
  }

  /**
   * Selects the given point.
   * @deprecated Since NeXtMidas 2.1.0: Use <code>setSelectedIndex(offset)</code>.
   */
  @Deprecated
  public void select(int offset) {
    // deprecated("select(index) method is deprecated use setSelectedIndex(index).");
    setSelectedIndex(offset);
  }

  /**
   * Gets the index of the selected point.
   * @deprecated Since NeXtMidas 2.1.0: Use <code>getSelectedIndex()</code>.
   */
  @Deprecated
  public int select() {
    // deprecated("select() method is deprecated use getSelectedIndex().");
    return getSelectedIndex();
  }

  /**
   * Selects the nearest point (POS) to the given position. This method is intended for use from
   * inside a macro.
   * @param mode The selection mode (either "Pixel" or "GreatArc").
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   */
  public Table selectNearest(String mode, Position position) {
    setSelectedIndex(getNearest(mode, position));

    return getSelected();
  }

  /**
   * Selects the nearest point (POS) to the given position. This method is intended for use from
   * inside a macro.
   * @param mode The selection mode (either "Pixel" or "GreatArc").
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @param range The maximum distance (or -1.0 for unlimited).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   */
  public Table selectNearest(String mode, Position position, double range) {
    setSelectedIndex(getNearest(mode, position, range));

    return getSelected();
  }

  /**
   * Selects the nearest point (POS) to the given position. This method simply calls
   * {@link #selectNearest(int,Position,double)} with a default range value. The defaults are:
   *
   * <pre>{@code
   *         mode={@link #NEAREST_BY_PIXELS}    --> range=625.0  [nxm200 default]
   *         mode={@link #NEAREST_BY_GREAT_ARC} --> range=-1.0   [unlimited range]
   * }</pre>
   */
  public Table selectNearest(int mode, Position position) throws IllegalArgumentException {
    setSelectedIndex(getNearest(mode, position));

    return getSelected();
  }

  /**
   * Selects the nearest point (POS) to the given position. No point will be selected if none is
   * within the given range (use <code>range=-1.0</code> for unlimited range), this can be used to
   * prevent selecting points that are far away from where the user selected on the plot.<br>
   * <br>
   * The following selection modes are available:
   *
   * <pre>
   *         {@link #NEAREST_BY_PIXELS}
   *           The nearest point is determined based on its pixel position in the
   *           plot (i.e. the point that is nearest "visually" will be selected).
   *           The range is set in pixels and should be an integer value. If the
   *           given position's reference frame is set to GEODETIC it will use
   *           the positions lon,lat,alt values to define the position; otherwise
   *           it will use the x,y,z values; no attempt will be made to perform
   *           a conversion of the coordinate system.
   *
   *         {@link #NEAREST_BY_GREAT_ARC}
   *           The nearest point is determined based on the great arc distance
   *           between points. This is the most accurate for using geodetic or
   *           Lon-Lat-Alt plots (which is the majority of the Type 5000 files
   *           plotted). The range is set in real-world coordinates. If the
   *           given position's reference frame is not set to GEODETIC, it will
   *           be converted to GEODETIC coordinates before computing the nearest
   *           point.
   * </pre>
   * @param mode The selection mode (either {@link #NEAREST_BY_PIXELS} or {@link #NEAREST_BY_GREAT_ARC}).
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @param range The maximum distance (or -1.0 for unlimited).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   * @throws IllegalArgumentException If the mode is invalid.
   */
  public Table selectNearest(int mode, Position position, double range)
      throws IllegalArgumentException {
    setSelectedIndex(getNearest(mode, position, range));

    return getSelected();
  }

  /**
   * Sets the selected point based on the index value of the point.
   * @param index The index in the file (0..n-1) of the point to select or -1 to select none.
   */
  public void setSelectedIndex(int index) {
    if (select != index) {
      select = index;
      refresh();
    }
  }

  /** Gets the index value of the selected point (0..n-1) or -1 if none is selected. */
  public int getSelectedIndex() {
    return (select < 0) ? -1 : select;
  }

  /**
   * Gets the data for the selected point as a table. The table will include the following fields:
   *
   * <pre>
   *         INDEX - The index of the point.                    (---ALWAYS---)
   *         NAME  - The name of the point.                     (If Available)
   *         POS   - The POS position.                          (If Available)
   *         POSP  - The POSP position.                         (If Available)
   *         ELPS  - The ellipse parameters.                    (If Available)
   *         CONE  - The cone parameters.                       (If Available)
   *         FLAG  - The set of flags setting the drawing mode. (If Available)
   * </pre>
   * @return The table of data for the selected point or <code>null</code> if none is selected.
   */
  public Table getSelected() {
    return (select < 0) ? null : getPoint(select);
  }

  // ////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods to manage the text displayed for the selected point.
  // ////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the keys to include when using {@link #getSelectedText()}. This should be a
   * comma-separated list of field names (in the order to be shown). <br>
   * <br>
   * Examples:
   *
   * <pre>
   *         setSelectionKeys("NAME,POS")
   *         setSelectionKeys("NAME,POS,POSP,CONE")
   * </pre>
   * @param keys The keys to show use <code>null</code> to show only the <code>NAME</code> field
   *          without the <code>"NAME="</code> prefix.
   */
  public void setSelectionKeys(String keys) {
    if (keys == null) {
      selectedKeys = null;
    }
    else {
      selectedKeys = keys.trim().toUpperCase().split(SPLIT_REGEX);
    }
  }

  /**
   * Gets the text of the selected entries.
   * @return The selected text or "" if none are selected.
   */
  public String getSelectedText() {
    return getSelectedText(selectedKeys);
  }

  /**
   * Gets the text of the selected entries.
   * @param keys The keys to show (often <code>{"NAME"}</code> or <code>{"NAME","POS"}</code>) use
   *          <code>null</code> to show only the <code>NAME</code> field without the
   *          <code>"NAME="</code> prefix.
   * @return The selected text or "" if none are selected.
   */
  private String getSelectedText(String[] keys) {
    final Table table = getSelected();
    final StringBuffer str = new StringBuffer();

    if (table != null) {
      if (keys == null) {
        str.append(table.get("NAME"));
      }
      else {
        for (int i = 0; i < keys.length; i++) {
          Object value = table.get(keys[i]);

          if (value != null) {
            if (value instanceof Position) { /* JC4: Make positions easy-to-read. */
              Position pos = (Position) value;

              if (pos.referenceFrame == Position.GEODETIC) {
                value = Convert.deg2dmsUnicode(pos.getLon(), "EW") + " "
                    + Convert.deg2dmsUnicode(pos.getLat(), "NS") + " " + ((float) pos.getAlt()) + "m";
              }
            }

            if (str.length() > 0) {
              str.append("\n");
            }
            str.append(StringUtil.padRight(keys[i], 4));
            str.append(" = ");
            str.append(value);
          }
        }
      }
    }

    return str.toString();
  }

  // ////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods that get the nearest point.
  // ////////////////////////////////////////////////////////////////////////////////////////////////
  /**
   * Selects the nearest point (POS) to the given position. <br>
   * <br>
   * This method simply calls {@link #selectNearest(int,Position,double)} with a default range
   * value. The defaults are:
   *
   * <pre>{@code
   *         mode="Pixel"    --> range=625.0  [nxm200 default]
   *         mode="GreatArc" --> range=-1.0   [unlimited range]
   * }</pre>
   * @param mode The selection mode (either {@link #NEAREST_BY_PIXELS} or {@link #NEAREST_BY_GREAT_ARC}).
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   * @throws IllegalArgumentException If the mode is invalid.
   */
  public int getNearest(String mode, Position position) throws IllegalArgumentException {
    return getNearest(Parser.find(selectionModes, mode, -1), position);
  }

  /**
   * Selects the nearest point (POS) to the given position. This method is intended for use from
   * inside a macro.
   * @param mode The selection mode (either "Pixel" or "GreatArc").
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @param range The maximum distance (or -1.0 for unlimited).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   * @see #selectNearest(int,Position)
   */
  public int getNearest(String mode, Position position, double range) {
    return getNearest(Parser.find(selectionModes, mode, -1), position, range);
  }

  /**
   * Selects the nearest point (POS) to the given position. <br>
   * <br>
   * This method simply calls {@link #selectNearest(int,Position,double)} with a default range
   * value. The defaults are:
   *
   * <pre>{@code
   *         mode={@link #NEAREST_BY_PIXELS}    --> range=625.0  [nxm200 default]
   *         mode={@link #NEAREST_BY_GREAT_ARC} --> range=-1.0   [unlimited range]
   * }</pre>
   * @param mode The selection mode (either {@link #NEAREST_BY_PIXELS} or {@link #NEAREST_BY_GREAT_ARC}).
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   * @throws IllegalArgumentException If the mode is invalid.
   */
  public int getNearest(int mode, Position position) throws IllegalArgumentException {
    switch (mode) {
      case NEAREST_BY_PIXELS:
        return getNearest(mode, position, 625.0);
      case NEAREST_BY_GREAT_ARC:
        return getNearest(mode, position, -1.0);
      default:
        return getNearest(mode, position, -1.0);
    }
  }

  /**
   * Selects the nearest point (POS) to the given position. No point will be selected if none is
   * within the given range (use <code>range=-1.0</code> for unlimited range), this can be used to
   * prevent selecting points that are far away from where the user selected on the plot.<br>
   * <br>
   * The following selection modes are available:
   *
   * <pre>
   *         {@link #NEAREST_BY_PIXELS}
   *           The nearest point is determined based on its pixel position in the
   *           plot (i.e. the point that is nearest "visually" will be selected).
   *           The range is set in pixels and should be an integer value. If the
   *           given position's reference frame is set to GEODETIC it will use
   *           the positions lon,lat,alt values to define the position; otherwise
   *           it will use the x,y,z values; no attempt will be made to perform
   *           a conversion of the coordinate system.
   *
   *         {@link #NEAREST_BY_GREAT_ARC}
   *           The nearest point is determined based on the great arc distance
   *           between points. This is the most accurate for using geodetic or
   *           Lon-Lat-Alt plots (which is the majority of the Type 5000 files
   *           plotted). The range is set in real-world coordinates. If the
   *           given position's reference frame is not set to GEODETIC, it will
   *           be converted to GEODETIC coordinates before computing the nearest
   *           point.
   * </pre>
   * @param mode The selection mode (either {@link #NEAREST_BY_PIXELS} or {@link #NEAREST_BY_GREAT_ARC}).
   * @param position The given position (usually the result of the user selecting a point on the plot).
   * @param range The maximum distance (or -1.0 for unlimited).
   * @return The table of data for the selected point or <code>null</code> if none is selected (see
   *         {@link #getSelected()} for details about the contents of the table).
   * @throws IllegalArgumentException If the mode is invalid.
   */
  public int getNearest(int mode, Position position, double range) throws IllegalArgumentException {
    switch (mode) {
      case NEAREST_BY_PIXELS:
        return getNearestByPixels(position, range);
      case NEAREST_BY_GREAT_ARC:
        return getNearestByGreatArc(position, range);
      default:
        throw new IllegalArgumentException("LayerT5: Unknown selection mode (" + mode + ").");
    }
  }

  /**
   * Selects the nearest point based on a pixel distance. Prior to NeXtMidas 2.1.0 the select method
   * used this same code but made the assumption that <code>pos.x</code>, <code>pos.y</code>, and
   * <code>pos.z</code> were the valid position coordinates (ignoring the reference frame set in the position).
   * @param pos The position.
   * @param range The maximum range.
   * @return The index of the nearest point or (-1 if none found within the given range).
   */
  private int getNearestByPixels(Position pos, double range) {
    double[] pp = new double[3];

    if (pos.referenceFrame == Position.GEODETIC) {
      pp[0] = pos.x;
      pp[1] = pos.y;
      pp[2] = pos.z;
    }
    else {
      pp[0] = pos.x;
      pp[1] = pos.y;
      pp[2] = pos.z;
    }

    return getNearestByPixels(pp, range);
  }

  /**
   * Selects the nearest point based on a pixel distance. Prior to NeXtMidas 2.1.0 the select method
   * used this code to select the nearest point except that the range was hard-coded to be 25*25.
   * @param pos The position.
   * @param range The maximum range.
   * @return The index of the nearest point or (-1 if none found within the given range).
   */
  private int getNearestByPixels(double[] pos, double range) {
    int select = -1;

    if (bPos == -1) {
      MPoint ppos = new MPoint();
      MP.viewer.rwc2pix(pos, ppos);
      int dist;
      int mdist = (int) Math.ceil(range); // maximum pixel radius
      for (int i = 0; i < numPoints; i++) {
        Pnt point = points[i];

        if (point != null) {
          points[i].getPosPix(pix);

          int ix = pix.x[0];
          int iy = pix.y[0];
          if (ix > MP.ix1 && ix < MP.ix2 && iy > MP.iy1 && iy < MP.iy2) {
            int dx = ix - ppos.x, dy = iy - ppos.y;
            dist = dx * dx + dy * dy;
            if (dist < mdist) {
              select = i;
              mdist = dist;
            }
          }
        }
      }
    }

    return select;
  }

  /**
   * Selects the nearest point (POS) in the layer to the given position.
   * @param pos The location to search from (must be in {@link Position#GEODETIC} coordinates).
   * @param range The maximum distance (or -1.0 for unlimited).
   * @return The index of the nearest point or (-1 if none found within the given range).
   * @since NeXtMidas 2.1.0
   */
  private int getNearestByGreatArc(Position pos, double range) {
    int select = -1;

    if (bPos == -1) {
      warn("Can not select by great arc, no POS component found.");
    }
    else {
      if (pos.referenceFrame == Position.NORMAL) {
        pos.car2geo();
      }

      Position point = new Position(Position.GEODETIC);
      double lastDist = (range < 0) ? Double.MAX_VALUE : range;

      for (int i = 0; i < numPoints; i++) {
        Pnt pnt = points[i];
        if (pnt != null) {
          point.lon = pnt.posLon;
          point.lat = pnt.posLat;
          point.alt = pnt.posAlt;

          point.referenceFrame = Position.GEODETIC;
          double dist = Transform.greatArc(pos, point);
          pos.car2geo(); // Reset pos
          if (dist <= lastDist) {
            select = i;
            lastDist = dist;
          }
        }
      }
    }
    return select;
  }

  private Pnt getPointAt(int index) {
    if (firstPoint != -1)
      index = (index + firstPoint) % numPoints; // Adjust index for pipes
    return points[index];
  }

  /** Gets the point at a given index, returns null if the point is not visible on the plot. */
  private Pnt getVisiblePoint(int index, boolean posp, boolean cones, boolean ellipses) {
    // JC4: Added next line to prevent errors when called when no points defined
    if ((points == null) || (points.length == 0)){
      System.out.println("getVisiblePoint points null or 0 points:"+points+" layerName:"+name);
      return null;
    }

    if (firstPoint != -1) {
      index = (index + firstPoint) % numPoints; // Adjust index for pipes
    }

    Pnt point = points[index];
    if (point == null) {
      return null;
    }
    if (quickDraw && MP.is2D) {
      if (!point.hit(MP.rx1, MP.rx2, MP.ry1, MP.ry2)) {
        return null;
      }
    }
    else {
      if (!isPointVisible(index) && !isPointVisible(index + 1) && !isPointVisible(index - 1)
          && !isLineVisibleToFromPoint(index)) { // isPointVisible and isLineVisible added for Bug
                                                 // 2488
        return null;
      }
    }

    // pulled the prep in here and return non-null so that we can someday
    // we can do partial draws from hidden to non-hidden points
    point.prep(posp);
    return point;
  }

  /**
   * test whether a given point is visible
   * @param index Index of the point
   * @return True if the point is visible
   * @since NeXtMidas 3.1.0
   */
  private boolean isPointVisible(int index) {
    if (index < 0 || index > (points.length - 1))
      return false;
    Pnt point = points[index];
    if (point == null)
      return false;
    if (MP.is2D && !point.hit(MP.rx1, MP.rx2, MP.ry1, MP.ry2))
      return false;
    return true;
  }

  /**
   * Test whether a line drawn to or from a given point is visible. If either line would be visible,
   * then returns true
   * @param index Index of the current point
   * @return True if the line needs to be drawn
   * @since NeXtMidas 3.1.0
   */
  private boolean isLineVisibleToFromPoint(int index) {
    boolean visNext = false;
    boolean visLast = false;
    if (index < 0)
      return false;
    Pnt point = points[index];
    if (point == null)
      return false;
    Pnt next = null;
    if ((index + 1) < points.length)
      next = points[index + 1];
    if (next != null) {
      visNext = true;
      if (((point.maxLon < MP.rx1) && (next.maxLon < MP.rx1))
          || ((point.minLon > MP.rx2) && (next.minLon > MP.rx2))
          || ((point.maxLat < MP.ry1) && (next.maxLat < MP.ry1))
          || ((point.minLat > MP.ry2) && (next.minLat > MP.ry2))) {
        visNext = false;
      }
    }
    Pnt last = null;
    if ((index - 1) >= 0)
      last = points[index - 1];
    if (last != null) {
      visLast = true;
      if (((point.maxLon < MP.rx1) && (last.maxLon < MP.rx1))
          || ((point.minLon > MP.rx2) && (last.minLon > MP.rx2))
          || ((point.maxLat < MP.ry1) && (last.maxLat < MP.ry1))
          || ((point.minLat > MP.ry2) && (last.minLat > MP.ry2))) {
        visLast = false;
      }
    }
    return visNext || visLast;
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final String getPointName(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : tbl.getS("NAME");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final String getPointNote(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : tbl.getS("NOTE");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final float[] getPointHeading(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (float[]) tbl.getKey("HDG");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final double[] getPointPos(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (double[]) tbl.getKey("POS");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final double[] getAllPointPos() {
    double[] allPos = new double[3 * numPoints];
    for (int i = 0, j = 0; j < allPos.length; i++, j += 3) {
      System.arraycopy(getPointPos(i), 0, allPos, j, 3);

    }
    return allPos;
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final double[] getAllPointPosp() {
    double[] allPos = new double[3 * numPoints];
    for (int i = 0, j = 0; j < allPos.length; i++, j += 3) {
      System.arraycopy(getPointPosp(i), 0, allPos, j, 3);

    }
    return allPos;
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final double[] getPointPosp(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (double[]) tbl.getKey("POSP");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final double[] getPointElps(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (double[]) tbl.getKey("ELPS");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final double[] getPointCone(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (double[]) tbl.getKey("CONE");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final byte[] getPointFlag(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (byte[]) tbl.getKey("FLAG");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final byte getPointFlag(int index, int id) {
    byte[] flags = getPointFlag(index);
    return (flags == null) ? 0 : flags[id];
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected final byte[] getPointConn(int index) {
    Table tbl = getPoint(index);
    return (tbl == null) ? null : (byte[]) tbl.getKey("CONN");
  }

  /** @deprecated Since NeXtMidas 2.5.0: use {@link #getPoint(int)} */
  @Deprecated
  protected int getPointCount() {
    return numPoints;
  }

  public Table getPoint(int index) {
    Table tbl = null;

    if ((index >= 0) && (index < numPoints)) {
      if (points[index] != null) {
        tbl = points[index].toTable();
      }
    }
    return tbl;
  }

  /**
   * Set/replace an existing point at specified index -or-
   * Add a new point, if index is negative or greater than current number of points.
     The input table can have the following fields:
   *
   * <pre>
   *         NAME=  - A String (any other object is ok, .toString() is used)
   *         NOTE=  - A String (any other object is ok, .toString() is used)
   *         CONN=  - A byte[1] or a Data (CONN)
   *         POS=   - A Position, an Data (LON,LAT,ALT), or Table {ALT=,LAT=,LON=}
   *         POSP=  - A Position, an Data (LON,LAT,ALT), or Table {ALT=,LAT=,LON=}
   *         ELPS=  - A double[3] or Data (semiMajor,semiMinor,tilt)
   *         CONE=  - A double[3] or Data (radius,angle,angle)
   *         FLAG=  - A byte[8] or Data (NAMELEN,NAMEPOS,COLOR,BGCOLOR,CONE,
   *                  INVERT,SYMBOLID,SYMBOLSIZE) or Table {NAMELEN=,NAMEPOS=,
   *                  COLOR=,BGCOLOR=,CONE=,INVERT=,SYMBOLID=,SYMBOLSIZE=}
   * </pre>
   * @param index The index of the point (-1 to add new point to the end).
   * @param table The point.
   */
  public void setPoint(int index, Table table) {
    if (numPoints == 0) {
      // initialize component fields if this is the first point for this LayerT5 - since 3.7.0 bug 2965
      updatePointsComponents(table);
    }
    else {
      // validate point components match existing points (merging pt:false) - since 3.7.0 for bug 2965
      validatePointComponents(table,false);
    }
    Pnt point = new Pnt(table);

    if (index < 0) {
      index = numPoints;
    }

    if (index >= numPoints) {
      while (index + 1 < numPoints) {
        addPoint(null, true);
      }
      addPoint(point, true);
      setSize(points.length);
    }
    else {
      points[index] = point;
    }
    // process(); // commented out in 3.3.2 since it is unnecessary and grows the points array with dupes
    refresh();
  }

  /** Merge new point properties with an existing point. The input table can have the following fields:
     <pre>
       NAME=  - A String (any other object is ok, .toString() is used)
       NOTE=  - A String (any other object is ok, .toString() is used)
       CONN=  - A byte[1] or a Data (CONN)
       POS=   - A Position, an Data (LON,LAT,ALT), or Table {ALT=,LAT=,LON=}
       POSP=  - A Position, an Data (LON,LAT,ALT), or Table {ALT=,LAT=,LON=}
       ELPS=  - A double[3] or Data (semiMajor,semiMinor,tilt)
       CONE=  - A double[3] or Data (radius,angle,angle)
       FLAG=  - A byte[8] or Data (NAMELEN,NAMEPOS,COLOR,BGCOLOR,CONE,
                  INVERT,SYMBOLID,SYMBOLSIZE) or Table {NAMELEN=,NAMEPOS=,
                  COLOR=,BGCOLOR=,CONE=,INVERT=,SYMBOLID=,SYMBOLSIZE=}
     </pre>
     @param index index of point to merge point properties
     @param table input table of point properties to merge with existing point, see above for expected key/values
     @since 3.5.1 for BUG 2823  (major update/fix 3.7.0 for BUG 2965)
   */
  public void mergePoint (int index, Table table) {
    // only merging with existing point if index in range of current points - since 3.7.0 bug 2965
    boolean mergingWithExistingPt = index < numPoints;

    if (numPoints == 0) {
      // initialize component fields if this is the first point for this LayerT5 - since 3.7.0 bug 2965
      updatePointsComponents(table);
    }
    else {
      // validate point components match existing points - since 3.7.0 for bug 2965
      if (numPoints > 0) validatePointComponents(table,mergingWithExistingPt);
    }

    if ((debugMask & 0x1) != 0) System.out.println("******** LayerT5.mergePoint********: index=" + index + ", #pts=" + numPoints);
    if (index < 0)
      index = numPoints;

    if (!mergingWithExistingPt) {
      Pnt point = new Pnt(table);
      while (index + 1 < numPoints) {
        addPoint(null, true);
      }
      addPoint(point, true);
      setSize(points.length);
    }
    else { // simplified, since updates to other methods use existing values as defaults
      if (points[index] == null) {
        points[index] = new Pnt(table);
      }
      else {
        points[index].fromTable(table);
      }
    }

    refresh();
  }

  /** Merge all points in new table with properties in the existing existing points.
      The input table can have the following fields:
     <pre>
        NAME=  - A String (any other object is ok, .toString() is used)
        NOTE=  - A String (any other object is ok, .toString() is used)
        CONN=  - A byte[1] or a Data (CONN)
        POS=   - A Position, an Data (LON,LAT,ALT), or Table {ALT=,LAT=,LON=}
        POSP=  - A Position, an Data (LON,LAT,ALT), or Table {ALT=,LAT=,LON=}
        ELPS=  - A double[3] or Data (semiMajor,semiMinor,tilt)
        CONE=  - A double[3] or Data (radius,angle,angle)
        FLAG=  - A byte[8] or Data (NAMELEN,NAMEPOS,COLOR,BGCOLOR,CONE,
                 INVERT,SYMBOLID,SYMBOLSIZE) or Table {NAMELEN=,NAMEPOS=,
                 COLOR=,BGCOLOR=,CONE=,INVERT=,SYMBOLID=,SYMBOLSIZE=}
     </pre>
     @param table input table of point properties to merge with existing point, see above for expected key/values
     @since 3.5.1 (major update/fix 3.7.0 for BUG 2965)
   */
  public void mergeAllPoints (Table table) {

    // validate point components match existing points (merging pt:true) - since 3.7.0 for bug 2965
    if (numPoints > 0) validatePointComponents(table,true);

    if ((debugMask & 0x1) != 0) System.out.println("******LayerT5.mergeAllPoints********: #pts=" + numPoints);

    for (int index = 0; index < numPoints; index++) {
      if (points[index] == null) {
        points[index] = new Pnt(table);
      }
      else {
        points[index].fromTable(table);
      }
    }

    refresh();
  }

  /**
   * Update component fields to allow for point components not specified in the file
   * (Points added via {@link #setPoint(int, Table)}, {@link #mergePoint(int, Table)},
   * or {@link #mergeAllPoints(Table)} may have additional point components)
   * Point components contained only in memory will have their "byte position"
   * set to -2 (-1: component not exist, >0 = byte position of component in file)
   * @param table table containing Point components
   * @since NeXtMidas3.7.0 for Bugs 2917 and 2965
   */
  private void updatePointsComponents (Table table) {
    if (bName == -1  && table.containsKey("NAME")) {bName = -2;}
    if (bPos  == -1  && table.containsKey("POS" )) {bPos  = -2;}
    if (bPosp == -1  && table.containsKey("POSP")) {bPosp = -2;} // position pointer
    if (bFlag == -1  && table.containsKey("FLAG")) {bFlag = -2;}
    if (bNote == -1  && table.containsKey("NOTE")) {bNote = -2;}
    if (bElps == -1  && table.containsKey("ELPS")) {bElps = -2;}
    if (bConn == -1  && table.containsKey("CONN")) {bConn = -2;} // connect flag
    if (bHdg  == -1  && table.containsKey("HDG" )) {bHdg  = -2;} // heading
  }

  /**
   * Validate Point Components
   * @param table table containing either info describing new point or point update info
   * @param isMerge true:information is being merged with an existing point, false:completely new point
   * @throws MidasException if this update would result in non-homogeneous points
   * @since NextMidas 3.7.0 for Bugs 2917 and 2965
   */
  private void validatePointComponents (Table table, boolean isMerge) {

    if (table == null) { return; } // null points are allowed

    StringBuilder additionalKeys = new StringBuilder();
    StringBuilder missingKeys = new StringBuilder();
    boolean ptHasName = table.containsKey("NAME");
    boolean ptHasPos  = table.containsKey("POS" );
    boolean ptHasPosp = table.containsKey("POSP" );
    boolean ptHasFlag = table.containsKey("FLAG" );
    boolean ptHasNote = table.containsKey("NOTE" );
    boolean ptHasElps = table.containsKey("ELPS" );
    boolean ptHasConn = table.containsKey("CONN" );
    boolean ptHasHdg  = table.containsKey("HDG" );

    // New/Updated point can not contain more Type 5000 Components than the existing points
    if (bName == -1 && ptHasName) { additionalKeys.append(" \"NAME\""); }
    if (bPos  == -1 && ptHasPos ) { additionalKeys.append(" \"POS\"");  }
    if (bPosp == -1 && ptHasPosp) { additionalKeys.append(" \"POSP\""); }
    if (bFlag == -1 && ptHasFlag) { additionalKeys.append(" \"FLAG\""); }
    if (bNote == -1 && ptHasNote) { additionalKeys.append(" \"NOTE\""); }
    if (bElps == -1 && ptHasElps) { additionalKeys.append(" \"ELPS\""); }
    if (bConn == -1 && ptHasConn) { additionalKeys.append(" \"CONN\""); }
    if (bHdg  == -1 && ptHasHdg ) { additionalKeys.append(" \"HDG\"");  }

    // If this point information is NOT being merged, it needs to contain all the
    // Type 5000 components contained in the existing points (e.g. check for missing Components)
    if (!isMerge) {
      if (bName != -1 && !ptHasName) { missingKeys.append(" \"NAME\""); }
      if (bPos  != -1 && !ptHasPos ) { missingKeys.append(" \"POS\"");  }
      if (bPosp != -1 && !ptHasPosp) { missingKeys.append(" \"POSP\""); }
      if (bFlag != -1 && !ptHasFlag) { missingKeys.append(" \"FLAG\""); }
      if (bNote != -1 && !ptHasNote) { missingKeys.append(" \"NOTE\""); }
      if (bElps != -1 && !ptHasElps) { missingKeys.append(" \"ELPS\""); }
      if (bConn != -1 && !ptHasConn) { missingKeys.append(" \"CONN\""); }
      if (bHdg  != -1 && !ptHasHdg ) { missingKeys.append(" \"HDG\"");  }
    }

    boolean invalidExtraComponents = additionalKeys.length() != 0;
    boolean missingComponents      = missingKeys.length()    != 0;
    if (invalidExtraComponents ||  missingComponents) {
      StringBuilder sb = new StringBuilder();
      sb.append("LayerT5: Invalid new/updated Point with table: ").append(table);
      sb.append("\n New/updated Point MUST match existing Point Components so that they are homogeneous.");

      if (invalidExtraComponents) {
        sb.append("\n - new/updated Point contains these invalid additional Type 5000 Components: ").append(additionalKeys);
      }
      if (missingComponents) {
        sb.append("\n - new/replaced Point is missing these Type 5000 Components: ").append(missingKeys);
      }
      throw new IllegalArgumentException(sb.toString());
    }

  }

  /** Removes a point.
      @param index The index of the point.
   */
  public synchronized void deletePoint (int index) {
    if ((index >= 0) && (index < numPoints)) {
      points[index] = null;
      nullPoints++;
      process();
      refresh();
    }
  }

  /** Removes a range of points.
      @param indexFirst index of the first point.
      @param indexLast  index of last point to remove (if -1 means last point)
      @since NeXtMidas 3.7.0
   */
  @InternalUseOnly
  public synchronized void deletePoints (int indexFirst, int indexLast) {
   if (indexLast == -1) {
     indexLast = numPoints -1;
   }
   if (indexLast < indexFirst) {
     warning("LayerT5.deletePoints indexLast:"+indexLast+" < indexFirst:"+indexFirst);
   }
   else if (indexFirst < 0 || indexLast >= numPoints) {
     if (indexFirst < 0) {
       warning("LayerT5.deletePoints indexFirst="+indexFirst+" < 0 setting value to 0");
       indexFirst = 0;
     }
     if (indexFirst < 0) {
       warning("LayerT5.deletePoints indexLast="+indexLast+
           " beyond point range, setting value to index of last point:"+(numPoints-1));
       indexLast = numPoints - 1;
     }
   }

   for (int index = indexFirst; index <= indexLast; index++) {
     points[index] = null;
     nullPoints++;
   }

   process();
   refresh();

  }

  /**
   * Make copy of existing point at specified index
   * @param indexOrig    index of original
   * @param indexForCopy index to contain copy (-1 - append to end)
   * @since NeXtMidas 3.7.0 for bug 2917
   */
  @InternalUseOnly
  public void copyPoint (int indexOrig,int indexForCopy) {
    if (indexOrig < 0 || indexOrig >= numPoints) {
      throw new MidasException("LayerT5.copyPoint invalid index of original:"+
                                indexOrig+" valid range 0 - "+(numPoints-1));
    }
    if (indexForCopy == -1) {
      // append copy to end
      addPoint(new Pnt(points[indexOrig].toTable()), true);
    }
    else if (indexForCopy >= 0 && indexForCopy < numPoints){
      // overwrite point at indexForCopy
      if (points[indexOrig] != null) {
        points[indexForCopy] = new Pnt(points[indexOrig].toTable());
      }
      else {
        points[indexForCopy] = null;
      }
    }
    else {
      // indexForCopy is beyond end so fill with null points and then append
      if (indexForCopy >= numPoints) {
        while (indexForCopy + 1 < numPoints) {
          addPoint(null, true);
        }
        addPoint(new Pnt(points[indexOrig].toTable()), true);
        setSize(points.length);
      }
    }
  }

  /** Adds a point to the array of points, resizing as needed, returns insertion index. */
  private int addPoint (Pnt pnt) {
    return addPoint(pnt, false);
  }

  /** Adds a point to the array of points, resizing as needed, returns insertion index. */
  private synchronized int addPoint (Pnt pnt, boolean addToEnd) {
    int index = -1;
    if ((points == null) || (points.length == 0)) {
      if (size < 1) {
        warning("LayerT5: Got size < 1");
        size = 1;
      }
      points = new Pnt[size];
      numPoints = 1;
      index = 0;
      firstPoint = (firstPoint == -1) ? -1 : 0;
      points[index] = pnt;
      return index;
    }
    else if (firstPoint == -1) { // variable size
      if (!addToEnd && (nullPoints > 0)) { // Reuse one of the existing slots that is null
        for (int i = 0; (i < numPoints) && (index == -1); i++) {
          if (points[i] == null) {
            index = i;
            nullPoints--;
          }
        }
      }
      else {
        if (numPoints + 1 >= points.length) { // Grow the array

          Pnt[] array = new Pnt[points.length * 2];
          System.arraycopy(points, 0, array, 0, points.length);
          points = array;
        }
        index = numPoints;
        numPoints++;
      }
    }
    else if (numPoints == points.length) { // fixed size, drop oldest
      index = firstPoint;
      firstPoint = (firstPoint + 1) % points.length; // wrap around as needed
      if (points[index] == null)
        nullPoints--;
    }
    else { // fixed size, still has room
      index = numPoints;
      numPoints++;
    }
    points[index] = pnt;
    return index;
  }

  // ////////////////////////////////////////////////////////////////////////////////////////////////
  // Misc Methods
  // ////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the flags for plotting and refreshes the plot.
   * @param value The new plotting flags (see {@link #flagsList}).
   */
  public void setFlags(int value) {
    if (flags != value) {
      flags = value;
      refresh();
    }
  }

  /**
   * Sets the flags for plotting and refreshes the plot.
   * @param value The new plotting flags (see {@link #flagsList}).
   */
  public void setFlags(String value) {
    setFlags(Parser.mask(flagsList, value, flags));
  }

  /** Gets the flags for plotting (see {@link #flagsList}). */
  public String getFlags() {
    return Parser.mask2s(flagsList, flags);
  }

  /**
   * Outputs a warning message.
   * @param msg The message.
   */
  private void warn(String msg) {
    warning("LayerT5: " + msg);
  }

  /**
   * Indicates if a given field should be shown.
   * @param flag The flag for the field (e.g. {@link #LABELS}).
   * @return true if shown, flase if not.
   */
  private boolean showField(int flag) {
    return (field & flag) != 0;
  }

  /** issuer configuration menus */
  @Override
  public void configure() {
    new GMenu(MP, "Layer.cfg", Layer.configureList + ",AutoSelect", 0, 0, this);
  }

  @Override
  public int processMessage(Message msg) {
    if (msg.name.equals("LAYER.CFG") && ((String) msg.data).equals("AUTOSELECT")) {
      new GMenu(MP, "AutoSelect.Options", "Enable,SelectionMode,SelectionRange", 0, 0, this);
    }
    else if (msg.name.equals("AUTOSELECT.OPTIONS")) {
      String text = (String) msg.data;
      if (text.equals("ENABLE")) {
        new GMenu(MP, "AutoSelect.Enable", plot.pointerMessageList, autoSelEnable, GMenu.TOGGLE,
            this);
      }
      else if (text.equals("SELECTIONMODE")) {
        new GMenu(MP, "AutoSelect.SelectionMode", selectionModes, autoSelMode, 0, this);
      }
      else if (text.equals("SELECTIONRANGE")) {
        new GPrompt(MP, "AutoSelect.SelectionRange", Double.toString(autoSelRange), 0, this);
      }
    }
    else if (msg.name.equals("AUTOSELECT.ENABLE")) {
      autoSelEnable = msg.info;
    }
    else if (msg.name.equals("AUTOSELECT.SELECTIONMODE")) {
      autoSelMode = msg.info;
    }
    else if (msg.name.equals("AUTOSELECT.SELECTIONRANGE")) {
      autoSelRange = Convert.o2d(msg.data);
    }
    else {
      return super.processMessage(msg);
    }
    return 1;
  }

  /**
   * Sets the AutoSelect enable flags.
   * @param enable The flags (see {@link plot#pointerMessageList}).
   * @since NeXtMidas 2.1.0
   */
  public void setAutoSelectEnable(String enable) {
    autoSelEnable = Parser.mask(plot.pointerMessageList, enable, autoSelEnable);
  }

  /**
   * Gets the AutoSelect enable flags (see {@link plot#pointerMessageList}).
   * @since NeXtMidas 2.1.0
   */
  public String getAutoSelectEnable() {
    return Parser.mask2s(plot.pointerMessageList, autoSelEnable);
  }

  /**
   * Sets the AutoSelect mode.
   * @param mode The mode (see {@link #selectionModes}).
   * @since NeXtMidas 2.1.0
   */
  public void setAutoSelectMode(String mode) {
    autoSelMode = Parser.find(selectionModes, mode, autoSelMode, 0, Parser.E_WARNING);
  }

  /**
   * Gets the AutoSelect mode (see {@link #selectionModes}).
   * @since NeXtMidas 2.1.0
   */
  public String getAutoSelectMode() {
    return Parser.get(selectionModes, autoSelMode);
  }

  /**
   * Sets the AutoSelect range.
   * @param range The range (units are determined based on AutoSelect mode).
   * @since NeXtMidas 2.1.0
   */
  public void setAutoSelectRange(double range) {
    autoSelRange = range;
  }

  /**
   * Gets the AutoSelect range (units are determined based on AutoSelect mode).
   * @since NeXtMidas 2.1.0
   */
  public double getAutoSelectRange() {
    return autoSelRange;
  }

  /**
   * Processes mouse movement messages. If the message type matches the autoselect mode then a
   * selection is made as a result of the message.
   * @param msg {@inheritDoc}
   * @param type {@inheritDoc}
   * @since NeXtMidas 2.1.0
   */
  @Override
  public void processMouseMessage(Message msg, int type) {
    if ((autoSelEnable & type) != 0) {
      Position pos = (Position) msg.getData();
      selectNearest(autoSelMode, pos, autoSelRange);
    }
  }

  // ////////////////////////////////////////////////////////////////////////////
  // A given point
  // ////////////////////////////////////////////////////////////////////////////
  /**
   * A point in the layer. This is named Pnt to avoid conflicts with either java.awt.Point or
   * nxm.sys.lib.MPoint.
   */
  private class Pnt {
    private String name = ""; // The name of the point

    private String note = ""; // The notes for the point

    private boolean posSet = false; // Has POS been set?

    private double posAlt = 0.0; // The POS position (alt,lat,lon)

    private double posLat = 0.0; // The POS position (alt,lat,lon)

    private double posLon = 0.0; // The POS position (alt,lat,lon)

    private boolean pospSet = false; // Has POSP been set?

    private double pospAlt = 0.0; // The POSP position (alt,lat,lon)

    private double pospLat = 0.0; // The POSP position (alt,lat,lon)

    private double pospLon = 0.0; // The POSP position (alt,lat,lon)

    private boolean elpsSet = false; // Has ellipse been set?

    private double elpsSemiMajor = 0.0; // The ellipse parameters (major,minor,tilt)

    private double elpsSemiMinor = 0.0; // The ellipse parameters (major,minor,tilt)

    private double elpsTilt = 0.0; // The ellipse parameters (major,minor,tilt)

    private boolean coneSet = false; // Has cone been set?

    private double coneA = 0.0; // The cone parameters (radius,angle,angle)

    private double coneB = 0.0; // The cone parameters (radius,angle,angle)

    private double coneAngle = 0.0; // The cone parameters (radius,angle,angle)

    private boolean conn = false; // The connect flags setting the drawing mode

    private float hdg = 0.0f; // Symbol Heading

    private byte flagNameLen = 0; // Flag 0: Name length in characters)

    private byte flagNamePos = 0; // Flag 1: Name position (not supported)

    private byte flagFgColor = 0; // Flag 2: foreground color (X-Midas color index number)

    private byte flagBgColor = 0; // Flag 3: background color (not supported)

    private byte flagCone = 0; // Flag 4: Cone/ellipse flag (0=hidden, -9=visible)

    private byte flagInvert = 0; // Flag 5: Invert (not supported)

    private byte flagSymCode = 0; // Flag 6: Symbol (X-Midas symbol ID number)

    private byte flagSymSize = 0; // Flag 7: Symbol size

    private int _ixPos = 0; // Only used by draw() - Pixel X for pos

    private int _iyPos = 0; // Only used by draw() - Pixel Y for pos

    private int _ixPosp = 0; // Only used by draw() - Pixel X for posp

    private int _iyPosp = 0; // Only used by draw() - Pixel Y for posp

    private double[] coords = null; // Only used by prep(..)

    private double minLon = 0.0; // Internal use only: max longitude used in hit(..)

    private double maxLon = 0.0; // Internal use only: max longitude used in hit(..)

    private double minLat = 0.0; // Internal use only: min latitude used in hit(..)

    private double maxLat = 0.0; // Internal use only: min latitude used in hit(..)

    private double minAlt = 0.0; // Internal use only: min altitude used in hit(..)

    private double maxAlt = 0.0; // Internal use only: min altitude used in hit(..)

    private boolean hidden;

    /** Creates a new instance. */
    public Pnt() {
      coords = new double[6];
    }

    /** Creates a new instance. */
    Pnt(Table tbl) {
      this();
      this.fromTable(tbl);
    }

    /** Prepare for use inside draw() -- only used by draw() method. */
    void prep(boolean doPosp) {
      int k = 0;
      if (doPosp) {
        k = MP.viewer.rwc2pix(coords, 0, 3, 2, pix);
        _ixPosp = pix.x[1];
        _iyPosp = pix.y[1];
      }
      else {
        k = MP.viewer.rwc2pix(coords, 0, 3, 1, pix);
      }
      hidden = (MP.view == MPlot.V_GEODETIC) && k < 1;
      _ixPos = pix.x[0];
      _iyPos = pix.y[0];
    }

    /** Done using in draw() -- only used by draw() method. */
    void done() {
      // may need to do something here in the future
    }

    /** Is this point visible -- only used by draw() method. */
    boolean hit(double lon1, double lon2, double lat1, double lat2) {
      if (coneSet)
        return true; // JC4: Not sure how to optimize this one

      if ((lat2 < minLat) || (lat1 > maxLat))
        return false; // Out of range (Lat)
      if ((lon2 < minLon) || (lon1 > maxLon))
        return false; // Out of range (Lon)

      return true;
    }

    /** Updates the min/max values for the point. */
    private void updateMinMax() {
      minLon = posLon;
      maxLon = posLon;
      minLat = posLat;
      maxLat = posLat;
      minAlt = posAlt;
      maxAlt = posAlt;
      if (pospSet) {
        minLon = Math.min(minLon, pospLon);
        maxLon = Math.max(maxLon, pospLon);
        minLat = Math.min(minLat, pospLat);
        maxLat = Math.max(maxLat, pospLat);
        minAlt = Math.min(minAlt, pospAlt);
        maxAlt = Math.max(maxAlt, pospAlt);
      }
      if (elpsSet) {
        double rad = Math.max(elpsSemiMajor, elpsSemiMinor);
        double radLat = rad * METERS_TO_DEG_LAT * 2.0; // JC4: Add factor of 2 since this is an est.
        double radLon = rad * METERS_TO_DEG_LON * 2.0; // JC4: Add factor of 2 since this is an est.
        minLon = Math.min(minLon, posLon - radLon);
        maxLon = Math.max(maxLon, posLon + radLon);
        minLat = Math.min(minLat, posLat - radLat);
        maxLat = Math.max(maxLat, posLat + radLat);
      }
      updateRange(this);
    }

    /** Gets the pixel position of pos. */
    private Pix getPosPix(Pix p) {
      double[] pos = new double[] { posLon, posLat, posAlt };
      MP.viewer.rwc2pix(pos, 0, 3, 1, p);
      return p;
    }

    /** Gets the pixel position of pos. */
    private Pix getPospPix(Pix p) {
      double[] posp = new double[] { pospLon, pospLat, pospAlt };
      MP.viewer.rwc2pix(posp, 0, 3, 1, p);
      return p;
    }

    /** Sets the name of the point. */
    public void setName(String name) {
      this.name = name;
    }

    /** Gets the name of the point. */
    public String getName() {
      return name;
    }

    /** Sets the name of the point. */
    public void setNote(String note) {
      this.note = note;
    }

    /** Gets the name of the point. */
    public String getNote() {
      return note;
    }

    /** Gets the connection flag that indicates if this point is connected to the previous one. */
    boolean getConn() {
      return conn;
    }

    /** Sets the connection flag to indicate if this point is connected to the previous one. */
    void setConn(boolean conn) {
      this.conn = conn;
    }

    /** Sets the connection flag to indicate if this point is connected to the previous one. */
    void setConn(byte conn) {
      this.conn = (conn == 1);
    }

    /** Gets the heading of the symbol. */
    float getHdg() {
      return hdg;
    }

    /** Sets the heading of the symbol. */
    void setHdg(float hdg) {
      this.hdg = hdg;
    }

    /** Sets the connection flag to indicate if this point is connected to the previous one. */
    private void setConn(Object obj) {
      if (obj instanceof Data)
        obj = toArray(obj, new byte[1]);
      if (obj instanceof byte[])
        setConn(((byte[]) obj)[0] != 0);
      else if (obj != null)
        setConn(Convert.s2state(obj.toString(), conn));
    }

    /**
     * Sets the position of the point.
     * @param alt The altitude.
     * @param lat The latitude (in deg).
     * @param lon The longitude (in deg).
     */
    public void setPos(double alt, double lat, double lon) {
      coords[0] = posLon = lon;
      coords[1] = posLat = lat;
      coords[2] = posAlt = alt;
      posSet = true;
      updateMinMax();
    }

    /** Internal use only */
    private void setPos(Object obj) {
      if (obj instanceof Table)
        setPos((Table) obj);
      if (obj instanceof Position)
        setPos((Position) obj);
      if (obj instanceof double[])
        setPos((double[]) obj);
    }

    /**
     * Sets the position of the point with a table.
     * @param t Pointing position table <tt>{LON=lon,LAT=lat,ALT=alt}</tt>
     */
    public void setPos(Table t) {
      setPos(t.getD("ALT", posAlt), t.getD("LAT",posLat), t.getD("LON",posLon)); // default to exiting values - since 3.7.0 Bug 2965
    }

    /**
     * Sets the position of the point with a Position.
     * @param p Position in {@link Position#GEODETIC} mode.
     */
    public void setPos(Position p) {
      setPos(p.getAlt(), p.getLat(), p.getLon());
    }

    /**
     * Sets the position of the point with an array (lon,lat,alt).
     * @param a The pointing position <tt>{ lon, lat, alt }</tt>
     */
    public void setPos(double[] a) {
      setPos(a[2], a[1], a[0]);
    }

    /** Gets the position. */
    public Position getPos() {
      return Position.fromGeo(posAlt, posLat, posLon);
    }

    /**
     * Sets the pointing position of the point.
     * @param alt The altitude.
     * @param lat The latitude (in deg).
     * @param lon The longitude (in deg).
     */
    public void setPosp(double alt, double lat, double lon) {
      coords[3] = pospLon = lon;
      coords[4] = pospLat = lat;
      coords[5] = pospAlt = alt;
      pospSet = (bVec == -1);
      updateMinMax();
    }

    /** Internal use only */
    private void setPosp(Object obj) {
      if (obj instanceof Table)
        setPosp((Table) obj);
      if (obj instanceof Position)
        setPosp((Position) obj);
      if (obj instanceof double[])
        setPosp((double[]) obj);
    }

    /**
     * Sets the pointing position of the point with a table.
     * @param t Pointing position table <tt>{LON=lon,LAT=lat,ALT=alt}</tt>
     */
    public void setPosp(Table t) {
      setPosp(t.getD("ALT", pospAlt), t.getD("LAT", pospLat), t.getD("LON", pospLon)); // default to exiting values - since 3.7.0 Bug 2965
    }

    /**
     * Sets the pointing position of the point with a Position.
     * @param p Position in {@link Position#GEODETIC} mode.
     */
    public void setPosp(Position p) {
      setPosp(p.getAlt(), p.getLat(), p.getLon());
    }

    /**
     * Sets the pointing position of the point with an array (lon,lat,alt).
     * @param a The pointing position <tt>{ lon, lat, alt }</tt>
     */
    public void setPosp(double[] a) {
      setPosp(a[2], a[1], a[0]);
    }

    /** Gets the pointing position. */
    public Position getPosp() {
      return Position.fromGeo(pospAlt, pospLat, pospLon);
    }

    /** Internal use only */
    private void setFlags(Object flags) {
      if (flags instanceof Data)
        flags = toArray(flags, new byte[8]);
      if (flags instanceof byte[])
        setFlags((byte[]) flags);
      if (flags instanceof Table)
        setFlags((Table) flags, (byte) name.length());
    }

    /** Internal use only */
    private void setFlags(byte[] flags) {
      flagNameLen = flags[0];
      flagNamePos = flags[1];
      flagFgColor = flags[2];
      flagBgColor = flags[3];
      flagCone = flags[4];
      flagInvert = flags[5];
      flagSymCode = flags[6];
      flagSymSize = flags[7];
    }

    /** Sets the name length flag. */
    void setFlagNameLen(byte val) {
      flagNameLen = val;
    }

    /** Sets the name position flag. */
    void setFlagNamePos(byte val) {
      flagNamePos = val;
    }

    /** Sets the foreground color flag. */
    void setFlagFgColor(byte val) {
      flagFgColor = val;
    }

    /** Sets the background color flag. */
    void setFlagBgColor(byte val) {
      flagBgColor = val;
    }

    /** Sets the cone/ellipse flag. */
    void setFlagCone(byte val) {
      flagCone = val;
    }

    /** Sets the invert flag. */
    void setFlagInvert(byte val) {
      flagInvert = val;
    }

    /** Sets the symbol code flag. */
    void setFlagSymCode(byte val) {
      flagSymCode = val;
    }

    /** Sets the symbol size flag. */
    void setFlagSymSize(byte val) {
      flagSymSize = val;
    }

    /** Gets the name length flag. */
    byte getFlagNameLen() {
      return flagNameLen;
    }

    /** Gets the name position flag. */
    byte getFlagNamePos() {
      return flagNamePos;
    }

    /** Gets the foreground color flag. */
    byte getFlagFgColor() {
      return flagFgColor;
    }

    /** Gets the background color flag. */
    byte getFlagBgColor() {
      return flagBgColor;
    }

    /** Gets the cone/ellipse flag. */
    byte getFlagCone() {
      return flagCone;
    }

    /** Gets the invert flag. */
    byte getFlagInvert() {
      return flagInvert;
    }

    /** Gets the symbol code flag. */
    byte getFlagSymCode() {
      return flagSymCode;
    }

    /** Gets the symbol size flag. */
    byte getFlagSymSize() {
      return flagSymSize;
    }

    /** Sets the flags via a table. */
    private void setFlags(Table t) {
      setFlags(t, flagNameLen);
    }

    /** Internal use only. */
    private void setFlags(Table tab, byte nameLen) {
      flagNameLen = tab.getB("NAMELEN", nameLen);
      flagNamePos = tab.getB("NAMEPOS", flagNamePos);
      flagFgColor = tab.getB("COLOR", flagFgColor);
      flagBgColor = tab.getB("BGCOLOR", flagBgColor);
      flagCone    = tab.getB("CONE", flagCone);
      flagInvert  = tab.getB("INVERT", flagInvert);
      flagSymCode = tab.getB("SYMBOLID", flagSymCode);
      flagSymSize = tab.getB("SYMBOLSIZE", flagSymSize);
    }

    /** Gets the flag as a data object. */
    private Data getFlagsData() {
      byte[] array = { flagNameLen, flagNamePos, flagFgColor, flagBgColor, flagCone, flagInvert,
          flagSymCode, flagSymSize };
      return new Data(array);
    }

    /** Converts a Data object to an array. */
    private Object toArray(Object obj, Object array) {
      Data data = (Data) obj;
      data.toArray(array);
      return array;
    }

    /**
     * Sets the ellipse parameters with a table.
     * @param t ellipse parameters table <tt>{SEMIMAJOR=elpsSemeiMajor,SEMIMINOR=elpsSemiMinor,TILT=elpsTilt}</tt>
     * @since NeXtMidas 3.7.0 for Bug 2965
     */
    private void setElps(Table t) {
      setElps(t.getD("SEMIMAJOR", elpsSemiMajor), t.getD("SEMIMINOR", elpsSemiMinor), t.getD("TILT",elpsTilt)); // default to exiting values - since 3.7.0 Bug 2965
    }

    /**
     * Sets the ellipse parameters from an array (a,b,c).
     * @param semiMajor The semi-major axis (in meters).
     * @param semiMinor The semi-minor axis (in meters).
     * @param tilt The tilt (in deg).
     */
    public void setElps(double semiMajor, double semiMinor, double tilt) {
      elpsSemiMajor = semiMajor;
      elpsSemiMinor = semiMinor;
      elpsTilt = tilt;
      elpsSet = true;
      updateMinMax();
    }

    /**
     * Sets the ellipse parameters from an array (a,b,c).
     * @param a The ellipse parameters <tt>{ semiMajor, semiMinor, tilt }</tt>.
     */
    public void setElps(double[] a) {
      setElps(a[0], a[1], a[2]);
    }

    /** Internal use only. */
    private void setElps(Object obj) {
      if (obj instanceof Table) // Since 3.7.0 for bug 2965
        setElps((Table) obj);
      if (obj instanceof Data)
        obj = toArray(obj, new double[3]);
      if (obj instanceof double[])
        setElps((double[]) obj);
    }

    /**
     * Sets the cone parameters
     * @param a (saved, but not used at this time)
     * @param b (saved, but not used at this time)
     * @param angle Cone angle.
     */
    public void setCone(double a, double b, double angle) {
      coneA = a;
      coneB = b;
      coneAngle = angle;
      coneSet = true;
      updateMinMax();
    }

    /** Gets the elps as a data object. */
    private Data getElpsData() {
      double[] array = { elpsSemiMajor, elpsSemiMinor, elpsTilt };
      return new Data(array);
    }

    /**
     * Sets the cone parameters from an array (a,b,c).
     * @param a The cone parameters <tt>{ a, b, angle }</tt>.
     */
    public void setCone(double[] a) {
      setCone(a[0], a[1], a[2]);
    }

    /** Internal use only. */
    private void setCone(Object obj) {
      if (obj instanceof Data)
        obj = toArray(obj, new double[3]);
      if (obj instanceof double[])
        setCone((double[]) obj);
    }

    private Data getConnData() {
      return (conn) ? new Data(1) : new Data(0);
    }

    /** Gets the cone as a data object. */
    private Data getConeData() {
      double[] array = { coneA, coneB, coneAngle };
      return new Data(array);
    }

    /** Returns whether this point is hidden by another object. */
    public boolean isHidden() {
      return hidden;
    }

    /** Gets the points values based as a table. */
    Table toTable() {
      Table table = new Table();

      table.put("NAME", getName());
      table.put("NOTE", getNote());
      table.put("CONN", getConnData());
      table.put("POS", getPos());
      table.put("POSP", getPosp());
      table.put("ELPS", getElpsData());
      table.put("CONE", getConeData());
      table.put("FLAG", getFlagsData());

      return table;
    }

    /** Sets the points values based on a table. */
    void fromTable(Table tbl) {
      setName(tbl.getS("NAME", name));
      setNote(tbl.getS("NOTE", note));

      setConn(tbl.getKey("CONN"));
      setPos(tbl.getKey("POS"));
      setPosp(tbl.getKey("POSP"));
      setFlags(tbl.getKey("FLAG"));
      setElps(tbl.getKey("ELPS"));
      setCone(tbl.getKey("CONE"));
    }

    @Override
    public String toString() {
      return toTable().toString();
    }
  }
}
