package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.Arrays;
import java.util.Map;

/**
  A class to run a Midas panel in a browser Applet.
  <p>
  Since NeXtMidas 3.3.0, it is possible to interface from Java/Macro to the web
  page by calling or evaluating JavaScript on page where this Applet is embedded on.
  See {@link #callJavaScript(String, Object...)}, {@link #setJScript(String, Table)},
  and/or {@link #evalJavaScript(String)}.
  <a name="javascriptIntefaceNote"></a>
  This requires web browsers to use the new Java plug-in from Java 6u10 and later.
  See <a href="http://jdk6.java.net/plugin2/liveconnect/">LiveConnect Specification</a> for more
  details on how it allows Java applets to communicate with the JavaScript engine in the browser,
  and JavaScript on the web page to interact with applets.
  </p>

  @author Jeff Schoen / Neon Ngo
  @version $Id: MApplet.java,v 1.46 2015/08/27 15:54:21 ntn Exp $
*/
public class MApplet extends Applet
  implements ComponentListener,MessageHandler,IDable,Keyable,ShellParent {
  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2012020930311L; // for Serializable interface

  /** @since NeXtMidas 3.3.0 */
  private static final String[][] PARAMETER_INFO = {
    // NAME              TYPE          DESCRIPTION
    { "AUX",             "String",     "NeXtMidas AUX entries in the form '<READ> <WRITE1>|<WRITE2>|...'" },
    { "CMD",             "String",     "NeXtMidas command to run" },
    { "DEBUG",           "Boolean",    "Turns on/off MApplet debugging" },
    { "ENV.THEME",       "String",     "Theme to use" },
    { "ENV.STYLE",       "String",     "Style to use" },
    { "HOME",            "String",     "HOME path" },
    { "NMROOT",          "String",     "NeXtMidas NMROOT" },
    { "NMS",             "String",     "NeXtMidas startup script (**DEPRECATED**)" },
    { "NMSTARTUP.USER",  "String",     "NeXtMidas startup script (overrides NMS)" },
    { "PATH",            "String",     "NeXtMidas option tree path" },
    { "OPTS",            "String",     "NeXtMidas option trees defined" },
    { "WIN",             "String",     "Window/registry name for applet" },

    // Standard Applet entries from, see
    // http://docs.oracle.com/javase/6/docs/technotes/guides/plugin/developer_guide/special_attributes.html
    { "boxmessage",      "String",     "Message to display in the applet box"},
    { "boxbgcolor",      "Color",      "Background color for applet box"},
    { "boxfgcolor",      "Color",      "Foreground color for applet box"},
    { "codebase_lookup", "Boolean",    "Look up the codebase? (This should be set to false if using .jar files.)"},

    // JNLPAppletLauncher parameters (see https://jogamp.org/applet-launcher/www/)
    { "jnlpNumExtensions",     "Integer",    "Number of extensions"},
    { "jnlpExtension1",        "String",     "JNLP extension #1"},
    { "jnlpExtension2",        "String",     "JNLP extension #2"},
    { "jnlpExtension3",        "String",     "JNLP extension #3"},
    { "jnlpExtension4",        "String",     "JNLP extension #4"},
    { "jnlpExtension5",        "String",     "JNLP extension #5"},
    { "jnlpExtension6",        "String",     "JNLP extension #6"},
    { "jnlpExtension7",        "String",     "JNLP extension #7"},
    { "jnlpExtension8",        "String",     "JNLP extension #8"},
    { "jnlpExtension9",        "String",     "JNLP extension #9"},
      // There could be more jnlpExtensionN entries...
    { "noddraw.check",         "Boolean",    "Check to see if DriectDraw is disabled on Windows"},
    { "noddraw.check.silent",  "Boolean",    "Don't show user dialogue box if DriectDraw check fails"},
    { "progressbar",           "Boolean",    "Show the progress bar?"},
    { "subapplet.classname",   "String",     "Class name of applet"},
    { "subapplet.displayname", "String",     "Display name for applet"},
  };

  private final Registry     registry = new Registry();
  private       String       registryName;
  private       MessageQueue queue;
  private       MWindow      MW;
  private       Shell        shell;
  private       boolean      shown   = false;
  private       boolean      conduit = false;
  private       boolean      debug   = false; // since 3.3.0
  private final boolean      appletLauncher;

  /** netscape.javascript.JSObject (JavaScript Object) - reference to web page's window,
      used for interacting from Java/Macro to web page via JavaScript callback(s).
   */
  private Object jswin; // since 3.3.0
  /** netscape.javascript.JSObject.call(java.lang.String,java.lang.Object[]) Method to invoke via reflection */
  Method jsoCallMethod; // since 3.3.0
  /** netscape.javascript.JSObject.eval(java.lang.String) Method to invoke via reflection */
  Method jsoEvalMethod; // since 3.3.0
  /** The Terminal to use. */
  private Terminal terminal; // since 3.3.0

  // this static b/c Shell.java:1208 need to call Applet.getProperty()
  // FIXME: there are better and safer ways to do this...
  private static MApplet currentApplet = null; // The currently running applet.

  public MApplet () {
    super();
    currentApplet  = this;
    String isJnlpAppletLauncher = null;
    try {
      isJnlpAppletLauncher = System.getProperty("sun.jnlp.applet.launcher", "false");
    } catch (SecurityException se) {
      System.err.println("WARN: cannot determine if using JNLPAppletLauncher! exception=" + se + " ["+getName()+"]");
    }
    appletLauncher = Convert.s2z(isJnlpAppletLauncher);
  }

  /** <b>Internal use only:</b> Was the applet launched from JNLPAppletLauncher?
      <i>This should only be called internal to the {@link Shell}.</i>
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  public boolean isJNLPAppletLauncher () {
    return appletLauncher;
  }

  /** <b>Internal use only:</b> Loads a library via the JNLPAppletLauncher (returns
      false if the applet was not launched with the JNLPAppletLauncher). <i>This
      should only be called internal to {@link Shell#loadLibraryUnsafe}.</i>
      @param libName Name of the library to load.
      @return true if loaded, false otherwise.
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  public boolean loadLibrary (String libName) {
    if (!isJNLPAppletLauncher()) return false;

    try {
      Class<?>  clazz       = Class.forName("org.jdesktop.applet.util.JNLPAppletLauncher");
      Method    loadLibrary = clazz.getDeclaredMethod("loadLibrary", String.class);

      loadLibrary.invoke(null, libName);
      return true;
    }
    catch (ClassNotFoundException e) {
      Shell.printStackTrace("loadLibrary(" + libName + ")", e);
      Shell.warning("Attempting to use System.loadLibrary instead");
      System.loadLibrary(libName);
      return true;
    }
    catch (Throwable t) {
      if (t instanceof InvocationTargetException) {
        t = ((InvocationTargetException)t).getTargetException();
      }

      if (t instanceof Error           ) throw (Error)t;
      if (t instanceof RuntimeException) throw (RuntimeException)t;

      // Throw UnsatisfiedLinkError for best compatibility with System.loadLibrary()
      throw (UnsatisfiedLinkError)new UnsatisfiedLinkError().initCause(t);
    }
  }

  /** @since NeXtMidas 3.3.0 */
  @Override
  public String[][] getParameterInfo() {
    return PARAMETER_INFO;
  }

  @Override
  public void init () {
    String command = getParameter("CMD");
    String regName = getParameter("WIN");
    if (regName == null) {
      regName = "MApplet";
    }
    init0(regName, command, this);
  }

  public void init (String win, String command, Object parent) {
    init0(win, command, (ShellParent)parent);
  }

  private void init0 (String win, String command, ShellParent parent) {
    super.init();
    addComponentListener(this);
    setLayout(null);

    registryName = win; // <-- must set this before using debug(..)
    debug        = Convert.s2z(getParameter("DEBUG")); // since 3.3.0

    if (debug) {
      debug("MApplet.init() "+this);
      debug("  System.getProperties() =>");
      try {
        for (Map.Entry<?,?> e : System.getProperties().entrySet()) {
          String key = e.getKey().toString();
          Object val = e.getValue();
          debug("    "+StringUtil.padRight(key, 32)+" = "+val);
        }
      }
      catch (Exception e) {
        debug("    Unable to list properties "+e);
      }
      debug("  MApplet.getParameter(..) => [only most common parameters shown]");
      for (String[] nameTypeDescr : getParameterInfo()) {
        String key = nameTypeDescr[0];
        String val = getParameter(key);
        debug("    "+StringUtil.padRight(key, 32)+" = "+val);
      }
    }

    registry.put(registryName, this);
    initJavaScriptIF();     // since 3.3.0: initialize JavaScript interface
    if (terminal == null) { // since 3.3.0: use Java Terminal in Applet
      terminal = new Terminal();
    }
    if (command != null) {
      debug("  Shell.runNew(..)");
      debug("    parent  = "+parent);
      debug("    command = "+command);
      debug("    bg      = "+true);
      shell = Shell.runNew(parent, command, true);
    }
  }

  @Override
  public void start () {
    super.start();
    debug("MApplet.start() "+this);
    shown = true;
    process("WINDOW",1,"OPENED");
  }

  @Override
  public void update (Graphics g) {
    paint(g);
  }

  @Override
  public void paint (Graphics g) {
    Rectangle rect=null;
    try { rect = g.getClipBounds(); } catch (Exception e) {}
    if (rect!=null && (rect.width<=1 || rect.height<=1)) return;
    if (MW != null) {
      MW.process("PAINT",0,g);
    } else {
      g.drawString(registryName,10,20);
    }
    super.paint(g);
  }

  @Override
  public void stop () {
    debug("MApplet.stop() "+this);
    shown = false;
    process ("WINDOW",-1,"CLOSED");
    super.stop();
  }

  public void dispose () {
    MW = null;
    debug("MApplet.dispose() "+this);
  }

  @Override
  public void destroy () {
    debug("MApplet.destroy() "+this);
    if (shell!=null && !conduit) shell.exit();
    super.destroy();
  }

  //////////////////////////////////////////////////////////////////////////////
  // Begin implement ShellParent interface
  //////////////////////////////////////////////////////////////////////////////

  /** Return the value of NMRoot for this applet. Applets can be deployed in
      two contexts, one where everything is stored on the server and one where
      the files are stored locally. If the NMROOT parameter is specified then
      it must point to a jar file that is stored locally or on the server
      where the applet's is running from.

      @return The path that should be used for NMROOT
  */
  @Override
  public String getNmRoot() {
    String nmroot = getParameter("NMROOT");
    if (nmroot == null || nmroot.trim().isEmpty()) {
      nmroot = OptionTree.findOptRoot("SYS");
    }
    // fall back to the codebase() as NMROOT if above two scenarios fails
    if (nmroot == null) {
      URL codeBaseURL = getCodeBase();
      if (codeBaseURL != null) { // since 3.4.0 for BUG 2791
        nmroot = codeBaseURL.toExternalForm();
      }
    }
    return nmroot;
  }

  /** Return the value of the AUX path as set by the applet. This is
      determined by the value of the AUX applet parameter.  If this parameter
      is not set then this method will return a default string.

      @return The arguments to the AUX command for configuring the aux path.
  */
  @Override
  public String getAux () {
    final String defAux = "RAM RAM|DAT";
    String aux = getParameter("AUX");
    if (aux == null) aux = defAux;
    return aux;
  }

  /** Return the option trees that are to be loaded by the Shell on startup.
      The list of these is taken from the OPTS parameter in the applet tag.
      The String is a list of comma separated option tree names.
      @return A comma separated list of option trees to configure on startup
              or null if there are none.
    */
  @Override
  public String getOpts () {
    String opts = getParameter("OPTS");
    return OptionTree.getOptsHelper(opts, this, true);
  }

  @Override
  public String getHome () { return getParameter("HOME"); }

  @Override
  public String getNMS () {
    String startupmacro = getParameter("NMSTARTUP.USER"); // since 3.1.3: this takes precedence over "NMS"
    if (startupmacro == null) startupmacro = getParameter("NMS");
    return startupmacro;
  }

  @Override
  public String getCWD () {
    URL docBase = getDocumentBase();
    if (docBase == null) return null;  // added in 3.2.2 / 3.4.0 to prevent NPEs
    String cwd = docBase.toString();
    int i=cwd.indexOf('?'); if (i>0) cwd = cwd.substring(0,i);
    i=cwd.lastIndexOf('/'); if (i>0) cwd = cwd.substring(0,i+1);
    return cwd;
  }

  @Override
  public String getServer () {
    // since 3.3.4: use docBase.host to fix Bug 2761
    URL docBase = getDocumentBase();
    if (docBase == null) return null;  // added in 3.2.2 / 3.4.0 to prevent NPEs
    String srv = docBase.getHost();
    if (srv.isEmpty()) srv = null; // null for localhost
    return srv;
  }

  /** Gets theme based on "ENV.THEME" applet parameter, otherwise returns default of "BROWSER".
      @since NeXtMidas 2.9.3
   */
  @Override
  public String getTheme () {
    String theme = getParameter("ENV.THEME");
    if (theme == null) theme = "BROWSER";
    return theme;
  }

  /** Gets style based on "ENV.STYLE" applet parameter, otherwise returns default of "BROWSER".
      @since NeXtMidas 2.9.3
   */
  @Override
  public String getStyle () {
    String style = getParameter("ENV.STYLE");
    if (style == null) return "BROWSER";
    return style;
  }

  /**  @since NeXtMidas 3.3.0 */
  @Override
  public Terminal getTerminal () {
    return terminal;
  }

  //////////////////////////////////////////////////////////////////////////////
  // End implement ShellParent interface
  //////////////////////////////////////////////////////////////////////////////

  //////////////////////////////////////////////////////////////////////////////
  // Begin implement ComponentListener interface
  //////////////////////////////////////////////////////////////////////////////

  @Override public void componentShown   (ComponentEvent e) { process("WINDOW",1,"SHOWN");   }
  @Override public void componentHidden  (ComponentEvent e) { process("WINDOW",-1,"HIDDEN"); }
  @Override public void componentMoved   (ComponentEvent e) {  }
  @Override public void componentResized (ComponentEvent e) {  }

  //////////////////////////////////////////////////////////////////////////////
  // End implement ComponentListener interface
  //////////////////////////////////////////////////////////////////////////////

  public void setMWindow (MWindow mw) {
    MW = mw;
    if (shown) process ("WINDOW",1,"OPENED");
  }

  private void process (String name, int info, Object data) {
    if (MW!=null) MW.process(name,info,data);
  }

  /** Send a message to the main MWindow that is on this applet. */
  public void msg (String name, int info, String value) {
    if (MW != null) MW.process (name,info,value);
  }

  /** Send a message to the specified id and process it. */
  public void msg (String name, int info, String value, String id) {
    MessageHandler handler = registry.getMessageHandler(id,5);
    Message msg = new Message(name,info,value,handler,this);
    if (handler!=null) handler.processMessage (msg);
  }

  /** Runs the specified command in a new Shell, where this applet will be conduit. */
  public void cmd (String cmd, int flag) {
    if (cmd.length()<=0) return;
    conduit = true; // I am being used to possible start another applet
    shell = Shell.runNew (this,cmd,true);
  }

  /** Runs the specified command in a new Shell */
  public void cmd (String cmd) {
    if (cmd.length()<=0) return;
    shell = Shell.runNew (this,cmd,true);
  }

  /** Put specified id(key) / object(value) into this applet's registry */
  public void register (String id, Object object) {
    registry.put(id,object);
  }

  @Override
  public int processMessage (Message msg) {
    if (queue==null) queue = new MessageQueue(this);
    queue.put(msg);
    return 1;
  }

  /** Gets the next message on this applet's MessageQueue (null if none). */
  public Message getMessage () {
    if (queue==null) return null;
    return queue.get();
  }

  public String getMessageKeyValuePair () {
    if (queue==null) return null;
    Message msg = queue.get();
    if (msg==null) return null;
    String string = msg.name+"="+msg.data.toString();
    return string;
  }

  @Override
  public String getID () {
    return registryName;
  }

  private static String[] keys = {"ID","REG"};

  @Override
  public String[] getKeys() { return keys; }

  @Override
  public Object setKey (String name, Object value) {
    if (name.equals("ID")) registryName = (String)value;
    else if (name.equals("REG")) registry.table = (Table)value;
    else return null;
    return value;
  }

  @Override
  public Object getKey (String name) {
    Object value = null;
    if (name.equals("ID")) value = registryName;
    else if (name.equals("REG")) value = registry.table;
    return value;
  }

  /** Gets the current applet, or null if no applet is running. */
  public static MApplet getCurrent() {
    return currentApplet;
  }

  /** Invoke JavaScript function from web page this applet is embedded on from Java/Macro.
      @param jsMethodName JavaScript method name to call (NOTE: JavaScript is Case-Sensitive)
      @param args Table of arguments to call JavaScript method with; (keys are ignored),
                  values in the Table are the arguments in order specified (and it is cast to a String),
                  use null if method has no arguments.
      @return result of calling the specified
      @see #callJavaScript(String, Object...)
   */
  public Object setJScript (String jsMethodName, Table args) throws MidasException {
    Table.Iterator ti=args.iterator();
    String[] sargs = new String[ti.size];
    for (int i=0; i<ti.size; i++) {
      sargs[i] = (String)ti.values[i];
    }
    if (ti.size==0) sargs=null;
    return callJavaScript(jsMethodName, (Object)sargs);
  }

  /** Invoke JavaScript function from web page this applet is embedded on from Java/Macro.
      @param jsMethodName JavaScript method name to call (NOTE: JavaScript is Case-Sensitive)
      @param args arguments to call JavaScript method with
      @return result of calling the specified
      @since NeXtMidas 3.3.0
      @see #evalJavaScript(String)
   */
  public Object callJavaScript (String jsMethodName, Object... args) throws MidasException {
    Object rtnobj = null;
    debug("callJavaScript() calling methodName = " + jsMethodName+" with args = "+ Arrays.toString(args));
    if (jswin != null && jsoCallMethod != null) {
      try {
        // rtnobj = ((JSObject)jswin).call(jsMethodName, args);
        rtnobj = jsoCallMethod.invoke(jswin, new Object[] {jsMethodName, args});
        debug("called method returned following: "+rtnobj);
      } catch (Exception e) {
        throw new MidasException("Unable to call JavaScript method ["+jsMethodName+"]", e);
      }
    }
    return rtnobj;
  }

  /** Evaluates a JavaScript expression on web page this applet is embedded on from Java/Macro.
      @param expression JavaScript expression to evaluate on web page (NOTE: JavaScript is Case-Sensitive)
      @return result of evaluating the specified JavaScript expression
      @since NeXtMidas 3.3.0
   */
  public Object evalJavaScript (String expression) throws MidasException {
    Object rtnobj = null;
    debug("evalJavaScript() expression = " + expression);
    if (jswin != null && jsoEvalMethod != null) {
      try {
        // rtnobj = ((JSObject)jswin).eval(expression);
        rtnobj = jsoEvalMethod.invoke(jswin, expression);
        debug("evaluated expression returned following: "+rtnobj);
      } catch (Exception e) {
        throw new MidasException("Unable to evaluate JavaScript expression ["+expression+"]", e);
      }
    }
    return rtnobj;
  }

  /** Initialize Java to JavaScript interface. This allows Java/Macro to call
      JavaScript methods on web page that this applet is embedded on.
      @since NeXtMidas 3.3.0
   */
  @SuppressWarnings("deprecation")
  private void initJavaScriptIF() {
    try {
      // use reflection so we don't need to include <java6u10+>/jre/lib/plugins.jar in CLASSPATH
      jswin = KeyObject.invokeStaticMethod("netscape.javascript.JSObject", "getWindow", new Object[] {this});

      if (jswin == null) {
        System.out.println("WARN: initJavaScriptIF() - unable to obtain JavaScript Object of window this applet is embedded on.");
      }
      else {
        debug("initJavaScriptIF() - got JavaScript object for window = "+jswin);

        jsoCallMethod = Util.getMethod("netscape.javascript.JSObject.call(java.lang.String,java.lang.Object[])", false);
        jsoEvalMethod = Util.getMethod("netscape.javascript.JSObject.eval(java.lang.String)", false);
        debug("initJavaScriptIF() - call Method = "+jsoCallMethod);
        debug("initJavaScriptIF() - eval Method = "+jsoEvalMethod);
     }

    } catch (Exception e) {
      System.err.println("WARN: initJavaScriptIF() - could not initialize Javacript Object(s), caught exception: " + e);
      e.printStackTrace();
    }
  }

  /** print debug message to System.out if debug is enabled via "DEBUG" parameter */
  private void debug (String msg) { // since 3.3.0
    if (debug) System.out.println(msg + " ["+ registryName +"]");
  }
}
