package nxm.sys.libg;

import nxm.sys.lib.Shell;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.*;

import nxm.sys.libg.LargeJTable.LargeTableModel;

/** TableSorter is a decorator for TableModels; adding sorting functionality to a supplied
    TableModel. TableSorter does not store or copy the data in its TableModel; instead it
    maintains a map from the row indexes of the view to the row indexes of the model. As
    requests are made of the sorter (like getValueAt(row, col)) they are passed to the
    underlying model after the row numbers have been translated via the internal mapping
    array. This way, the TableSorter appears to hold another copy of the table with the rows
    in a different order. <br>
    <br>
    Clicking the mouse clears any previous sorting state and advances the sorting status of
    that column through three values: {NOT_SORTED, ASCENDING, DESCENDING} (then back to
    NOT_SORTED again). <br>
    <br>
    <i>The original version of this class was provided in the "How to Use Tables" section of "The
    Java Tutorial": <a href="http://java.sun.com/docs/books/tutorial/uiswing/components/table.html">
    http://java.sun.com/docs/books/tutorial/uiswing/components/table.html</a></i> However it has
    been heavily revised since then.

    @author  P. Milne, B. McLean, D. van Enckevort, P. Sekhon (Java Tutorial Version 2.0, 02/27/04)
    @author  J. Czechowski IV (revised version)
    @version $Id: TableSorter.java,v 1.28 2015/08/27 15:26:50 ntn Exp $
 */
public class TableSorter extends AbstractTableModel implements LargeTableModel {
  /* JC4: Implementation note:
            For performance reasons the realIndex and viewIndex maps are never generated for
            DESCENDING order, rather they are always generated for ASCENDING order. This is
            to provide for quick switching between ASCENDING and DESCENDING. When viewed in
            DESCENDING order the ASCENDING order maps are simply read in reverse (i.e. use
            size-1-index as the index into the map).

            Additionally, the maps are kept around even after switching back to NOT_SORTED
            mode so that they will be immediately available when returning to a sorted mode.
            The maps are only cleared in two situations (1) the "real" model changes and
            (2) when clearSortingState() is called.

          Suggested Future Enhancements:
            (1) Modification of a single row in a table should not result in resorting of
                the entire table.
            (2) Modification of a cell that is not in the sort-by column should have no
                affect on the sorting.
            (3) If the table remains in the NOT_SORTED state for a long time (2+ min) the
                clearSortingState() could be called to release the memory used by the maps.
  */
  /** The table is sorted in descending order. */
  public static final int DESCENDING = -1;
  /** The table is not sorted. */
  public static final int NOT_SORTED = 0;
  /** The table is sorted in ascending order. */
  public static final int ASCENDING  = 1;

  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2012021020311L; // for Serializable interface

  private LargeJTable        table         = null;         // The table
  private LargeTableModel    tableModel    = null;         // The "real" table model
  private JTableHeader       tableHeader   = null;         // Renders the header with sort arrows
  private MouseListener      mouseListener = null;         // Listens for mouse events to set sort
  private TableModelListener modelListener = null;         // Listens to events in the "real" model
  private Comparator<Number> sorter        = null;         // Sorts the table in ASCENDING order
  private Symbol             upTriangle    = null;         // UP arrow symbol
  private Symbol             dnTriangle    = null;         // DOWN arrow symbol
  private int                maxSize       = 0;            // The maximum sortable table size
  private boolean            warnMaxSize   = true;         // Warn the user when above maxSize
  private long               column        = 0L;           // The index of the sort column
  private int                direction     = NOT_SORTED;   // The sort direction
  private boolean            upToDate      = false;        // Indicates if the maps are up-to-date
  private Vector<Number>     realIndex     = new Vector<Number>(); // Maps view -> real
  private Vector<Number>     viewIndex     = new Vector<Number>(); // Maps real -> view
  private boolean            allowSorting  = true;         // Indicates if sorting is allowed.

  /** Creates a new instance.
      @param table       The actual table.
      @param tableModel  The "real" table model that accesses the file.
      @param tableHeader Renders the header with sort arrows.
      @param maxSize     The maximum table size that is sortable.
   */
  public TableSorter(LargeJTable table, LargeTableModel tableModel,
                     JTableHeader tableHeader, int maxSize) {
    this.table         = table;
    this.mouseListener = new MouseHandler();
    this.modelListener = new TableModelHandler();
    this.sorter        = new TableValueComparator();
    this.maxSize       = maxSize;
    this.warnMaxSize   = (maxSize > 0);
    this.upTriangle    = new Symbol.UTriangle(); upTriangle.setFilled(true);
    this.dnTriangle    = new Symbol.DTriangle(); dnTriangle.setFilled(true);
    setTableHeader(tableHeader);
    setTableModel(tableModel);
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // LargeTableModel Methods
  //////////////////////////////////////////////////////////////////////////////////////////////////
  public void setValueAt(Object val, long row, long col) {
    tableModel.setValueAt(val, toReal(row), col);
  }

  public boolean isCellEditable(long row, long col) {
    return tableModel.isCellEditable(toReal(row), col);
  }

  public Object getValueAt(long row, long col) {
    return tableModel.getValueAt(toReal(row), col);
  }

  public int getRowCount() {
    return tableModel.getRowCount();
  }

  public int getColumnCount() {
    return tableModel.getColumnCount();
  }

  public String getColumnName(int column) {
    return tableModel.getColumnName(column);
  }

  public Class<?> getColumnClass(int column) {
    return tableModel.getColumnClass(column);
  }

  public boolean isCellEditable(int rowIndex, int columnIndex) {
    return isCellEditable(table.getRealRowIndex(rowIndex), table.getRealColIndex(columnIndex));
  }

  public Object getValueAt(int rowIndex, int columnIndex) {
    return getValueAt(table.getRealRowIndex(rowIndex), table.getRealColIndex(columnIndex));
  }

  public void setValueAt(Object val, int rowIndex, int columnIndex) {
    setValueAt(val, table.getRealRowIndex(rowIndex), table.getRealColIndex(columnIndex));
  }

  public void fireNeedsRedraw(long top, long btm) {
    if (isSorting()) {
      getTableModel().fireNeedsRedraw(-1L, -1L);
    } else {
      getTableModel().fireNeedsRedraw(top, btm);
    }
  }

  public void fireRowsInserted(long top, long btm) {
    if (isSorting()) {
      /* It's faster to refresh all the rows than to figure out which
         rows need to be refreshed based on their current ordering.
      */
      getTableModel().fireRowsInserted(-1L, -1L);
    } else {
      getTableModel().fireRowsInserted(top, btm);
    }
  }

  public void fireRowsDeleted(long top, long btm) {
    if (isSorting()) {
      /* It's faster to refresh all the rows than to figure out which
         rows need to be refreshed based on their current ordering.
      */
      getTableModel().fireRowsDeleted(-1L, -1L);
    } else {
      getTableModel().fireRowsDeleted(top, btm);
    }
  }

  public void fireStructureChanged() {
    getTableModel().fireStructureChanged();
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Public Methods
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Gets the "normal" table model. */
  public LargeTableModel getTableModel() {
    return tableModel;
  }

  /** Gets the table header. */
  public JTableHeader getTableHeader() {
    return tableHeader;
  }

  /** Sets the table header. */
  public void setTableHeader(JTableHeader tableHeader) {
    if (this.tableHeader != null) {
      this.tableHeader.removeMouseListener(mouseListener);
      TableCellRenderer defaultRenderer = this.tableHeader.getDefaultRenderer();
      if (defaultRenderer instanceof SortableHeaderRenderer) {
        this.tableHeader.setDefaultRenderer(((SortableHeaderRenderer) defaultRenderer).renderer);
      }
    }
    this.tableHeader = tableHeader;
    if (this.tableHeader != null) {
      this.tableHeader.addMouseListener(mouseListener);
      this.tableHeader.setDefaultRenderer(
              new SortableHeaderRenderer(this.tableHeader.getDefaultRenderer()));
    }
  }

  /** Indicates if sorting is being done at this time. */
  public boolean isSorting() {
    return (direction != NOT_SORTED);
  }

  /** Gets the sorting status for a given column. */
  public int getSortingStatus(int column) {
    return (column == this.column)? direction: NOT_SORTED;
  }

  /** Sets the sorting status.
      @param column    The column to sort by.
      @param direction One of {@link #NOT_SORTED}, {@link #ASCENDING}, {@link #DESCENDING}.
   */
  public void setSortingStatus(int column, int direction) {
    if ((column < 0) || !allowSorting) {
      direction = TableSorter.NOT_SORTED;
      column    = 0;
    }

    if ((this.direction == NOT_SORTED) && (direction == NOT_SORTED)) {
      if (this.column != column) {
        clearSortingState();
        this.column = column;
      }

    } else if (table.getNumberOfRows() > maxSize) {
      if (warnMaxSize) {
        Shell.warning("WARNING: Table is too large to sort."); // XXX Change this

        warnMaxSize = false; // Only issue the warning once.
      }

    } else if (this.column == column) {
      if (this.direction != direction) {
        this.direction = direction;
        sortingStatusChanged();
      }
    } else {
      clearSortingState();
      this.column    = column;
      this.direction = direction;
      sortingStatusChanged();
    }
  }

  /** Sets the sorting status.
      @param column  The name of the column to sort by prefixed by a '+' for ASCENDING or a
                     '-' for DESCENDING. Use a null string or "" to disable sorting.
   */
  public void setSortingStatus(String column) {
    int  dir = ASCENDING;  // Sort direction
    long col = -1;         // Sort column

    if ((column != null) && (column.length() > 1)) {
      if (column.startsWith("+")) {
        dir = ASCENDING;
        col = table.getColumnIndex(column.substring(1));
      } else if (column.startsWith("-")) {
        dir = DESCENDING;
        col = table.getColumnIndex(column.substring(1));
      } else {
        dir = ASCENDING;
        col = table.getColumnIndex(column);
      }
    } else if ((column != null) && (column.length() > 0)) {
      dir = ASCENDING;
      col = table.getColumnIndex(column);
    }

    setSortingStatus((int)col, dir);
  }

  /** Enables or disables sorting.
      @param enabled If this is false all sorting is disabled.
   */
  public void setSortingEnabled(boolean enabled) {
    if (enabled != allowSorting) {
      allowSorting = enabled;
      if (!allowSorting) { setSortingStatus(null); } // Clears previous sorting status.
    }
  }

  /** Indicates if sorting is enabled. */
  public boolean getSortingEnabled() {
    return allowSorting;
  }

  /** Gets the sorting status.
      @return The name of the column to sort by prefixed by a '+' for ASCENDING or a
              '-' for DESCENDING. This will return "" if NOT_SORTED.
   */
  public String getSortingStatus() {
    String status = "";

         if (direction == ASCENDING ) { status = "+" + getColumnName((int)column); }
    else if (direction == DESCENDING) { status = "-" + getColumnName((int)column); }
    else                              { status = "";                               }

    return status;
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Private Methods
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Get sort index. This is only used by toReal(long) and toView(long). */
  private long getSortIndex(Vector<Number> mapping, long row) {
    if (direction != NOT_SORTED) {
      int index = (direction == ASCENDING)? (int)row : (int)(table.getNumberOfRows() - 1 - row);

      if (!isUpToDate()) { resort(); }

      row = mapping.get(index).longValue();
    }
    return row;
  }


  /** Converts sorted index to real index. */
  long toReal(long row) {
//System.out.println("  toReal("+row+") = "+getSortIndex(realIndex, row));
    return getSortIndex(realIndex, row);
  }

  /** Converts real index to sorted index. */
  long toView(long row) {
    return getSortIndex(viewIndex, row);
  }

  /** Gets the Arrow icon used when rendering a given column.
      @param column The column number.
      @param size   The size of the icon in pixels (note the icon will be square).
   */
  protected Icon getHeaderRendererIcon(int column, int size) {
    int direction = getSortingStatus(column);
    Symbol icon = null;

         if (direction == ASCENDING ) { icon = upTriangle; }
    else if (direction == DESCENDING) { icon = dnTriangle; }

    if (icon != null) { icon.setSize(size/2); }

    return icon;
  }

  /** Resorts the column. */
  private synchronized void resort() {
    if (!isUpToDate()) {
      realIndex.setSize((int)table.getNumberOfRows());
      viewIndex.setSize((int)table.getNumberOfRows());

      /* Sort the realIndex. */
      for(int i = 0; i < realIndex.size(); i++) {
        realIndex.set(i, Integer.valueOf(i));
      }
      Collections.sort(realIndex, sorter);

      /* Build the viewIndex from the realIndex. */
      for(int i = 0; i < viewIndex.size(); i++) {
        int index = realIndex.get(i).intValue();

        viewIndex.set(index, Integer.valueOf(i));
      }

      upToDate = true;
    }
  }

  /** Gets the up-to-date status. */
  private synchronized boolean isUpToDate() {
    return upToDate && (table.getNumberOfRows() == realIndex.size());
  }

  /** Clears the sorting state but does not refresh the display. */
  private synchronized void clearSortingState() {
    direction = NOT_SORTED;
    upToDate  = false;
    column    = 0L;
    realIndex.clear();
    viewIndex.clear();
  }

  /** Sets the sort direction to NOT_SORTED, discards the current sort information and refreshes
      the display.
   */
  private void clearAll() {
    clearSortingState();
    sortingStatusChanged();
  }

  /** Sets the sort direction to NOT_SORTED and refreshes the display but does not discard the
      current sort information.
   */
  private void cancelSorting() {
    if (direction != NOT_SORTED) {
      direction = NOT_SORTED;
      sortingStatusChanged();
    }
  }

  /** Clears the sorting state and refreshes the display. */
  private void invalidateSorting() {
    if (upToDate) {
      upToDate = false;
      sortingStatusChanged();
    }
  }

  /** Makes appropriate updates to reflect changes in the sorting state. */
  private void sortingStatusChanged() {
    fireTableDataChanged();
    if (tableHeader != null) {
      tableHeader.repaint();
      table.getRowHeader().repaint();
    }
  }

  /** Sets the "normal" table model.
      @param tableModel The table model.
   */
  public void setTableModel(LargeTableModel tableModel) {
    if (this.tableModel != null) {
      this.tableModel.removeTableModelListener(modelListener);
    }

    this.tableModel = tableModel;
    if (this.tableModel != null) {
      this.tableModel.addTableModelListener(modelListener);
    }

    clearSortingState();
    sortingStatusChanged();
  }

  /** Compares two rows. */
  @SuppressWarnings({ "unchecked", "rawtypes" })
  private int compareRows(long row1, long row2) {
    int returnValue = 0;

    Object val1 = tableModel.getValueAt(row1,column);
    Object val2 = tableModel.getValueAt(row2,column);

    if (val1 instanceof String) {  // Fix to Bug 1973 ClassCastException for when val1 is String vs val2 is Data or a non-String
      returnValue = ((String) val1).compareTo(val2.toString());
    }
    else if (val1 instanceof Comparable) {
      returnValue = ((Comparable) val1).compareTo(val2);
    }
    else {
      returnValue = val1.toString().compareTo(val2.toString());
    }

    return returnValue;
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Nested Classes
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** An implementation of the {@link Comparator} interface that takes in two numbers (as row
      numbers in the table) and compares the corresponding values in the table.
   */
  protected class TableValueComparator implements Comparator<Number> {
    public int compare(Number o1, Number o2) {
      long row1 = o1.longValue();
      long row2 = o2.longValue();

      return compareRows(row1, row2);
    }
  }

  /** Handles mouse events for the one-click-sorting. */
  private class MouseHandler extends MouseAdapter {
    public void mouseClicked(MouseEvent e) {
      JTableHeader h = (JTableHeader) e.getSource();
      TableColumnModel columnModel = h.getColumnModel();
      int viewColumn = columnModel.getColumnIndexAtX(e.getX());
      if (viewColumn == -1) return; // x is out of bounds, Bug 843

      int column = columnModel.getColumn(viewColumn).getModelIndex();
      if (column != -1) {
        int status = getSortingStatus(column);
        if (!e.isControlDown()) {
          cancelSorting();
        }
        // Cycle the sorting states through {NOT_SORTED, ASCENDING, DESCENDING} or
        // {NOT_SORTED, DESCENDING, ASCENDING} depending on whether shift is pressed.
        status = status + (e.isShiftDown() ? -1 : 1);
        status = (status + 4) % 3 - 1; // signed mod, returning {-1, 0, 1}
        setSortingStatus(column, status);
      }
    }
  }

  /** Renders the table header with the appropriate direction icon. */
  private class SortableHeaderRenderer implements TableCellRenderer {
    private TableCellRenderer renderer; // The default renderer.

    /** Creates a new instance.
        @param renderer The default renderer (without a direction icon).
     */
    public SortableHeaderRenderer(TableCellRenderer renderer) {
      this.renderer = renderer;
    }

    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSel,
                                                   boolean hasFocus, int row, int col) {
      Component comp = renderer.getTableCellRendererComponent(table,value,isSel,hasFocus,row,col);
      if (comp instanceof JLabel) {
        int modelColumn = table.convertColumnIndexToModel(col);

        JLabel jlabel = ((JLabel)comp);
        jlabel.setHorizontalTextPosition(JLabel.LEFT);
        jlabel.setIcon(getHeaderRendererIcon(modelColumn, jlabel.getFont().getSize()));
      }
      return comp;
    }
  }

  /** The listener used to listen for changes in the "real table model". */
  private class TableModelHandler implements TableModelListener {
    public void tableChanged(TableModelEvent e) {
      if (!isSorting()) {
        /* If we're not sorting by anything, just pass the event along. */
        fireTableChanged(e);

      } else if (e.getFirstRow() == TableModelEvent.HEADER_ROW) {
        /* If the table structure has changed, clear everything. */
        // JC4: This was causing issues when scrolling, see Bug 427 (24APR06)
        //clearAll();

      } else {
        /* Something has happened to the data that may have invalidated the sort order. */
        boolean invalidate = true;

        if (e.getFirstRow() == e.getLastRow()) {
          long numRows = table.getNumberOfRows();

          if (numRows == realIndex.size()) {
            long rowNum = toReal(e.getFirstRow());

            if (numRows <= 1) {
              invalidate = false;
            } else if (rowNum == 0L) {
              invalidate = (compareRows(rowNum, rowNum+1) > 0);
            } else if (rowNum == numRows-1) {
              invalidate = (compareRows(rowNum-1, rowNum) > 0);
            } else {
              invalidate = (compareRows(rowNum-1, rowNum) > 0)
                        || (compareRows(rowNum, rowNum+1) > 0);
            }
          }
        }
        if (invalidate) { invalidateSorting(); }
      }
    }
  }
}
