package nxm.sys.libm;

import nxm.sys.libm.MMath;
import nxm.sys.inc.Constants;
/**
    Defines interface for arithmetic operators.

    This class allows a subclass to be fully usable by the NeXtMidas primitives
    <code>CALC</code> and <code>FCALC</code>.
    primitives. <br>
    <br>
    See the NeXtMidas help/explain file for the <code>GENERATE OPER</code>
    primitive for details on how to automatically generating an operator
    subclass.<br>
    <br>

    <a name="iomodes"><b>I/O Modes:</b></a>
       The following table summarizes the available I/O modes for use:
       <center><table border="1" summary="">
         <tr>
           <td align="center"><b>Mode</b></td>
           <td align="left"><b>Description</b></td>
         </tr>
         <tr>
           <td align="center">{@link #S}</td>
           <td align="left">Single scalar operand.</td>
         </tr>
         <tr>
           <td align="center">{@link #C}</td>
           <td align="left">Single complex operand.</td>
         </tr>
         <tr>
           <td align="center">{@link #V}</td>
           <td align="left">Single vector operand.</td>
         </tr>
          <tr>
           <td align="center">{@link #G}</td>
           <td align="left">Single complex-conjugate operand.</td>
         </tr>

       </table></center>
    <br>

    @author Jeff Schoen
    @version $Id: Operator.java,v 1.9 2014/03/25 14:48:24 nmg Exp $
*/
public class Operator extends Object implements Constants {

  /** data input/output modes S-scalar, C-complex, G-complexConjugate, V-vector(XYZ) */
  public static final int S=0x1, C=0x2, V=0x3, G=0xA;
  public static final int SS=0x11, CS=0x12, SC=0x21, CC=0x22, CG=0xA2, VS=0x13, VV=0x33;
  public static final int SSS=0x111, CSC=0x212, SCC=0x221, CCC=0x222, GCC=0xA22, CGC=0x2A2;
  public static final int VSV=0x313, VVV=0x333, CCS=0x122, VVS=0x133, SSC=0x211, CSS=0x112;
  public static final int SSSS=0x1111, SSSV=0x3111, VSSS=0x1113;

  /** modifier modes */
  public static final int CNJA=0x8, CNJB=0x80, CNJC=0x800;
  public static final int SNGA=0x4, SNGB=0x40, SNGC=0x400;

  /** Return value indicating that a given function is not supported.
      @see #process(double[][],int,int)
      @see #process(double[][],int,int,double)
      @see #process(float[][],int,int)
      @see #process(float[][],int,int,double)
      @see #process(int[][],int,int)
      @see #process(int[][],int,int,double)
      @see #validate(char,int,int)
   */
  public static final int NOSUPPORT=-1;

  /** Any processing flags to use. */
  protected int flags;

  /** The scale factor to apply. */
  protected double scale;

  /** Creates a new instance. <br>
      This is the same as using <code>Operator(0, 1.0)</code>.
      @see #Operator(int,double)
   */
  public Operator() {
    this (0,1.0);
  }

  /** Creates a new instance. <br>
      This is the same as using <code>Operator(flags, 1.0)</code>.
      @param flags Any processing flags to use.
      @see #Operator(int,double)
   */
  public Operator(int flags) {
    this (flags,1.0);
  }

  /** Creates a new instance. <br>
      This is the same as using <code>Operator(0, scale)</code>.
      @param scale The scale factor to use.
      @see #Operator(int,double)
   */
  public Operator(double scale) {
    this (0,scale);
  }

  /** Creates a new instance.
      @param flags Any processing flags to use.
      @param scale The scale factor to use.
   */
  public Operator(int flags, double scale) {
    setFlags (flags);
    setScale (scale);
  }

  /** Apply this operator to <code>double</code> arrays.
      @param v     The arrays to apply the operator to.
      @param n     The number of elements to operate on.
      @param mode  The input/output mode (see <a href="#iomodes">I/O Modes</a>).
      @return The value of <code>n</code> if this method is supported or
              {@link #NOSUPPORT} if the operation is not supported.
      @see #process(double[][],int,int,double)
   */
  public int process (double[][] v, int n, int mode) {
    return NOSUPPORT;
  }

  /** Apply this operator to <code>double</code> arrays.
      @param v     The arrays to apply the operator to.
      @param n     The return value to return if the computation completes.
      @param mode  The input/output mode (see <a href="#iomodes">I/O Modes</a>).
      @param scale The scale factor to apply.
      @return The value of <code>n</code> if this method is supported or
              {@link #NOSUPPORT} if the operation is not supported.
      @see #process(double[][],int,int)
   */
  public int process (double[][] v, int n, int mode, double scale) {
    return NOSUPPORT;
  }

  /** Apply this operator to <code>float</code> arrays.
      @param v     The arrays to apply the operator to.
      @param n     The return value to return if the computation completes.
      @param mode  The input/output mode (see <a href="#iomodes">I/O Modes</a>).
      @return The value of <code>n</code> if this method is supported or
              {@link #NOSUPPORT} if the operation is not supported.
      @see #process(float[][],int,int,double)
   */
  public int process (float[][] v, int n, int mode) {
    return NOSUPPORT;
  }

  /** Apply this operator to <code>float</code> arrays.
      @param v     The arrays to apply the operator to.
      @param n     The return value to return if the computation completes.
      @param mode  The input/output mode (see <a href="#iomodes">I/O Modes</a>).
      @param scale The scale factor to apply.
      @return The value of <code>n</code> if this method is supported or
              {@link #NOSUPPORT} if the operation is not supported.
      @see #process(float[][],int,int)
   */
  public int process (float[][] v, int n, int mode, double scale) {
    return NOSUPPORT;
  }

  /** Apply this operator to <code>int</code> arrays.
      @param v     The arrays to apply the operator to.
      @param n     The return value to return if the computation completes.
      @param mode  The input/output mode (see <a href="#iomodes">I/O Modes</a>).
      @return The value of <code>n</code> if this method is supported or
              {@link #NOSUPPORT} if the operation is not supported.
      @see #process(int[][],int,int,double)
   */
  public int process (int[][] v, int n, int mode) {
    return NOSUPPORT;
  }

  /** Apply this operator to <code>int</code> arrays.
      @param v     The arrays to apply the operator to.
      @param n     The return value to return if the computation completes.
      @param mode  The input/output mode (see <a href="#iomodes">I/O Modes</a>).
      @param scale The scale factor to apply.
      @return The value of <code>n</code> if this method is supported or
              {@link #NOSUPPORT} if the operation is not supported.
      @see #process(int[][],int,int)
   */
  public int process (int[][] v, int n, int mode, double scale) {
    return NOSUPPORT;
  }

  /** Get the number of inputs for the operator.
      @return The number of inputs for the operator
              or <code>-1</code> if the number of inputs is undefined.
   */
  public int getInputs() {
    return -1;
  }

  /** Get the number of outputs for the operator.
      @return The number of outputs for the operator
              or <code>-1</code> if the number of inputs is undefined.
   */
  public int getOutputs() {
    return -1;
  }

  /** Sets the processing flags to use.
      @param value Any processing flags to use.
   */
  public void setFlags (int value) {
    flags=value;
  }

  /** Sets the scale factor to use.
      @param value The scale factor to use.
   */
  public void setScale (double value) {
    scale=value;
  }

  /** Check to see if a particular configuration is supported.
      @param type  The type to use. <i>This is usually ignored.</i>
      @param n     The number of elements to operate on.
      @param mode  The input/output mode to use (see <a href="#iomodes">I/O Modes</a>).
      @return The value of <code>n</code> if the configuration is supported or
              {@link #NOSUPPORT} if the operation is not supported.
   */
  public int validate(char type, int n, int mode) {
    return NOSUPPORT;
  }
}
