package nxm.ucl.prim;

import nxm.sys.lib.*;
import nxm.sys.libg.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import nxm.ucl.lib.SpringUtilities;

/** Provides a way to display multiple time rastered plots
    of Type 1000 Midas files in a Scrollable Pane.

    @author  Lisa Tursky
    @author  Jay Hallam
    @version $Id: multiplot.java,v 1.18 2013/03/08 16:11:49 ntn Exp $
*/
public class multiplot extends GPrimitive implements KeyListener, MouseListener {
  /** Vertical scroll to options (value={@value}). See {@link #setScrollEnd(String)}. */
  public static final String vscrollList = "Off,Top,Bottom";

  private static final int VSCROLL_OFF    = 1; // Vertical scroll turned off
  private static final int VSCROLL_TOP    = 2; // Vertical scroll to top
  private static final int VSCROLL_BOTTOM = 3; // Vertical scroll to bottom

  private static final boolean DEBUG_ALIGN  = false;    // Used when debugging alignment issues.

  private boolean isDone     = false;       // Is the command done?
  private boolean dbg        = DEBUG_ALIGN; // Turn debug messages on?
  private boolean warn       = true;        // Turn warnings on?
  private int     vscrollEnd = VSCROLL_OFF; // Turn automatic scrolling sticky (OFF|BOTTOM|TOP)

  private String       switches;   // Switches for PLOT command.
  private String       params;     // Parameters (other than file name) for PLOT command.
  private JScrollPane  pane;       // The scroll pane to place panel in.
  private JPanel       panel;      // The top-level panel.
  private GridLayout   layout;     // The layout used by panel.

  private double      xMin;       // Minimum X-axis value.
  private double      xMax;       // Maximum X-axis value.
  private int         panWidth;   // Panel width in pixels (for alignment, may not match display).
  private int         plotHeight; // Panel height in pixels (for alignment, may not match display).
  private int         maxWidth;   // Max allowable width for the data area

  private MidasDisplay[] displays;

  private String idPrefix    = "MULTIPLOT_PLOT_"; // The prefix for the individual PLOT ids.

  public int open() {
    layout   = new GridLayout(1, 1);
    panel    = new JPanel(layout);
    pane     = new JScrollPane(panel);
    dbg      = MA.getState("/DEBUG", dbg);
    warn     = MA.getState("/WARN", warn);
    xMin     = MA.getD("/XMIN",   0.0);
    xMax     = MA.getD("/XMAX",  10.0);

    String title = MA.getS("/TITLE", "NeXtMidas Multiplot");

    MW = new MWindow(title, this);
    MW.open();
    MW.setTheme(MA.getO("/THEME"));

    int width  = MA.getL("WIDTH", 700);
    int height = MA.getL("HEIGHT", 500);
    maxWidth   = MA.getL("DATAWIDTH", 2500);
    if (width > maxWidth) width = maxWidth;

    MW.setSize(width,height);
    pane.setSize(width,height);

    int w = pane.getVerticalScrollBar().getWidth();
    int h = pane.getHorizontalScrollBar().getHeight();
    panel.setSize(MW.pos.w-w,MW.pos.h-h);
    panWidth   = MW.pos.w-w;
    plotHeight = 150;
    MW.addTo(this);

    panel.addKeyListener(this);
    panel.setFocusable(true);
    panel.addMouseListener(this);

    idPrefix = MA.getS("/IDPREFIX", idPrefix);

    info("my id "+this.id+" inner plots id prefix "+idPrefix);

    Args.SplitArg plotFiles = MA.getSplitArg(MA.getRaw("FILES"),'|');
    switches = MA.getS("SWITCHES","");
    switches = "/EXIT=NONE" + switches;
    params   = MA.getS("PARAMS","");
    displays = new MidasDisplay[plotFiles.getFields()];
    setScrollEnd(MA.getCS("/SCROLLEND","OFF"));

    // Create Midas Displays and add them to the panel.
    panel.addHierarchyBoundsListener(new HierarchyBoundsListener() {
      public void ancestorMoved(HierarchyEvent e)   { if (panel.isShowing()) { panel.validate(); } }
      public void ancestorResized(HierarchyEvent e) { if (panel.isShowing()) { panel.validate(); } }
    });
    pane.setVisible(true);
    MW.panel.add(pane);
    MW.setDisplay("FRONT");

    /* The frame and the MidasDisplay panels must be visible before creating
       the MidasBean's. If you do not do this you will experience refresh
       problems in PLOT.4.8828125E-4
       Create the beans that will occupy the given Midas Displays. */
    for (int i = 0; i < plotFiles.getFields(); i++) {
      addPlotToPanel(plotFiles.getField(i));
    }
    resize();
    return NORMAL;
  }

  public int process() {
    if (isDone) {
      return FINISH;
    } else {
      return NOOP;
    }
  }

  public int close() {
     return NORMAL;
  }

  public int processMessage(Message msg) {
    if        (msg.getName().equals("REFRESH")) {
      MW.refresh();
    } else if (msg.getName().equals("OPENFILE")) {
      addPlotToPanel(""+msg.data);
      resize();
    } else if (msg.getName().equals("RESIZE")) {
      resize();
    } else if (msg.getName().equals("VSCROLL")) {
      scrollVertical(""+msg.data);
    } else if (msg.name.equals("POP")) {
        MW.pop(msg.info);
    } else if (msg.getName().equals("WINDOW")) {
      if (msg.data.toString().equals("CLOSING")) {
        msg.name = "EXIT";
        MQ.put(msg);
      }
    }
    if (msg.getName().equalsIgnoreCase("EXIT")) {
      isDone=true;
      return FINISH;
    }

    return NORMAL;
  }

  /** Get the horizontal alignment information for a given file.
      @param filename The name of the file.
      @return Returns the start and width pixel values (in an array {start,width}). These do not
              take in to account the size of the plot axis.
   */
  private int[] getHorizAlign (String filename) {
    final int MIN_WIDTH = 20; // Minimum width allowed.

    DataFile hcb = new DataFile(M, filename,"", "", 0);
    hcb.open(BaseFile.INPUT);
    double xstart = hcb.getXStart();
    double xend   = xstart + (hcb.getXDelta() * hcb.getSize());
    double range  = (xMax - xMin);
    double mult   = (panWidth / range);
    hcb.close();

    int start = (int)((xstart - xMin) * mult);
    int end   = (int)((xend   - xMin) * mult);
    int width = end - start;

    debug("getHorizAlign("+filename+") xstart="+xstart+" xend="+xend+" range="+range+" mult="+mult);
    debug("  start="+start+" end="+end+" width="+width+" panWidth="+panWidth);
    if (start < 0) {
      warn("Can not properly align file "+filename+" with XS="+xstart+" when /XMIN="+xMin);
      start = 0;
    }
    if (width <= MIN_WIDTH) {
      warn("Can not properly display file "+filename+" with a range of "+range+" using minimum "
         + "allowable.");
      width = MIN_WIDTH;
    }

    return new int[] { start, width };
  }

  /** Get the number of plots that are in the panel. */
  private int getNumPlots() {
    return (panel==null)? 0: panel.getComponents().length;
  }

  /** Adds a given plot to the panel.
      @param filename The name of the plot.
   */
  private void addPlotToPanel(String filename) {
    final int LAXIS  = 16; // Width of left axis in pixels.
    final int RAXIS  = 22; // Width of right axis in pixels.

    // DETERMINE PLOT POSITION -------------------------------------------
    int[] align  = getHorizAlign(filename); // = { start, width }
    int dispWidth = align[1] + LAXIS + RAXIS;
    int leftPad   = align[0] - LAXIS;
    int rightPad  = panWidth - leftPad - dispWidth;

    // UPDATE GRID COUNTS ------------------------------------------------
    layout.setRows(getNumPlots() + 1); // Increase number of rows by one.

    // ADD PLOT TO DISPLAY -----------------------------------------------
    JPanel       subp = new JPanel(new SpringLayout()); // Subset panel for alignment.
    MidasDisplay disp = new MidasDisplay();             // The plot display.
    JPanel       padL = new JPanel();                   // Pad/spacer on left side of plot.
    JPanel       padR = new JPanel();                   // Pad/spacer on right side of plot.

    padL.setPreferredSize(new Dimension(Math.max(1, leftPad),  plotHeight));
    padR.setPreferredSize(new Dimension(Math.max(1, rightPad), plotHeight));
    disp.setPreferredSize(new Dimension(Math.max(1, dispWidth),plotHeight));

    if (DEBUG_ALIGN) {
      padL.setBackground(Color.red);
      padR.setBackground(Color.blue);
    }

    subp.add(padL);
    subp.add(disp);
    subp.add(padR);
    SpringUtilities.makeCompactGrid(subp, 1, 3, 0, 0, 0, 0);

    panel.setSize(MW.pos.w, plotHeight*getNumPlots());
    panel.add(subp);
    panel.validate();


    // RUN THE PLOT COMMAND ----------------------------------------------
    String    idPlot   = idPrefix + (getNumPlots() - 1);
    String    command  = "plot" + switches + " " + filename + params;
    MidasBean plotBean = MidasBean.createBean(M, command, disp, idPlot);
    scrollToVerticalSticky();
  }

  /** Split the file names on a '|' or a ','. */
  private String[] parseFileNames (String filenames) {
    return (filenames == null)? new String[0]: filenames.split("[|,]"); // Split on '|' or ','
  }

  /** Resize the scrollpane. */
  private void resize() {
    if (panel == null || pane == null) return;
    pane.setSize(MW.pos.w, MW.pos.h);
    MW.resize(0);
    pane.validate();
    if (panel.getWidth() > maxWidth) {
      panel.setPreferredSize(new Dimension(maxWidth, panel.getHeight()));
    }
    scrollToVerticalSticky();
    panel.repaint();
  }

  /** Get the vertical scroll end sticky mode for the multiplot scroll pane.
      @return return vertical scroll end sticky setting for the multiplot
   */
  public String getScrollEnd() {
    return Parser.get(vscrollList, vscrollEnd);
  }

  /** Set the vertical scroll end sticky mode for the multiplot scroll pane and then scroll to
      specified setting.
      @param scrollSticky vertical scroll sticky setting see {@link #vscrollList}.
   */
  private void setScrollEnd(int scrollSticky) {
    vscrollEnd = scrollSticky;
    scrollToVerticalSticky();
  }

  /** Set the vertical scroll end sticky mode for the multiplot scroll pane and then scroll to
      specified setting.
      @param scrollSticky vertical scroll sticky setting see {@link #vscrollList}.
   */
  public void setScrollEnd (String scrollSticky) {
    setScrollEnd(Parser.find(vscrollList,scrollSticky,vscrollEnd));
  }

  /** Vertical scroll to top, bottom, or none based on the current scrollEnd setting. */
  private void scrollToVerticalSticky() {
    scrollVertical(vscrollEnd);
  }

  /** Scroll to the specified vertical position in the scroll panel.
      @param vpos vertical scroll sticky position setting see {@link #vscrollList}.
   */
  private void scrollVertical(String vpos) {
    scrollVertical(Parser.find(vscrollList,vpos,vscrollEnd));
  }

  /** Scroll to the specified vertical position in the scroll panel.
      @param vpos vertical scroll sticky position setting see {@link #vscrollList}.
   */
  private void scrollVertical(int vpos) {
    if (pane.getComponentCount() < 1) {
      return; // cannot scroll pane when there are no child components, so return
    }

    int x = pane.getHorizontalScrollBar().getValue(); // save current x (horizontal) scroll position
    int h = panel.getHeight(); // get panel's height so that we scroll to bottom of the panel

    switch (vpos) {
      case VSCROLL_OFF:    /* Do Nothing */              break;
      case VSCROLL_TOP:
        pane.getVerticalScrollBar().setValue(0);
        pane.getHorizontalScrollBar().setValue(x);
        break;
      case VSCROLL_BOTTOM:
        pane.getVerticalScrollBar().setValue(h);
        pane.getHorizontalScrollBar().setValue(x);
        break;
      default: throw new MidasException("MULTIPLOT: Unknown vertical scroll position vpos="+vpos);
    }
  }

  private void debug (String str) {
    if (dbg) {
      System.out.println("MULTIPLOT: "+str);
    }
  }
  private void info (String str) {
    if (verbose) {
      M.info("MULTIPLOT: "+str);
    }
  }
  private void warn (String str) {
    if (warn) {
      M.warning("MULTIPLOT: "+str);
    }
  }

  public void keyPressed(KeyEvent e) {
    this.MW.keyPressed(e);
  }

  public void keyReleased(KeyEvent e) {
    this.MW.keyReleased(e);
  }

  public void keyTyped(KeyEvent e) {
    this.MW.keyTyped(e);
  }

  public void mouseClicked(MouseEvent e) {
    panel.requestFocus();
    MW.mouseClicked(e);
  }

  public void mouseEntered(MouseEvent e) {
    MW.mouseEntered(e);
  }

  public void mouseExited(MouseEvent e) {
    MW.mouseExited(e);
  }

  public void mousePressed(MouseEvent e) {
    MW.mousePressed(e);
  }

  public void mouseReleased(MouseEvent e) {
    MW.mouseReleased(e);
  }

}
