package nxm.ucl.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.net.Socket;
import java.net.ServerSocket;
import java.io.BufferedInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.Vector;

/**
  zpipe - NeXtMidas version of ZPIPE

  @author Jay Hallam
  @version $Id: zpipe.java,v 1.8 2014/04/15 15:58:28 ntn Exp $
*/
public class zpipe extends Primitive {

  private DataFile hcb;   // The input or output data file
  private Data data;      // The data read from file
  private int port;       // The socket port
  private String host;    // The hostname (client only)

  private Vector<SocketConnection> connections = new Vector<SocketConnection>(); // List of connections
  private ServerSocket server;               // The Server Socket
  private Socket client;                     // The Client Socket
  private InputStream inputStream = null;    // The input stream for reader
  private boolean isServer     = false;      // Write to a socket
  private boolean isSocketWriter = true;     // Write to a socket
  private double ConnAttemptInterval = 1.0;  // Maximum time to wait for connections
  private int NumBytesRead = 0;              //
  private double socketTimeout = 2.0;        // SO_TIMEOUT in seconds
  private boolean warn = true;               // Turn on/off all warnings
  public boolean debug = false;              // Turn on/off all debug

  public int open () {
    M.warning("This is a BETA version of ZPIPE.  Expect changes in future releases.  "
             +"Not all original X-Midas capabilities are implemented yet");
    // Get parameters
    debug = MA.getState("/DEBUG",debug);
    warn  = MA.getState("/WARN", warn);

    boolean reader = MA.getState("/XM_IN",  false);
    boolean writer = MA.getState("/XM_OUT", false);
    if (reader && writer) {
      M.error("You can not be both a socket reader and writer");
    }
    if (!reader && !writer) {
      isSocketWriter = true; //Default
    } else {
      isSocketWriter = writer;
    }

    host = MA.getCS("HOST", host);
    port = MA.getL("PORT", port);
    info("writer? "+isSocketWriter+" host "+host+" port "+port);

    ConnAttemptInterval = MA.getD("/ACCEPTTIME", ConnAttemptInterval);
    boolean reuseAddress = MA.getState("/REUSEADDR",true);
    socketTimeout       = MA.getD("/TIMEOUT",socketTimeout);
    String format = MA.getS("/FORMAT","SB");
    isServer = isEmptyString(host);
    isServer = MA.getState("/SERVER", isServer);
    int xfer = MA.getL("/TL",32768);
    boolean haveTLSwitch = MA.find("/TL");
    info("server? "+isServer);

    if  (isServer) {
      try {
        server = new ServerSocket(port);
        server.setReuseAddress(reuseAddress);
        server.setSoTimeout((int)(socketTimeout*1000.0)); // Timeout in millisecs
        acceptConnections();
      } catch (Exception e) {
        throw new MidasException("Unable to open server socket on port "+port, e);
      }
    } else {
      // Client
      client = connectToSocket(host, port);
    }

    // Open type-3000 file for INPUT
    hcb = MA.getDataFile("IN");
    hcb.setFormat(format);
    if (isSocketWriter) {
      M.info("Opening "+hcb.getURL()+" for INPUT");
      hcb.open(BaseFile.INPUT);
    } else {
      M.info("Opening "+hcb.getURL()+" for OUTPUT");
      hcb.open(BaseFile.OUTPUT);
    }
    if (!hcb.isOpen()) {
      M.warning ("'"+hcb.getFileName().getFullName()+"' does not exist");
      return FINISH;
    }
    // Set the transfer length if the user did not.
    if (!haveTLSwitch) xfer = Math.max(1,(int)(32768/hcb.dbpe));

    data = hcb.getDataBuffer(xfer);
    info("xfer "+xfer+" format "+format+" dbpe "+hcb.dbpe+" buffer length "+data.buf.length);

    return NORMAL;
  }

  public int process () {
    int status = NORMAL;
    if (isServer) {
      if (isSocketWriter ) {
        status = processServerSocketWriter();
      } else {
        warn("Server socket reader is TBD");
        status = FINISH;
      }
    } else {
      debug("process... client "+client);
      if (client == null) {
        client = connectToSocket(host, port);
        if (client == null) {
          Time.sleep(ConnAttemptInterval);
        }
      }
      if (client != null) {
        if (isSocketWriter) {
          M.warning("Client socket writer is TBD");
          status = FINISH;
        } else {
          status = processClientSocketReader();
        }
      }
    }
    return status;
  }

  public int close () {
    info("closing...");
    if (hcb != null) {
      hcb.close();
    }
    closeConnection(server);
    closeConnection(client);
    return NORMAL;
  }

  /* This is a client that reads a file/pipe and writes to a socket(s) */
  private int processClientSocketReader () {
    int status = NOOP;
    try {
      debug("reading from socket: have "+NumBytesRead+" of "+data.buf.length+" bytes");
      int totalBytesToRead = data.buf.length;
      int numToReadThisPass = totalBytesToRead - NumBytesRead;
      debug("reading from socket: bytes avail "+inputStream.available());
      int numRead = 0;
      //if (inputStream.available() > 0) {
        numRead = inputStream.read(data.buf, NumBytesRead, numToReadThisPass);
      //}
      NumBytesRead = NumBytesRead + numRead;
      debug("Num bytes read "+NumBytesRead);
      if (numRead > 0) {
        status = NORMAL;
      }
      if (NumBytesRead == totalBytesToRead) {
if (verbose) System.out.println("Writing to file NumBytesRead = "+NumBytesRead);
        hcb.write(data);
        NumBytesRead = 0;
      } else if (numRead < 0 && NumBytesRead != 0) {
        warn("End of socket: connection is in middle of read.");
        NumBytesRead = 0;
      }
    } catch (Exception e) {
      //jph -- TBD  -- count number of exceptions here and error out...
      warn("Exception "+e+" reading input stream");
      NumBytesRead = 0;
    }
    return status;
  }

  /* This is a server that reads a file/pipe and writes to a socket(s) */
  private int processServerSocketWriter () {
    int nelem = hcb.read(data);

    if (nelem > 0) {
      int numBytes = (int)hcb.getBPE()*nelem;
      int offset   = data.boff;
      debug("Got "+nelem+" elements, "+numBytes+" bytes");

      boolean wroteToSocket = false;
      //Iterate backwards so I can remove from connections vector
      for (int i =  connections.size()-1; i >=0; i--) {
        SocketConnection sc = connections.get(i);
        OutputStream     os = sc.getOutputStream();
        Socket sock = sc.getSocket();
        if (sock.isClosed()) {
          debug("------> SOCKET CLOSED removing a connection at "+i);
          connections.remove(i);
        } else if (!sock.isConnected()) {
          debug("------> SOCKET not CONNECTED removing a connection at "+i);
          connections.remove(i);
          closeConnection(sc);
        } else if (!sock.isBound()) {
          debug("------> SOCKET not BOUND removing a connection at "+i);
          connections.remove(i);
          closeConnection(sc);
        } else {
          try {
            os.write(data.buf, offset, numBytes);
            debug("Wrote "+nelem+"elements "+numBytes+" (bytes) to "+sc);
            wroteToSocket = true;
          } catch (Exception e) {
            e.printStackTrace();
            M.warning("Unable to write to socket output stream number "+i+" for "+sc);
            if (sock != null) {
              System.out.println("Connected? "+sock.isConnected()+" bound? "+sock.isBound()
                                +" closed? "+sock.isClosed());
              closeConnection(sc);
            }
          }
        }
      }
      if (wroteToSocket) {
        System.out.println("Wrote "+numBytes+" bytes to "+getNumConnections()+" socket(s)");
        return NORMAL;
      } else {
        return NOOP;    // Pause if nothing written to socket
      }
    } else {
      // no data
      debug("No data ... FINISHING");
      return FINISH;
    }
  }

  private Socket connectToSocket(String host, int port) {
    info("Attempting to connect to "+host+" port "+port);
    Socket sock = null;
    try {
      sock = new Socket(host, port);
      sock.setSoTimeout((int)(socketTimeout*1000.0));  // Timeout in millisecs
      // No need to do bind..
      inputStream = new BufferedInputStream (sock.getInputStream());
      info("Created socket to remotehost '"+host+" port "+port+"', client "+sock );
    } catch (Exception e) {
      debug("Unable to open client socket to host "+host+" port "+port+", is server up?");
    }
    return sock;
  }

  private void openConnection(Socket sock) {
    try {
      SocketConnection sc = new SocketConnection(sock);
      connections.add(sc);
    } catch (Exception e) {
      M.warning("Unable to open a connection for socket"+sock);
    }
  }

  /** Closes a socket connection. Socket can be null. */
  private void closeConnection(Socket sock) {
    if (sock != null) {
      try {
        sock.close();
      } catch (Exception e) {
        warn("Unable to socket connection to "+sock);
      }
    }
  }

  /** Closes a server socket connection. Socket can be null. */
  private void closeConnection(ServerSocket sock) {
    if (sock != null) {
      try {
        sock.close();
      } catch (Exception e) {
        warn("Unable to socket connection to "+sock);
      }
    }
  }

/* -------------------------------------------- Set/Get controls -------------------------------- */
  /** Return the number of connections */
  public int getNumConnections() {
    return connections.size();
  }

  /** Get the maximum time to wait for connections */
  public double getAcceptTime() {
    return ConnAttemptInterval;
  }
  /** Set the maximum time to wait for connections */
  public void setAcceptTime(double t) {
    ConnAttemptInterval = t;
  }


/* -------------------------------------------- Set/Get controls -------------------------------- */


  private synchronized void closeAllConnections() {
    for (SocketConnection socketConnection : connections) {
      closeConnection(socketConnection);
    }
    connections.clear();
  }

  private void closeConnection(SocketConnection sc) {
    M.info("Closing SocketConnection "+sc);
    Socket sock = sc.getSocket();
    if (!sock.isClosed()) {
      try {
        sock.close();
      } catch (Exception e) {
        M.warning("Unable to close connection to socket "+sc);
      }
    }
  }

  public void acceptConnections () {
    M.info("Starting ACCEPT connections thread, server closed? "+server.isClosed());
    (new ConnectionThread()).start();
  }


  private class ConnectionThread extends Thread {
    public ConnectionThread() {
      super("ConnectionThread"); // Give the thread a name
    }

    public void run () {
      M.info("ConnectionThread: running ...");
      while (!server.isClosed()) {
        try {
          M.info("Waiting for socket ...");
          Socket sock = server.accept();
          if (sock != null) {
            M.info("Accepted a socket "+sock);
            openConnection(sock);
          } else {
            M.info("No socket ... ");
          }
        } catch (Exception e) {
          System.out.println("Error trying to accept connection");
          e.printStackTrace();
        }
      }
    }
  }

  /** Inner class to hold the socket and output stream information */
  private class SocketConnection {
    private Socket       socket; // The socket to write on.
    private OutputStream os;   // The output stream to write to.

    public SocketConnection (Socket s) throws IOException{
      socket = s;
      os   = socket.getOutputStream();
    }

    /** Gets the output socket. */
    public Socket getSocket() {
      return socket;
    }

    /** Gets the output stream. */
    public OutputStream getOutputStream() {
      return os;
    }

    public String toString() {
      return "SocketConnection: socket="+socket+" address="+socket.getInetAddress();
    }

  }
  /* True is string is null, zero length or the string null or NULL. */
  private boolean isEmptyString(String str) {
    return (str == null || str.length() == 0 || str.equalsIgnoreCase("NULL"));
  }
  private void debug (String str) {
    if (debug) System.out.println("ZPIPE: "+str);
  }

  private void info(String str) {
    if (verbose) M.info("ZPIPE: "+str);
  }
  private void warn(String str) {
    if (warn) M.warning("ZPIPE: "+str);
  }
}
